/*
FILE:     _fileobj.C
*/
/*
VERSION:  8.120
*/
/*
DATE:     2/12/2008
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2008 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>

#include "ndb_misclib.h"
#include "_fileobj.h"
#include "erf_iolib.h"
#include "erf_util.h"
#include "_fileobj_ioutil.h"
#include "util.h"

static void get_symmetry_operators(CifString &oper, CifFileObj *fobj, const char *blockId);
static void get_pdb_chain_id(CifString &chn_id, ReVarCifArray<CifString>& pdb_chain_ids,
               ReVarCifArray<_PDB_ASYM_CORR> &chainids);

_FileObj::_FileObj(const int mxtokens, const int mxjrnltokens)
{
       memset(_input_filename, '\0', 200);
       memset(_output_filename, '\0', 200);
       memset(_real_filename, '\0', 200);
       memset(_temp_file, '\0', 200);
       memset(_StructureId, '\0', 100);
       _input_format = 0;
       _output_format = 0;
       _file_status = FILE_STATUS_NDB_PLUS_MOL;
       _experiment_type = EXPERIMENT_TYPE_BASIC;
       _num_mol = 0;
       _alloc_mol = 0;
       _rep_mol = 0;
       _CardInfo = new Cards_Info_Format[mxtokens];
       _JrnlCardInfo = new Cards_Info_Format[mxjrnltokens];
       _molecules = NULL;
       _entity_molecule = NULL;
       _CifObj = NULL;
       _Alloc_Seqs = 0;
       _Num_Seqs = 0;
       _Seqs = NULL;
       _Alloc_NPdb = 0;
       _Num_NPdb = 0;
       _NPdb = NULL;
       _Alloc_ICode = 0;
       _Num_ICode = 0;
       _ICode = NULL;
       _cell = NULL;
       _n_stbiol = 0;
       _stbiol = NULL;
       _alloc_ssbond = 0;
       _n_ssbond = 0;
       _ssbond = NULL;
       _error_message.Clear();
       _error_message_s.Clear();
       _unrecognized_residue.Clear();
       _is_binary = 0;
       for (int i = 0; i < MAX_NUM_ERROR; i++) {
            _Error_Check[i] = 0;
            _Error_Check_Message[i].Clear();
       }
       _chain_corr = NULL;
       _n_chain_corr = 0;
       _exist_pdb_chain_id = 1;
       _title_suppression = 0;
       _assign_chain_id_flag = 0;
       _is_exchange_format = 0;
       __maxit_version.Copy("2.3");
#ifdef IUPAC
       __maxit_version.Copy("3.0");
#ifdef EH99
       __maxit_version.Copy("3.1");
#endif
#endif
       __need_file_version_update = 0;
       __input_file_version = PDB_VERSION_23;
       __assembly_array.Clear();
       __encap_entity_poly = NULL;
       __encap_coordinates = NULL;
       __encap_ligand_info = NULL;
       __need_fix_atom_name = 0;
}

_FileObj::~_FileObj()
{
       int i;
       if (_error_message != "") {
            ndb_log_message(NDB_MSG_ERR, "%s\n", _error_message.Text());
       }
       if (_error_message_s != "") {
            ndb_log_message(NDB_MSG_ERR, "%s\n", _error_message_s.Text());
       }
       delete [] _CardInfo;
       delete [] _JrnlCardInfo;
       if (_molecules) {
            for (i = 0; i < _num_mol; i++)
                 if (_molecules[i]) delete _molecules[i];
            delete [] _molecules;
       }
       if (_entity_molecule) delete _entity_molecule;
       if(_CifObj) {
           if (_is_binary) _CifObj->CloseFile();
           delete _CifObj;
       }
       delete_Seqs();
       if (_cell) {
            if (_cell->symops) delete [] _cell->symops;
            delete _cell;
       }
       if (_stbiol) {
            for (i = 0; i < _n_stbiol; i++) if (_stbiol[i]) delete _stbiol[i];
            delete [] _stbiol;
       }
       if (strcmp(_temp_file, "")) {
            char command[100];
            struct stat stat1;
            if (!stat(_temp_file, &stat1)) {
                 sprintf(command, "rm %s", _temp_file);
                 system(command);
            }
       }
       if (_chain_corr) delete [] _chain_corr;
}

void _FileObj::_ManageSpace()
{
       int i = 0;
       if (_num_mol >= _alloc_mol) {
            _alloc_mol += 10;
            _MOLECULE **tmp = _molecules;
            _molecules = new _MOLECULE*[_alloc_mol];
            for (i = 0; i < _num_mol; i++) _molecules[i] = tmp[i];
            for (i = _num_mol; i < _alloc_mol; i++) _molecules[i] = NULL;
            delete [] tmp;
       }
       _molecules[_num_mol] = new _MOLECULE;
       _num_mol++;
}

void _FileObj::_ManageSeqSpace()
{
       int i = 0;
       if (_Num_Seqs >= _Alloc_Seqs) {
            _Alloc_Seqs += 10;
            SEQ** tmp = _Seqs;
            _Seqs = new SEQ*[_Alloc_Seqs];
            for (i = 0; i < _Num_Seqs; i++) _Seqs[i] = tmp[i];
            for (i = _Num_Seqs; i < _Alloc_Seqs; i++) _Seqs[i] = NULL;
            delete [] tmp;
       }
       _Seqs[_Num_Seqs] = new SEQ;
       _Seqs[_Num_Seqs]->ChainId.Clear();
       _Seqs[_Num_Seqs]->PolyType.Clear();
       _Seqs[_Num_Seqs]->AllocLen = 0;
       _Seqs[_Num_Seqs]->SeqLen = 0;
       _Seqs[_Num_Seqs]->chain_type = -1;
       _Seqs[_Num_Seqs]->is_used = 0;
       _Seqs[_Num_Seqs]->res = NULL;
       _Num_Seqs++;
}

void _FileObj::_ManagePdbSpace()
{
       int i = 0;
       if (_Num_NPdb >= _Alloc_NPdb) {
            _Alloc_NPdb += 10;
            SEQ** tmp = _NPdb;
            _NPdb = new SEQ*[_Alloc_NPdb];
            for (i = 0; i < _Num_NPdb; i++) _NPdb[i] = tmp[i];
            for (i = _Num_NPdb; i < _Alloc_NPdb; i++) _NPdb[i] = NULL;
            delete [] tmp;
       }
       _NPdb[_Num_NPdb] = new SEQ;
       _NPdb[_Num_NPdb]->ChainId.Clear();
       _NPdb[_Num_NPdb]->AllocLen = 0;
       _NPdb[_Num_NPdb]->SeqLen = 0;
       _NPdb[_Num_NPdb]->chain_type = -1;
       _NPdb[_Num_NPdb]->res = NULL;
       _Num_NPdb++;
}

void _FileObj::_ManageICodeSpace()
{
       int i = 0;
       if (_Num_ICode >= _Alloc_ICode) {
            _Alloc_ICode += 10;
            SEQ** tmp = _ICode;
            _ICode = new SEQ*[_Alloc_ICode];
            for (i = 0; i < _Num_ICode; i++) _ICode[i] = tmp[i];
            for (i = _Num_ICode; i < _Alloc_ICode; i++) _ICode[i] = NULL;
            delete [] tmp;
       }
       _ICode[_Num_ICode] = new SEQ;
       _ICode[_Num_ICode]->ChainId.Clear();
       _ICode[_Num_ICode]->AllocLen = 0;
       _ICode[_Num_ICode]->SeqLen = 0;
       _ICode[_Num_ICode]->chain_type = -1;
       _ICode[_Num_ICode]->res = NULL;
       _Num_ICode++;
}

void _FileObj::_ManageResSpace(SEQ *seq)
{
       int i = 0;
       if (seq->SeqLen >= seq->AllocLen) {
            seq->AllocLen += 100;
            RES **tmp = seq->res;
            seq->res = new RES*[seq->AllocLen];
            for (i = 0; i < seq->SeqLen; i++) seq->res[i] = tmp[i];
            for (i = seq->SeqLen; i < seq->AllocLen; i++) seq->res[i] = NULL;
            delete [] tmp;
       }
}

void _FileObj::get_temp_file()
{
       int fd;
       char templateFileName[9] = "./XXXXXX";

       if (strcmp(_temp_file, "")) return;
       strcpy(_temp_file, templateFileName);
       fd = mkstemp(_temp_file);
       close (fd);
}

void _FileObj::delete_Seqs()
{
       int i = 0, j = 0;
       if (_Seqs) {
            for (i = 0; i < _Num_Seqs; i++) {
                 if (_Seqs[i]) {
                      for (j = 0; j < _Seqs[i]->SeqLen; j++)
                           if (_Seqs[i]->res[j]) delete _Seqs[i]->res[j];
                      delete [] _Seqs[i]->res;
                      delete _Seqs[i];
                 }
            }
            delete [] _Seqs;
       }
       if (_NPdb) {
            for (i = 0; i < _Num_NPdb; i++) {
                 if (_NPdb[i]) {
                      for (j = 0; j < _NPdb[i]->SeqLen; j++)
                           if (_NPdb[i]->res[j]) delete _NPdb[i]->res[j];
                      delete [] _NPdb[i]->res;
                      delete _NPdb[i];
                 }
            }
            delete [] _NPdb;
       }
       if (_ICode) {
            for (i = 0; i < _Num_ICode; i++) {
                 if (_ICode[i]) {
                      for (j = 0; j < _ICode[i]->SeqLen; j++)
                           if (_ICode[i]->res[j]) delete _ICode[i]->res[j];
                      delete [] _ICode[i]->res;
                      delete _ICode[i];
                 }
            }
            delete [] _ICode;
       }
       _Num_Seqs = 0;
       _Alloc_Seqs = 0;
       _Num_NPdb = 0;
       _Alloc_NPdb = 0;
       _Num_ICode = 0;
       _Alloc_ICode = 0;
       _Seqs = NULL;
       _NPdb = NULL;
       _ICode = NULL;
}

void  _FileObj::set_input_filename(const char *string)  { strcpy(_input_filename, string); }
void  _FileObj::set_output_filename(const char *string) { strcpy(_output_filename, string); }
void  _FileObj::set_real_filename(const char *string)   { strcpy(_real_filename, string); }
void  _FileObj::set_StructureId(const char *string)     { strcpy(_StructureId, string); }
void  _FileObj::set_input_format(const int format)      { _input_format = format; }
void  _FileObj::set_output_format(const int format)     { _output_format = format; }
void  _FileObj::set_experiment_type(const int type)     { _experiment_type = type; }
void  _FileObj::set_file_status(const int status)       { _file_status = status; }
void  _FileObj::set_is_binary(const int binary)         { _is_binary = binary; }

void _FileObj::get_one_line_info(FILE *fp, int &TokenId, int &prevTokenId,
       char *PrevSeqNo, int &JrnlId,  const int PrevJrnlId, char *CardString, 
       char FieldInfo[MxField][MxLen], const int JrnlSerialNo,
       const int PrevJrnlSerialNo, const int input_format, const int is_new)
{
       int FieldNo = 1, i = 0, ret = 0, StrLen, length, newFlag;
       char FieldAnswer[MxStrLen], JrnlCardString[MxStrLen];
       char newFieldInfo[MxField][MxLen], symbol[4], charge[4];

       if (TokenId + 1 == JRNL_TOKEN) {
            FieldNo = 1;
            if (!get_jrnl_field(fp, 0, FieldNo, FieldAnswer)) return;
            FieldNo++;
            JrnlId = get_jrnl_token(fp, JrnlCardString);
     
            if (PrevJrnlId != -1 && PrevJrnlId != JrnlId ||
               (JrnlSerialNo == -1 && strcmp(FieldInfo[1], FieldAnswer) != 0
                || JrnlSerialNo != PrevJrnlSerialNo)) {
                 strcpy(FieldInfo[0], "JRNL");
                 put_card(TokenId, PrevJrnlId, FieldInfo);
                 strcpy(FieldInfo[0], CardString);
            }
            if (JrnlSerialNo == -1)
                 strcpy(FieldInfo[1], FieldAnswer);
            else if (JrnlSerialNo > 0)
                 sprintf(FieldInfo[1], "%d", JrnlSerialNo);
            else strcpy(FieldInfo[1], "");
        
            if (JrnlId == EOF) {
                 for (i = 0; i <= FieldNo; i++) FieldInfo[i][0] = '\0';
                 end_line(fp);
                 return;
            }
            strcpy(FieldInfo[FieldNo], JrnlCardString);

            for (FieldNo = 3; FieldNo < JrnlToken[JrnlId].NumField; FieldNo++) {
                 ret = get_jrnl_field(fp, JrnlId, FieldNo, FieldAnswer);
                 if (JrnlToken[JrnlId].ContinuedF == FieldNo + 1) {
                      StrLen = strlen(FieldInfo[FieldNo]);
                      if (StrLen != 0 && FieldInfo[FieldNo][StrLen-1] != '-' )
                           sprintf(&FieldInfo[FieldNo][StrLen], " %s", FieldAnswer);
                      else sprintf(&FieldInfo[FieldNo][StrLen], "%s", FieldAnswer);
                 } else if (!strcmp(FieldInfo[FieldNo], ""))
                      strcpy(FieldInfo[FieldNo], FieldAnswer);
                 if (ret == FALSE) {
                      StrLen = strlen (FieldInfo[FieldNo]);
                      for (; FieldNo < JrnlToken[JrnlId].NumField; FieldNo++)
                           FieldInfo[FieldNo][StrLen] = '\0';
                      break;
                 }
            }
       } else {
            for (FieldNo = 0; FieldNo < NdbToken[TokenId].NumField; FieldNo++)
                 strcpy(newFieldInfo[FieldNo], "");
            strcpy(newFieldInfo[0], CardString);
            for (FieldNo = 1; FieldNo < NdbToken[TokenId].NumField; FieldNo++) {
                 FieldAnswer[0] = '\0';
                 if (input_format == NDB_FILE_FORMAT_PDB &&
                     NdbToken[TokenId].FieldList[FieldNo].NdbOnly) {
                      if (!strcmp(CardString, "ATOM") ||
                          !strcmp(CardString, "HETATM")) {
                           memset(symbol, 0, 4);
                           memset(charge, 0, 4);
                           // get_atom_symbol_and_charge(fp, TokenId, FieldNo, symbol, charge);
                           strcpy(newFieldInfo[23], symbol);
                           strcpy(newFieldInfo[24], charge);
                      }
                      end_line(fp);
                      break;
                 } else ret = get_ndbfile_field(fp, TokenId, FieldNo, FieldAnswer, 
                        length);
                 strcpy(newFieldInfo[FieldNo], FieldAnswer);
                 if (ret == FALSE) break;
            }

            if (!is_new && prevTokenId != -1 && 
                TokenId != REMARK_TOKEN-1 &&  TokenId != PREMRK_TOKEN-1 &&
                TokenId != PCOMPN_TOKEN-1 &&  TokenId != PSOURC_TOKEN-1 &&
               (NdbToken[TokenId].ConField == 0 ||
                NdbToken[TokenId].PdbConField != 0 ||
                NdbToken[TokenId].SeqField == 0 ||
                strcmp(newFieldInfo[NdbToken[TokenId].SeqField-1], PrevSeqNo) != 0)) {
                 newFlag = TRUE;
                 if (NdbToken[TokenId].PdbConField != 0) {
                      newFlag = FALSE;
                      if (NdbToken[TokenId].SeqField != 0 &&
                         !strcmp(newFieldInfo[NdbToken[TokenId].SeqField-1], "")) {
                           StrLen = strlen(newFieldInfo[NdbToken[TokenId].ContinuedF-1]);
                           int has_comma = 0;
                           for (i = 0; i < StrLen; i++) {
                                if (!strncmp(&newFieldInfo[NdbToken[TokenId].ContinuedF
                                        - 1][i], ": ", 2)) break;
                                if (!strncmp(&newFieldInfo[NdbToken[TokenId].ContinuedF
                                        - 1][i], ":", 1)) has_comma = 1;
                                if (!strncmp(&newFieldInfo[NdbToken[TokenId].ContinuedF
                                        - 1][i], ";", 1) && has_comma) break;
                           }
                           if (i != StrLen) newFlag = TRUE;
                      } else if (NdbToken[TokenId].SeqField != 0 &&
                         (strcmp(newFieldInfo[NdbToken[TokenId].SeqField-1], PrevSeqNo) ||
                         !strcmp(newFieldInfo[NdbToken[TokenId].SeqField-1], "") &&
                         !strcmp(PrevSeqNo, ""))) newFlag = TRUE;
                 } else if (NdbToken[TokenId].ContinuedF != 0 &&
                            NdbToken[TokenId].MatrixType == 0) {
                      if (TokenId == HETNAM_TOKEN - 1 || TokenId == HETSYN_TOKEN - 1) {
                           if (!strcmp(newFieldInfo[2], FieldInfo[2])) newFlag = FALSE;
                      } else if (NdbToken[TokenId].SeqField != 0) {
                           if ((strcmp(newFieldInfo[NdbToken[TokenId].SeqField-1], "") ||
                                strcmp(FieldInfo[NdbToken[TokenId].SeqField-1], "")) &&
                               !strcmp(newFieldInfo[NdbToken[TokenId].SeqField-1],
                                   FieldInfo[NdbToken[TokenId].SeqField-1]))
                               newFlag = FALSE;
                      } else newFlag = FALSE;
                 }

                 if (newFlag) {
                      put_card(TokenId, 0, FieldInfo);
                      strcpy(FieldInfo[0], CardString);  
                 }
            }

            /* Concatenats continuation field */
            for (FieldNo = 0; FieldNo < NdbToken[TokenId].NumField; FieldNo++) {
                 if (NdbToken[TokenId].ContinuedF == FieldNo + 1) {
                      StrLen = strlen(FieldInfo[FieldNo]);
                      if (StrLen != 0 && FieldInfo[FieldNo][StrLen-1] != '-' )
                           sprintf(&FieldInfo[FieldNo][StrLen], " %s", 
                                             newFieldInfo[FieldNo]);
                      else sprintf(&FieldInfo[FieldNo][StrLen], "%s",
                                             newFieldInfo[FieldNo]);
                 } else if (!strcmp(FieldInfo[FieldNo], ""))
                      strcpy(FieldInfo[FieldNo], newFieldInfo[FieldNo]);
                 newFieldInfo[FieldNo][0] = '\0';
            }
       }
}

void _FileObj::put_card(const int CardId, const int JrnlId, char FieldInfo[MxField][MxLen])
{
       One_Card_Info_Format *pCard = NULL;
       int i = 0;
       if (is_atom_card(CardId) || CardId == MODEL_TOKEN-1) return;

       if (CardId +1 != JRNL_TOKEN && CardId != -1) {
            pCard = _CardInfo[CardId].add_a_new_card(NdbToken[CardId], 0);
            for (i = 0; i < NdbToken[CardId].NumField; i++) {
                 pCard->AddElementToField(FieldInfo[i], i);
                 memset(FieldInfo[i], 0, MxLen);
            }
       } else if (JrnlId != -1) {
            pCard = _JrnlCardInfo[JrnlId].add_a_new_card(JrnlToken[JrnlId], 1);
            for (i = 0; i < JrnlToken[JrnlId].NumField; i++) {
                 pCard->AddElementToField(FieldInfo[i], i);
                 memset(FieldInfo[i], 0, MxLen);
            }
       }
}

void _FileObj::get_sequence_from_seqres()
{
       if (_Seqs && _Num_Seqs) return;

       int i = 0;
       char prevChainId[5], name[5];
       strcpy(prevChainId, "XXX");
       One_Card_Info_Format *pCard = _CardInfo[SEQRES_TOKEN-1].Head();
       while (pCard) {
            if (strcmp(prevChainId, pCard->Field(2))) {
                 _ManageSeqSpace();
                 _Seqs[_Num_Seqs-1]->ChainId.Copy(pCard->Field(2));
                 strcpy(prevChainId, pCard->Field(2));
                 for (i = 4; i < NdbToken[SEQRES_TOKEN-1].NumField; i++) {
                      if (!strcmp(pCard->Field(i), "")) break;
                      if (!strcmp(pCard->Field(i), "?")) continue;
                      _ManageResSpace(_Seqs[_Num_Seqs-1]);
                      _Seqs[_Num_Seqs-1]->res[_Seqs[_Num_Seqs-1]->SeqLen] = new RES;
                      strcpy(name, pCard->Field(i));
                      find_obsolete_supercede_residue(name);
                      strcpy(_Seqs[_Num_Seqs-1]->res[_Seqs[_Num_Seqs-1]->SeqLen]->name, name);
                             // pCard->Field(i));
                      _Seqs[_Num_Seqs-1]->SeqLen++;
                 }
            } else {
                 for (i = 4; i < NdbToken[SEQRES_TOKEN-1].NumField; i++) {
                      if (!strcmp(pCard->Field(i), "")) break;
                      if (!strcmp(pCard->Field(i), "?")) continue;
                      _ManageResSpace(_Seqs[_Num_Seqs-1]);
                      _Seqs[_Num_Seqs-1]->res[_Seqs[_Num_Seqs-1]->SeqLen] = new RES;
                      strcpy(name, pCard->Field(i));
                      find_obsolete_supercede_residue(name);
                      strcpy(_Seqs[_Num_Seqs-1]->res[_Seqs[_Num_Seqs-1]->SeqLen]->name, name);
                             // pCard->Field(i));
                      _Seqs[_Num_Seqs-1]->SeqLen++;
                 }
            }
            pCard = pCard->Next();
       }
       get_sequence_types();
}

void _FileObj::get_sequence_types()
{
       int i = 0, type, j = 0, max_num = -1, index = -1, types[4], exist_unknown = 0;
       for (i = 0; i < _Num_Seqs; i++) {
            for (j = 0; j < 4; j++) types[j] = 0;
            for (j = 0; j < _Seqs[i]->SeqLen; j++) {
                 type = find_residue_type(_Seqs[i]->res[j]->name, exist_unknown);
                 if (type == ATOMN_TOKEN) types[0]++;
                 else if (type == ATOMP_TOKEN) types[1]++;
                 else if (type == ATOMS_TOKEN) types[2]++;
                 else if (type == HETAIN_TOKEN) types[3]++;
            }
            max_num = -1;
            index = -1;
            for (j = 0; j < 4; j++) {
                 if (types[j] > max_num) {
                      max_num = types[j];
                      index = j;
                 }
            }
            if (index >= 0) {
                 if (index == 0) _Seqs[i]->chain_type = ATOMN_TOKEN;
                 else if (index == 1) _Seqs[i]->chain_type = ATOMP_TOKEN;
                 else if (index == 2) _Seqs[i]->chain_type = ATOMS_TOKEN;
                 else if (index == 3) _Seqs[i]->chain_type = HETAIN_TOKEN;
            }
       }
}

void _FileObj::get_seqres_from_mapping(const int format)
{
       if (_num_mol == 0) return;

       int SequenceTokenids[5];
       SequenceTokenids[0] = SEQRES_TOKEN-1;
       SequenceTokenids[1] = PDBSQN_TOKEN-1;
       SequenceTokenids[2] = PDBSQR_TOKEN-1;
       SequenceTokenids[3] = ALTSQN_TOKEN-1;
       SequenceTokenids[4] = ALTSQR_TOKEN-1;
       for (int i = 0; i < 5; i++) {
            One_Card_Info_Format *pCard = _CardInfo[SequenceTokenids[i]].Head();
            while (pCard) pCard = _CardInfo[SequenceTokenids[i]].delete_a_card(pCard);
       }
       _CHAIN *chain = _molecules[0]->GetFirstChain();
       while (chain) {
            chain->get_seqres(format, SEQRES_TOKEN-1, &_CardInfo[SEQRES_TOKEN-1]);
            chain = _molecules[0]->GetNextChain();
       }

       int seqNo = 1;
       char chnid[5];
       strcpy(chnid, "XXX");
       One_Card_Info_Format *pCard = _CardInfo[SEQRES_TOKEN-1].Head();
       while (pCard) {
            if (strcmp(chnid, pCard->Field(2))) {
                 strcpy(chnid, pCard->Field(2));
                 seqNo = 1;
            } else seqNo++;
            pCard->AddElementToField(seqNo, 1);
            pCard = pCard->Next();
       }
}

void _FileObj::seq_alignment(const int predefined)
{
       int /* i = 0, */ j = 0, m = 0;
       SEQ *seq = NULL, *pdb = NULL, *icode = NULL;
       _CHAIN *chain = NULL;

       if (_num_mol == 0 || _molecules == NULL) return;

       for (m = 0; m < _num_mol; m++) {
            for (j = 0; j < _Num_Seqs; j++) _Seqs[j]->is_used = 0;
            chain = _molecules[m]->GetFirstChain();
            // i = 0;
            while (chain) {
                 // chain->set_order(++i);
                 chain->set_biol_id(1);
                 seq = NULL;
                 for (j = 0; j < _Num_Seqs; j++) {
                      if (_Seqs[j]->ChainId == chain->ChainID() &&
                           _Seqs[j]->chain_type == chain->chain_type() &&
                          !_Seqs[j]->is_used) {
                           _Seqs[j]->is_used = 1;
                           break;
                      }
                 }
                 if (j < _Num_Seqs) seq = _Seqs[j];
                 if (seq == NULL) chain->set_missing_sequence();

                 pdb = NULL;
                 if (predefined && _Num_NPdb && _NPdb) {
                      for (j = 0; j < _Num_NPdb; j++) {
                           if (_NPdb[j]->ChainId == chain->ChainID()) {
                                pdb = _NPdb[j];
                                break;
                           }
                      }
                 }
                 icode = NULL;
                 if (predefined && _Num_ICode && _ICode) {
                      for (j = 0; j < _Num_ICode; j++) {
                           if (_ICode[j]->ChainId == chain->ChainID()) {
                                icode = _ICode[j];
                                break;
                           }
                      }
                 }
                 chain->seq_alignment(seq, pdb, icode);
                 chain = _molecules[m]->GetNextChain();
            }
       }
}

void _FileObj::find_missing_residues()
{
       int add_missing_residue = 0, i = 0, m = 0;
       _CHAIN *chain = NULL;
       One_Card_Info_Format *newCard = NULL;

       if (_num_mol == 0 || _molecules == NULL) return;
       if (_CardInfo[MISRES_TOKEN-1].Head() == NULL) add_missing_residue = 1;
       if (!add_missing_residue) return;

       for (m = 0; m < _num_mol; m++) {
            chain = _molecules[m]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() == ATOMP_TOKEN ||
                     chain->chain_type() == ATOMN_TOKEN) {
                      for (i = 0; i < chain->SeqLen(); i++) {
                           if (chain->SeqRes(i)->Deleted) continue;
                           if (chain->SeqRes(i)->Residue) continue;
                           newCard = _CardInfo[MISRES_TOKEN-1]
                                   .add_a_new_card(NdbToken[MISRES_TOKEN-1], 0);
                           newCard->AddElementToField(i+1, 1);
                           if (_num_mol > 1)
                                newCard->AddElementToField(_molecules[m]->Mol_ID(), 2);
                           newCard->AddElementToField(chain->PDB_ChainID(), 3);
                           newCard->AddElementToField(chain->SeqRes(i)->Field[0], 4);
                           newCard->AddElementToField(chain->SeqRes(i)->Field[4], 5);
                           newCard->AddElementToField(chain->SeqRes(i)->InsCode,  6);
                      }
                 }
                 chain = _molecules[m]->GetNextChain();
            }
       }
}

void _FileObj::get_seqifo(const int validation_flag, const int together)
{
       FILE *fp = NULL;
       int first = 1, j = 0, k = 0, l = 0, m = 0, z_value, z_index, mismatch;
       char Asym_ID[100], buffer[200];
       double dist;
       _ATOM *atom1 = NULL, *atom2 = NULL;
       _CHAIN *chain = NULL, *chain1 = NULL;;
       CifString string, message(""), message1, message2, message3;
       ISTable *Table = NULL;
       _RESIDUE *residue = NULL;
       
       if (_num_mol == 0 || _molecules == NULL) return;

       _rep_mol = 0;
       if (_entity_molecule) delete _entity_molecule;
       _entity_molecule = new _MOLECULE;
       int changed_zero_occupancy = 0;
       int mol_id = -1;
       for (m = 0; m < _num_mol; m++) {
            if (_num_mol > 1) mol_id = _molecules[m]->Mol_ID();
            chain = _molecules[m]->GetFirstChain();
            while (chain) {
                 chain1 = _entity_molecule->GetFirstChain();
                 while (chain1) {
                      if (*chain == *chain1) break;
                      chain1 = _entity_molecule->GetNextChain();
                 }
                 if (!chain1) {
                      _entity_molecule->add_a_chain(chain);
                      _rep_mol = m;
                 }
                 if (_experiment_type == EXPERIMENT_TYPE_NMR) {
                      residue = chain->GetFirstResidue();
                      while (residue) {
                           changed_zero_occupancy += residue->changed_zero_occupancy();
                           residue = chain->GetNextResidue();
                      }
                 }
                 if (__need_fix_atom_name) {
                      int first_residue = 1;
                      for (j = 0; j < chain->SeqLen(); j++) {
                           residue = chain->SeqRes(j)->Residue;
                           if (!residue) continue;

                           residue->correction_atom_name(mol_id, _error_message_s);

                           if (first_residue &&
                              (chain->chain_type() == ATOMP_TOKEN ||
                               chain->chain_type() == ATOMN_TOKEN)) {
                                k = 0;
                                if (j == 0) k = 1;
                                if (chain->chain_type() == ATOMP_TOKEN)
                                     residue->fix_n_terminal_hydrogen(k);
                                else residue->fix_5_terminal_hydrogen();
                           }
                           first_residue = 0;
                      }
                 }
                 if (chain->chain_type() == ATOMP_TOKEN) {
                      residue = chain->GetLastResidue();
                      if (residue) {
                           atom1 = residue->find_atom("C");
                           atom2 = residue->find_atom("OXT");
                           if (atom1 && atom2) {
                                dist = cal_distance(atom1, atom2);
                                if (dist > 1.6) {
                                     message += "WARNING: Distance of C-OXT bond for residue ";
                                     if (_num_mol > 1) 
                                          sprintf(buffer, "(%s %d %s (Model %d)) is %.2f\n\n",
                                               residue->ResName(), residue->pdb_res_no(),
                                               chain->PDB_ChainID(), _molecules[m]->Mol_ID(), dist);
                                     else sprintf(buffer, "(%s %d %s) is %.2f\n\n",
                                               residue->ResName(), residue->pdb_res_no(),
                                               chain->PDB_ChainID(), dist);
                                     message += buffer;
                                }
                           }
                      }
                 }
                 chain = _molecules[m]->GetNextChain();
            }
       }
       _entity_molecule->get_entity_id();

       if (changed_zero_occupancy) {
            message += "WARNING: The occupancies were zero and have been changed ";
            message += "to 1.0 according to PDB policy.\n\n";
       }

       mismatch = 0;
       for (m = 0; m < _num_mol; m++) {
            chain = _molecules[m]->GetFirstChain();
            message1.Copy("");
            message3.Copy("");
            while (chain) {
                 message2.Copy("");
                 mismatch += chain->print_mismatch(message2, first, _Error_Check[4], _Error_Check[19],
                               _Error_Check_Message[19]);
                 message1 += message2;
                 if ((m == 0 || message2 != "") && validation_flag != 2) {
                      chain->print_alignment(message3, 15);
                 }
                 chain = _molecules[m]->GetNextChain();
            }
            message1 += message3;
            if (message1 != "") {
                 if (_num_mol > 1) {
                      sprintf(buffer, "\nModel %d:\n\n", _molecules[m]->Mol_ID());
                      message += buffer;
                 }
                 message += message1;
            }
       }
       if (mismatch) {
            message += "\nThere were mismatches between the deposited sequence and sequence\n";
            message += "in coordinates involving alanines and/or glycines. The residue\n";
            message += "names have been changed to match the deposited sequence and/or\n";
            message += "the sequence database unless engineered mutation were specified\n";
            message += "in the deposition. The missing side chain atoms are listed in\n";
            message += "remark 470.\n\n";
       }

       for (j = 0; j < _n_stbiol; j++) _stbiol[j]->internal_order = j + 1;
       if (_n_stbiol) {
            if (_chain_corr) delete [] _chain_corr;
            _chain_corr = new _CHAIN_CORR[_n_stbiol];
       }
       _n_chain_corr = 0;
       for (m = 0; m < _num_mol; m++) {
            for (j = 0; j < _n_stbiol; j++) _stbiol[j]->is_change = 0;
            chain = _molecules[m]->GetFirstChain();
            while (chain) {
                 for (k = 0; k < _n_stbiol; k++) {
                      if (_stbiol[k]->Asym_ID == chain->ChainID() &&
                          !strcmp(_stbiol[k]->Symmetry, "1_555") &&
                          !_stbiol[k]->is_change) break;
                 }
                 if (k < _n_stbiol) {
                      chain->set_internal_order(k + 1);
                      chain->set_order(_stbiol[k]->order);
                      chain->set_biol_id(_stbiol[k]->Biol_ID);
                      chain->set_ChainID(_stbiol[k]->NDB_Strand_ID.Text());
                      if (_exist_pdb_chain_id)
                           chain->set_PDB_ChainID(_stbiol[k]->PDB_Strand_ID);
                      chain->set_PDB_ChainID_Flag(_stbiol[k]->PDB_Strand_ID_Flag);
                      if (m == _num_mol - 1) {
                           strcpy(_chain_corr[_n_chain_corr].old_chn, _stbiol[k]->Asym_ID.Text());
                           strcpy(_chain_corr[_n_chain_corr].new_chn, _stbiol[k]->NDB_Strand_ID.Text());
                           _n_chain_corr++;
                           strcpy(Asym_ID, _stbiol[k]->Asym_ID.Text());
                           for (l = 0; l < _n_stbiol; l++) {
                                if (_stbiol[l]->Asym_ID == Asym_ID &&
                                    !_stbiol[l]->is_change) {
                                     _stbiol[l]->Asym_ID.Copy(_stbiol[k]->NDB_Strand_ID);
                                     _stbiol[l]->is_change = 1;
                                }
                           }
                      }
                      _stbiol[k]->is_change = 1;
                 }
                 chain = _molecules[m]->GetNextChain();
            }
            _molecules[m]->get_entity_id1(_entity_molecule);
       }
       if (!_cell->is_artifical) {
            z_value = _molecules[_rep_mol]->MaxNumEntity() * _cell->nop;
            if (z_value) {
                 if (_cell->z && z_value != _cell->z) {
                      sprintf(buffer, "\n%s: Error: Z value is %d. It should be %d.\n\n",
                                 _input_filename, _cell->z, z_value);
                      _error_message += buffer;
                 }
                 if (!_cell->z) {
                      _cell->z = z_value;
                      if (_CardInfo[CRYST1_TOKEN-1].Head())
                           _CardInfo[CRYST1_TOKEN-1].Head()->AddElementToField(z_value, 8);
                      else if (_CifObj) {
                           Table = _CifObj->GetTablePtr(_StructureId, "cell");
                           if (Table) {
                                z_index = Table->GetColumnIndex("Z_PDB");
                                if (z_index >= 0) {
                                     sprintf(buffer, "%d", z_value);
                                     string.Copy(buffer);
                                     Table->UpdateCell(string, z_index, 0);
                                }
                           }
                           _CifObj->WriteTable(Table, _StructureId, "cell");
                      }
                 }
            }
       }
       if (validation_flag == 1) {
            fp = fopen("validation.alignment", "w");
            fprintf(fp, "%s\n", message.Text());
            fclose (fp);
       } else if (message != "") {
            if (!together) ndb_log_message(NDB_MSG_WARN, "%s\n", message.Text());
       }
}

void _FileObj::get_cell(void)
{
       int i = 0, j = 0, exist_scale_matrix = 0, numBlock = 0, is_change = 0, name_index = -1;
       int full_name = -1, a_index = -1, b_index = -1, c_index = -1, alpha_index = -1;
       int beta_index = -1, gamma_index = -1, z_index = 1, index_11 = -1, index_12 = -1;
       int index_13 = -1, index_21 = -1, index_22 = -1, index_23 = -1, index_31 = -1;
       int index_32 = -1, index_33 = -1, index_v1 = -1, index_v2 = -1, index_v3 = -1;
       double scale[3][3], vec[3];
       char new_name[50], **blockName = NULL;
       CifString cifstring, a, b, c, alpha, beta, gamma, space_group;
       ISTable *Table = NULL;
       One_Card_Info_Format *pCard = NULL;

       if (_cell) return;

       a.Clear();
       b.Clear();
       c.Clear();
       alpha.Clear();
       beta.Clear();
       gamma.Clear();
       space_group.Clear();
       _cell = new NDBCELLCONST;
       _cell->a = _cell->b = _cell->c = _cell->alpha = _cell->beta = _cell->gamma = 0.0;
       _cell->space_group[0] = '\0';
       _cell->full_space_group[0] = '\0';
       _cell->z = 0;
       _cell->is_artifical = 0;

       if (_CifObj) {
            blockName = _CifObj->GetBlockNames(numBlock);
            if (_CifObj->IsTablePresent(blockName[0], "symmetry")) {
                 Table = _CifObj->GetTablePtr(blockName[0], "symmetry");
                 name_index = Table->GetColumnIndex("space_group_name_H-M");
                 full_name = Table->GetColumnIndex("ndb_full_space_group_name_H-M");
                 cifstring.Clear();
                 Table->GetCell(cifstring, name_index, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?") {
                      check_space_group(cifstring.Text(), new_name);
                      if (strcmp(cifstring.Text(), new_name)) {
                           cifstring.Copy(new_name);
                           Table->UpdateCell(cifstring, name_index, 0);
                           is_change  = 1;
                      }
                      strcpy(_cell->space_group, new_name);
                 }
                 cifstring.Copy("");
                 Table->GetCell(cifstring, full_name, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?") {
                      check_space_group(cifstring.Text(), new_name);
                      if (strcmp(cifstring.Text(), new_name))
                           cifstring.Copy(new_name);
                      Table->UpdateCell(cifstring, name_index, 0);
                      cifstring.Copy("");
                      Table->UpdateCell(cifstring, full_name, 0);
                      strcpy(_cell->space_group, new_name);
                      is_change = 1;
                 }
                 space_group.Copy(_cell->space_group);
                 if (!strcmp(_cell->space_group, "C 2")) {
                      strcpy(_cell->space_group, "C 1 2 1");
                      cifstring.Copy("C 1 2 1");
                      Table->UpdateCell(cifstring, name_index, 0);
                      is_change = 1;
                 } else if (!strcmp(_cell->space_group, "I 2")) {
                      strcpy(_cell->space_group, "I 1 2 1");
                      cifstring.Copy("I 1 2 1");
                      Table->UpdateCell(cifstring, name_index, 0);
                      is_change = 1;
                 } else if (!strcmp(_cell->space_group, "P 2")) {
                      strcpy(_cell->space_group, "P 1 2 1");
                      cifstring.Copy("P 1 2 1");
                      Table->UpdateCell(cifstring, name_index, 0);
                      is_change = 1;
                 } else if (!strcmp(_cell->space_group, "P 21")) {
                      strcpy(_cell->space_group, "P 1 21 1");
                      cifstring.Copy("P 1 21 1");
                      Table->UpdateCell(cifstring, name_index, 0);
                      is_change = 1;
                 }
                 if (is_change) _CifObj->WriteTable(Table, blockName[0], "symmetry");
            }
            if (_CifObj->IsTablePresent(blockName[0], "cell")) {
                 Table = _CifObj->GetTablePtr(blockName[0], "cell");
                 a_index = Table->GetColumnIndex("length_a");
                 b_index = Table->GetColumnIndex("length_b");
                 c_index = Table->GetColumnIndex("length_c");
                 alpha_index = Table->GetColumnIndex("angle_alpha");
                 beta_index = Table->GetColumnIndex("angle_beta");
                 gamma_index = Table->GetColumnIndex("angle_gamma");
                 z_index = Table->GetColumnIndex("Z_PDB");
                 cifstring.Clear();
                 Table->GetCell(cifstring, a_index, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?") {
                      _cell->a = atof(cifstring.Text());
                      a.Copy(cifstring);
                 }
                 cifstring.Copy("");
                 Table->GetCell(cifstring, b_index, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?") {
                      _cell->b = atof(cifstring.Text());
                      b.Copy(cifstring);
                 }
                 cifstring.Copy("");
                 Table->GetCell(cifstring, c_index, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?") {
                      _cell->c = atof(cifstring.Text());
                      c.Copy(cifstring);
                 }
                 cifstring.Copy("");
                 Table->GetCell(cifstring, alpha_index, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?") {
                      _cell->alpha = atof(cifstring.Text());
                      alpha.Copy(cifstring);
                 }
                 cifstring.Copy("");
                 Table->GetCell(cifstring, beta_index, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?") {
                      _cell->beta = atof(cifstring.Text());
                      beta.Copy(cifstring);
                 }
                 cifstring.Copy("");
                 Table->GetCell(cifstring, gamma_index, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?") {
                      _cell->gamma = atof(cifstring.Text());
                      gamma.Copy(cifstring);
                 }
                 cifstring.Copy("");
                 Table->GetCell(cifstring, z_index, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      _cell->z = atoi(cifstring.Text());
            }
            if (_CifObj->IsTablePresent(blockName[0], "atom_sites")) {
                 Table = _CifObj->GetTablePtr(blockName[0], "atom_sites");
                 index_11 = Table->GetColumnIndex("fract_transf_matrix[1][1]");
                 index_12 = Table->GetColumnIndex("fract_transf_matrix[1][2]");
                 index_13 = Table->GetColumnIndex("fract_transf_matrix[1][3]");
                 index_21 = Table->GetColumnIndex("fract_transf_matrix[2][1]");
                 index_22 = Table->GetColumnIndex("fract_transf_matrix[2][2]");
                 index_23 = Table->GetColumnIndex("fract_transf_matrix[2][3]");
                 index_31 = Table->GetColumnIndex("fract_transf_matrix[3][1]");
                 index_32 = Table->GetColumnIndex("fract_transf_matrix[3][2]");
                 index_33 = Table->GetColumnIndex("fract_transf_matrix[3][3]");
                 index_v1 = Table->GetColumnIndex("fract_transf_vector[1]");
                 index_v2 = Table->GetColumnIndex("fract_transf_vector[2]");
                 index_v3 = Table->GetColumnIndex("fract_transf_vector[3]");
                 exist_scale_matrix = 1;
                 cifstring.Clear();
                 Table->GetCell(cifstring, index_11, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      scale[0][0] = atof(cifstring.Text());
                 else exist_scale_matrix = 0;

                 cifstring.Copy("");
                 Table->GetCell(cifstring, index_12, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      scale[0][1] = atof(cifstring.Text());
                 else exist_scale_matrix = 0;

                 cifstring.Copy("");
                 Table->GetCell(cifstring, index_13, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      scale[0][2] = atof(cifstring.Text());
                 else exist_scale_matrix = 0;

                 cifstring.Copy("");
                 Table->GetCell(cifstring, index_21, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      scale[1][0] = atof(cifstring.Text());
                 else exist_scale_matrix = 0;

                 cifstring.Copy("");
                 Table->GetCell(cifstring, index_22, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      scale[1][1] = atof(cifstring.Text());
                 else exist_scale_matrix = 0;

                 cifstring.Copy("");
                 Table->GetCell(cifstring, index_23, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      scale[1][2] = atof(cifstring.Text());
                 else exist_scale_matrix = 0;

                 cifstring.Copy("");
                 Table->GetCell(cifstring, index_31, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      scale[2][0] = atof(cifstring.Text());
                 else exist_scale_matrix = 0;

                 cifstring.Copy("");
                 Table->GetCell(cifstring, index_32, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      scale[2][1] = atof(cifstring.Text());
                 else exist_scale_matrix = 0;

                 cifstring.Copy("");
                 Table->GetCell(cifstring, index_33, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      scale[2][2] = atof(cifstring.Text());
                 else exist_scale_matrix = 0;

                 cifstring.Copy("");
                 Table->GetCell(cifstring, index_v1, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      vec[0] = atof(cifstring.Text());
                 else exist_scale_matrix = 0;

                 cifstring.Copy("");
                 Table->GetCell(cifstring, index_v2, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      vec[1] = atof(cifstring.Text());
                 else exist_scale_matrix = 0;

                 cifstring.Copy("");
                 Table->GetCell(cifstring, index_v3, 0);
                 if (cifstring != "" && cifstring != "." && cifstring != "?")
                      vec[2] = atof(cifstring.Text());
                 else exist_scale_matrix = 0;
            }
            strcpy(new_name, blockName[0]);
            for (i = 0; i < numBlock; i++) delete [] blockName[i];
            delete [] blockName;
       }
       if (_CardInfo[CRYST1_TOKEN-1].Head()) {
            _cell->a     = atof(_CardInfo[CRYST1_TOKEN-1].Head()->Field(1));
            _cell->b     = atof(_CardInfo[CRYST1_TOKEN-1].Head()->Field(2));
            _cell->c     = atof(_CardInfo[CRYST1_TOKEN-1].Head()->Field(3));
            _cell->alpha = atof(_CardInfo[CRYST1_TOKEN-1].Head()->Field(4));
            _cell->beta  = atof(_CardInfo[CRYST1_TOKEN-1].Head()->Field(5));
            _cell->gamma = atof(_CardInfo[CRYST1_TOKEN-1].Head()->Field(6));
            check_space_group(_CardInfo[CRYST1_TOKEN-1].Head()->Field(7), new_name);
            strcpy(_cell->space_group, new_name);
            _cell->z     = atoi(_CardInfo[CRYST1_TOKEN-1].Head()->Field(8));
            _cell->unique_axis = _CardInfo[CRYST1_TOKEN-1].Head()->Field(9)[0];
            strcpy(_cell->full_space_group, _CardInfo[CRYST1_TOKEN-1].Head()->Field(10));
            if (strcmp(_cell->full_space_group, "")) {
                 strcpy(_cell->space_group, _cell->full_space_group);
                 strcpy(_cell->full_space_group, "");
                 _CardInfo[CRYST1_TOKEN-1].Head()->AddElementToField("", 10);
            }
            a.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(1));
            b.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(2));
            c.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(3));
            alpha.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(4));
            beta.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(5));
            gamma.Copy(_CardInfo[CRYST1_TOKEN-1].Head()->Field(6));
            space_group.Copy(_cell->space_group);
       }
       if ((_cell->a == 1.0 || _cell->b == 1.0 || _cell->c == 1.0) &&
           _cell->alpha == 90.0 && _cell->beta == 90.0 && _cell->gamma == 90.0 &&
           !strcmp(_cell->space_group, "P 1")) {
            _cell->a = 1000.0;
            _cell->b = 1000.0;
            _cell->c = 1000.0;
            _cell->is_artifical = 1;
       } else if (_cell->a == 0.0 || _cell->b == 0.0 || _cell->c == 0.0 ||
                  _cell->alpha == 0.0 || _cell->beta == 0.0 || _cell->gamma == 0.0 ||
                  !strcmp(_cell->space_group, "")) {
            _cell->a = 1000.0;
            _cell->b = 1000.0;
            _cell->c = 1000.0;
            _cell->alpha = 90.0;
            _cell->beta = 90.0;
            _cell->gamma = 90.0;
            strcpy(_cell->space_group, "P 1");
            _cell->is_artifical = 1;
       }

       if (_cell->a == 1.0 || _cell->b == 1.0 || _cell->c == 1.0 ||
           _cell->a == 0.0 || _cell->b == 0.0 || _cell->c == 0.0) {
            ndb_log_message(NDB_MSG_ERR, "Wrong crystal info: a=%s b=%s c=%s alpha=%s beta=%s gamma=%s space group=%s",
              a.Text(), b.Text(), c.Text(), alpha.Text(), beta.Text(), gamma.Text(), space_group.Text());
            _Error_Check[52]++;
            _Error_Check_Message[52] += "Crystal info. error: a=";
            _Error_Check_Message[52] += a;
            _Error_Check_Message[52] += " b=";
            _Error_Check_Message[52] += b;
            _Error_Check_Message[52] += " c=";
            _Error_Check_Message[52] += c;
            _Error_Check_Message[52] += " alpha=";
            _Error_Check_Message[52] += alpha;
            _Error_Check_Message[52] += " beta=";
            _Error_Check_Message[52] += beta;
            _Error_Check_Message[52] += " gamma=";
            _Error_Check_Message[52] += gamma;
            _Error_Check_Message[52] += " space group=";
            _Error_Check_Message[52] += space_group;
            _cell->is_artifical = 2;
       } else if ((_experiment_type & EXPERIMENT_TYPE_XRAY) &&
                  (a == "" || b == "" || c == "" || alpha == "" || beta == "" || gamma == "" ||
                  !strncmp(a.Text(), "1.0", 3) || !strncmp(b.Text(), "1.0", 3) || !strncmp(c.Text(), "1.0", 3))) {
            ndb_log_message(NDB_MSG_ERR, "Wrong crystal info: a=%s b=%s c=%s alpha=%s beta=%s gamma=%s space group=%s",
              a.Text(), b.Text(), c.Text(), alpha.Text(), beta.Text(), gamma.Text(), space_group.Text());
            _Error_Check[52]++;
            _Error_Check_Message[52] += "Crystal info. error: a=";
            _Error_Check_Message[52] += a;
            _Error_Check_Message[52] += " b=";
            _Error_Check_Message[52] += b;
            _Error_Check_Message[52] += " c=";
            _Error_Check_Message[52] += c;
            _Error_Check_Message[52] += " alpha=";
            _Error_Check_Message[52] += alpha;
            _Error_Check_Message[52] += " beta=";
            _Error_Check_Message[52] += beta;
            _Error_Check_Message[52] += " gamma=";
            _Error_Check_Message[52] += gamma;
            _Error_Check_Message[52] += " space group=";
            _Error_Check_Message[52] += space_group;
       }

       if (ndb_get_sym_data(_cell)) {
            _cell->is_artifical = 2;
            ndb_log_message(NDB_MSG_ERR, "Wrong space group name '%s'\n", _cell->space_group);
       }
       ndb_ortho_to_frac_matrix(_cell);
       ndb_frac_to_ortho_matrix(_cell);

       if (_CardInfo[SCALE_TOKEN-1].Head() && !exist_scale_matrix) {
            pCard = _CardInfo[SCALE_TOKEN-1].Head();
            for (i = 0; i < 3; i++) {
                 if (pCard == NULL) break;
                 for (j = 0; j < 3; j++)
                      scale[i][j] = atof(pCard->Field(j+1));
                 vec[i] = atof(pCard->Field(j+1));
                 pCard = pCard->Next();
            }
            if (i == 3) exist_scale_matrix = 1;
       }

       if (!_cell->is_artifical && exist_scale_matrix) {
            if (check_scale_token(_cell, scale, vec)) {
                 if (_CifObj) _CifObj->DeleteTable(new_name, "atom_sites");
                 pCard = _CardInfo[SCALE_TOKEN-1].Head();
                 while (pCard) pCard = _CardInfo[SCALE_TOKEN-1].delete_a_card(pCard);
            }
       }
}

void _FileObj::delete_null_cards()
{
       int i = 0, j = 0, card_id;
       One_Card_Info_Format *pCard = NULL;

       card_id = find_card_id("PDBRMK");
       for (i = 0; i < MxTokens; i++) {
            if (i == REMARK_TOKEN-1 || i == PREMRK_TOKEN-1 ||
                i == PCOMPN_TOKEN-1 || i == PSOURC_TOKEN-1 ||
                i == NCSGRO_TOKEN-1 || i == card_id) continue;

            pCard = _CardInfo[i].Head();
            while (pCard) {
                 for (j = 1; j < NdbToken[i].NumField; j++) {
                      if (j != NdbToken[i].SeqField-1 && j != NdbToken[i].ConField-1 &&
                          strcmp(pCard->Field(j), "") && (j != 2 || (i != FORMUL_TOKEN-1 &&
                          i != HETSYN_TOKEN-1 && i != HETNAM_TOKEN-1))) break;
                 }
                 if (j == NdbToken[i].NumField)
                      pCard = _CardInfo[i].delete_a_card(pCard);
                 else pCard = pCard->Next();
            }
       }
       for (i = 0; i < MxJrnlTokens; i++) {
            pCard = _JrnlCardInfo[i].Head();
            while (pCard) {
                 for (j = 4; j < JrnlToken[i].NumField; j++) {
                      if (strcmp(pCard->Field(j), "")) break;
                 }
                 if (j == JrnlToken[i].NumField)
                      pCard = _JrnlCardInfo[i].delete_a_card(pCard);
                 else pCard = pCard->Next();
            }
       }
}

void _FileObj::convert_between_orthogonal_and_fractional(const int iflag)
{
       if (_num_mol == 0 || _molecules == NULL) return;
       _molecules[0]->convert_between_orthogonal_and_fractional(_cell, iflag);
}

void _FileObj::print_master_card(FILE *fp, const int numRemark)
{
       int pos = 6, countNo, field_no;
       char FieldFormat[82], temp_line[82];
       int card_id = MASTER_TOKEN-1;
       int FTNOTE_TOKEN = find_card_id("FTNOTE");
       int CONECT_TOKEN = find_card_id("CONECT");

       sprintf(temp_line, "%-6s", NdbToken[card_id].TokenName);
       for (field_no = 1; field_no < NdbToken[card_id].NumField; field_no++) {
            print_space_between_field(card_id, field_no, temp_line, pos, 0);
            get_field_format(card_id, field_no, FieldFormat, 0, NDB_FILE_FORMAT_PDB);
            countNo = 0;
            if (field_no == 1) countNo = numRemark;
            else if (field_no == 2)
                 countNo = ndb_count_general_line_no(FTNOTE_TOKEN);
            else if (field_no == 3)
                 countNo = ndb_count_general_line_no(HET_TOKEN-1);
            else if (field_no == 4)
                 countNo = ndb_count_general_line_no(HELIX_TOKEN-1);
            else if (field_no == 5)
                 countNo = ndb_count_general_line_no(SHEET_TOKEN-1);
            else if (field_no == 6)
                 countNo = ndb_count_general_line_no(TURN_TOKEN-1);
            else if (field_no == 7)
                 countNo = ndb_count_general_line_no(SITE_TOKEN-1);
            else if (field_no == 8) {
                 countNo = ndb_count_general_line_no(ORIGX_TOKEN-1);
                 countNo += ndb_count_general_line_no(SCALE_TOKEN-1);
                 countNo += ndb_count_general_line_no(MTRIX_TOKEN-1);
            } else if (field_no == 9) 
                 countNo = ndb_count_atom_line_no();
            else if (field_no == 10)
                 countNo = ndb_count_num_chain();
            else if (field_no == 11)
                 countNo =  ndb_count_general_line_no(CONECT_TOKEN);
            else if (field_no == 12)
                 countNo =  ndb_count_general_line_no(SEQRES_TOKEN-1);
            sprintf(&temp_line[pos], FieldFormat, (double) countNo);
            pos += NdbToken[card_id].FieldList[field_no].FieldWidth;
       }
       fprintf(fp,"%s          \n", temp_line);
}

int _FileObj::ndb_count_general_line_no(const int card_id)
{
       int countNo = 0;
       One_Card_Info_Format *pCard = _CardInfo[card_id].Head();
       while (pCard) {
            countNo++;
            pCard = pCard->Next();
       }
       return countNo;
}

int _FileObj::ndb_count_atom_line_no()
{
       int countNo = 0, i = 0;
       for (i = 0; i < _num_mol; i++) {
            countNo += _molecules[i]->count_atom_no(-1);
       }
       return countNo;
}

int _FileObj::ndb_count_num_chain()
{
       int countNo = 0, i = 0;
       _CHAIN *chain = NULL;
       for (i = 0; i < _num_mol; i++) {
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() == ATOMP_TOKEN ||
                     chain->chain_type() == ATOMN_TOKEN)
                      countNo++;
                 chain = _molecules[i]->GetNextChain();
            }
       }
       return countNo;
}

void _FileObj::add_chain_and_solvent()
{
       int found_chain_atoms = 0, found_solvent_atoms = 0, card_id;
       One_Card_Info_Format *pCard = NULL;

       if (_num_mol == 0 || _molecules == NULL) return;
       _MOLECULE *mol = _molecules[0];
       _CHAIN *chain = mol->GetFirstChain();
       while (chain) {
            if (chain->chain_type() == HETAS_TOKEN)
                 found_solvent_atoms = 1;
            if (chain->chain_type() == ATOMP_TOKEN ||
                chain->chain_type() == ATOMN_TOKEN)
                 found_chain_atoms = 1;
            chain = mol->GetNextChain();
       }
       if (found_chain_atoms) {
            card_id = find_card_id("COORDA");
            pCard = _CardInfo[card_id].Head();
            if (!pCard) pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
            pCard->AddElementToField("Y", 1);
       }
       if (found_solvent_atoms) {
            card_id = find_card_id("COORDS");
            pCard = _CardInfo[card_id].Head();
            if (!pCard) pCard = _CardInfo[card_id].add_a_new_card(NdbToken[card_id], 0);
            pCard->AddElementToField("Y", 1);
       }
}

void _FileObj::merge_waters()
{
       for (int i = 0; i < _num_mol; i++) _molecules[i]->merge_waters(1);
}

int _FileObj::exist_cryst_info()
{
       if (_cell == NULL) return 0;
       if (_cell->is_artifical) return 0;
       return 1;
}

void _FileObj::get_assembly_info()
{
       if (_experiment_type == EXPERIMENT_TYPE_NMR) return;

       if (!_CifObj) return;

       char *blockId = _CifObj->GetBlockName(0);
       ISTable *Table = _CifObj->GetTablePtr(blockId, "pdbx_struct_assembly");
       if (!is_empty_table(Table)) {
            delete [] blockId;
            return;
       }

       Table = _CifObj->GetTablePtr(blockId, "struct_biol_gen");
       ISTable *Table1 = _CifObj->GetTablePtr(blockId, "struct_asym");
       if (is_empty_table(Table) || is_empty_table(Table1)) {
            delete [] blockId;
            return;
       }

       int rowNo = Table->GetNumRows();
       int idx1 = Table->GetColumnIndex("biol_id");
       int idx2 = Table->GetColumnIndex("asym_id");
       int idx3 = Table->GetColumnIndex("symmetry");
       int idx6 = Table->GetColumnIndex("ndb_new_pdb_asym_id");

       int rowNo1 = Table1->GetNumRows();
       int idx4 = Table1->GetColumnIndex("id");
       int idx5 = Table1->GetColumnIndex("ndb_type");

       int num_assembly = 0;
       CifString cs;
       for (int i = 0; i < rowNo; i++) {
            get_value(cs, Table, idx1, i);
            if (atoi(cs.Text()) > num_assembly) {
                 num_assembly = atoi(cs.Text());
            }
       }

       if (num_assembly > 0) {
            char assembly_id[10];
            CifString chainid, symmetry, cs1, cs2;
            _ASSEMBLY_ARRAY assembly;
            for (int i = 0; i < num_assembly; i++) {
                 sprintf(assembly_id, "%d", i + 1);
                 assembly.assembly_id.Copy(assembly_id);
                 assembly.method.Copy("author_defined_assembly");
                 assembly.software.Clear();
                 assembly.absa.Clear();
                 assembly.ssa.Clear();
                 assembly.more.Clear();
                 assembly.assembly_chains.Clear();
                 for (int j = 0; j < rowNo; j++) {
                      get_value(cs, Table, idx1, j);
                      if (cs == assembly.assembly_id) {
                           get_value(chainid, Table, idx2, j);
                           get_value(symmetry, Table, idx3, j);
                           int is_polymer = 0;
                           for (int k = 0; k < rowNo1; k++) {
                                get_value(cs1, Table1, idx4, k);
                                get_value(cs2, Table1, idx5, k);
                                if (cs1 == chainid &&
                                   (cs2 == "ATOMP" ||
                                    cs2 == "ATOMN")) {
                                     is_polymer = 1;
                                     break;
                                }
                           }
                           if (!is_polymer) continue;
                           int found = 0;
                           for (unsigned int l = 0; l < assembly.assembly_chains.Length(); l++) {
                                if (assembly.assembly_chains[l].PDB_chain_ID == chainid) {
                                     assembly.assembly_chains[l].sym_operators += ", ";
                                     assembly.assembly_chains[l].sym_operators += symmetry;
                                     found = 1;
                                     break;
                                }
                           }
                           if (!found) {
                                _ASSEMBLY_CHAIN assembly_chain;
                                assembly_chain.PDB_chain_ID.Copy(chainid);
                                assembly_chain.sym_operators.Copy(symmetry);
                                assembly.assembly_chains.Add(assembly_chain);
                           }
                      }
                 }

                 for (unsigned int k = 0; k < assembly.assembly_chains.Length(); k++) {
                      for (int j = 0; j < rowNo; j++) {
                           get_value(chainid, Table, idx2, j);
                           get_value(symmetry, Table, idx3, j);
                           get_value(cs, Table, idx6, j);
                           if (assembly.assembly_chains[k].PDB_chain_ID == chainid &&
                               symmetry == "1_555") {
                                assembly.assembly_chains[k].PDB_chain_ID.Copy(cs);
                                break;
                           }
                      }
                 }

                 if (assembly.assembly_chains.Length() > 0) __assembly_array.Add(assembly);
            }
       }

       _CifObj->DeleteTable(blockId, "struct_biol_gen");
       delete [] blockId;
}

void _FileObj::get_assembly_info_from_assembly_category()
{
       if (!_CifObj) return;

       char *blockId = _CifObj->GetBlockName(0);

       ISTable *t = _CifObj->GetTablePtr(blockId, "pdbx_struct_assembly");
       if (is_empty_table(t)) {
            delete [] blockId;
            return;
       }

       int is_special_symmetry = 0;
       t = _CifObj->GetTablePtr(blockId, "pdbx_helical_symmetry");
       if (!is_empty_table(t)) is_special_symmetry = 1;
       t = _CifObj->GetTablePtr(blockId, "pdbx_point_symmetry");
       if (!is_empty_table(t)) is_special_symmetry = 1;

       if (is_special_symmetry) {
            delete [] blockId;
            return;
       }

       t = _CifObj->GetTablePtr(blockId, "pdbx_struct_assembly");

       ReVarCifArray<_PDB_ASYM_CORR> chainids;
       get_pdb_asym_correspondence(_CifObj, blockId, chainids);

       __assembly_array.Clear();

       int rowNo = t->GetNumRows();
       int idx1 = t->GetColumnIndex("id");
       int idx2 = t->GetColumnIndex("details");
       int idx3 = t->GetColumnIndex("method_details");
       _ASSEMBLY_ARRAY assembly;
       CifString cs1, cs2, cs3;
       for (int i = 0; i < rowNo; i++) {
            assembly.assembly_id.Clear();
            assembly.method.Clear();
            assembly.software.Clear();
            assembly.absa.Clear();
            assembly.ssa.Clear();
            assembly.more.Clear();
            assembly.assembly_chains.Clear();
            cs1.Clear();
            t->GetCell(cs1, idx1, i);
            if (cs1 == "?" || cs1 == ".") cs1.Clear();
            assembly.assembly_id.Copy(cs1);
            cs2.Clear();
            t->GetCell(cs2, idx2, i);
            if (cs2 == "?" || cs2 == ".") cs2.Clear();
            assembly.method.Copy(cs2);
            cs3.Clear();
            t->GetCell(cs3, idx3, i);
            if (cs3 == "?" || cs3 == ".") cs3.Clear();
            if ((assembly.method == "author_and_software_defined_assembly" ||
                 assembly.method == "software_defined_assembly") && cs3 == "") {
                 cs3.Copy("PISA");
            }
            assembly.software.Copy(cs3);
            __assembly_array.Add(assembly);
       }

       t = _CifObj->GetTablePtr(blockId, "pdbx_struct_assembly_prop");
       if (!is_empty_table(t)) {
            rowNo = t->GetNumRows();
            idx1 = t->GetColumnIndex("biol_id");
            idx2 = t->GetColumnIndex("type");
            int idx3 = t->GetColumnIndex("value");
            CifString cs3;
            for (int i = 0; i < rowNo; i++) {
                 cs1.Clear();
                 t->GetCell(cs1, idx1, i);
                 if (cs1 == "?" || cs1 == ".") cs1.Clear();
                 cs2.Clear();
                 t->GetCell(cs2, idx2, i);
                 if (cs2 == "?" || cs2 == ".") cs2.Clear();
                 cs3.Clear();
                 t->GetCell(cs3, idx3, i);
                 if (cs3 == "?" || cs3 == ".") cs3.Clear();
                 if (cs3 == "") continue;
                 for (unsigned int j = 0; j < __assembly_array.Length(); j++) {
                      if (__assembly_array[j].assembly_id == cs1) {
                           if (cs2 == "ABSA (A^2)") __assembly_array[j].absa.Copy(cs3);
                           if (cs2 == "SSA (A^2)") __assembly_array[j].ssa.Copy(cs3);
                           if (cs2 == "MORE") __assembly_array[j].more.Copy(cs3);
                           break;
                      }
                 }
            }
       }

       t = _CifObj->GetTablePtr(blockId, "pdbx_struct_assembly_gen");
       if (!is_empty_table(t)) {
            rowNo = t->GetNumRows();
            idx1 = t->GetColumnIndex("assembly_id");
            idx2 = t->GetColumnIndex("oper_expression");
            int idx3 = t->GetColumnIndex("asym_id_list");
            CifString cs3;
            _ASSEMBLY_CHAIN assembly_chain;
            ReVarCifArray<CifString> pdb_chain_ids;
            for (int i = 0; i < rowNo; i++) {
                 cs1.Clear();
                 t->GetCell(cs1, idx1, i);
                 if (cs1 == "?" || cs1 == ".") cs1.Clear();
                 cs2.Clear();
                 t->GetCell(cs2, idx2, i);
                 if (cs2 == "?" || cs2 == ".") cs2.Clear();
                 cs3.Clear();
                 t->GetCell(cs3, idx3, i);
                 if (cs3 == "?" || cs3 == ".") cs3.Clear();
                 if (cs3 == "") continue;
                 for (unsigned int j = 0; j < __assembly_array.Length(); j++) {
                      if (__assembly_array[j].assembly_id == cs1) {
                           get_symmetry_operators(cs2, _CifObj, blockId);
                           get_pdb_chain_id(cs3, pdb_chain_ids, chainids);
                           for (unsigned int l = 0; l < pdb_chain_ids.Length(); l++) {
                                int found = 0;
                                for (unsigned int k = 0; k < __assembly_array[j].assembly_chains.Length(); k++) {
                                     if (__assembly_array[j].assembly_chains[k].PDB_chain_ID == pdb_chain_ids[l]) {
                                          __assembly_array[j].assembly_chains[k].sym_operators += ", ";
                                          __assembly_array[j].assembly_chains[k].sym_operators += cs2;
                                          found = 1;
                                          break;
                                     }
                                }
                                if (!found) {
                                     assembly_chain.PDB_chain_ID.Copy(pdb_chain_ids[l]);
                                     assembly_chain.sym_operators.Copy(cs2);
                                     __assembly_array[j].assembly_chains.Add(assembly_chain);
                                }
                           }
                           break;
                      }
                 }
            }
       }
}

void get_pdb_asym_correspondence(CifFileObj *fobj, const char *blockId,
                ReVarCifArray<_PDB_ASYM_CORR>& chainids)
{
       chainids.Clear();
       ISTable *Table = fobj->GetTablePtr(blockId, "struct_biol_gen");
       CifString cifstring, cifstring1;
       _PDB_ASYM_CORR asym_corr;
       if (Table) {
            int rowNo = Table->GetNumRows();
            int idx1 = Table->GetColumnIndex("symmetry");
            int idx2 = Table->GetColumnIndex("ndb_new_pdb_asym_id");
            int idx3 = Table->GetColumnIndex("asym_id");
            for (int i = 0; i < rowNo; i++) {
                 Table->GetCell(cifstring, idx1, i);
                 if (cifstring != "1_555") continue;

                 Table->GetCell(cifstring, idx2, i);
                 Table->GetCell(cifstring1, idx3, i);
                 if (cifstring != "" && cifstring != "?" && cifstring != ".") {
                      int found = 0;
                      for (unsigned int j = 0; j < chainids.Length(); j++) {
                           if (chainids[j].PDB_chain_ID == cifstring) {
                                found = 1;
                                chainids[j].Asym_IDs.Add(cifstring1);
                                break;
                           }
                      }
                      if (!found) {
                           asym_corr.PDB_chain_ID.Copy(cifstring);
                           asym_corr.Asym_IDs.Clear();
                           asym_corr.Asym_IDs.Add(cifstring1);
                           chainids.Add(asym_corr);
                      }
                 }
            }
       }
}

static void get_symmetry_operators(CifString &oper, CifFileObj *fobj, const char *blockId)
{
       ISTable *t = fobj->GetTablePtr(blockId, "pdbx_struct_oper_list");
       if (is_empty_table(t)) {
            oper.Clear();
            return;
       }

       int rowNo = t->GetNumRows();
       int idx1 = t->GetColumnIndex("id");
       int idx2 = t->GetColumnIndex("name");

       int len = strlen(oper.Text()) + 1;
       char *value = new char[len];
       memset(value, 0, len);
       strcpy(value, oper.Text());
       for (int i = 0; i < len - 1; i++) {
            if (value[i] == ',') value[i] = ' ';
       }
       ndb_clean_string(value);
       oper.Clear();
       CifString cs1, cs2;
       char buffer[100];
       char *p = value;
       while (p) {
            p = get_word(p, buffer);
            if (!strcmp(buffer, "")) continue;
            for (int i = 0; i < rowNo; i++) {
                 cs1.Clear();
                 t->GetCell(cs1, idx1, i);
                 if (cs1 == "?" || cs1 == ".") cs1.Clear();
                 cs2.Clear();
                 t->GetCell(cs2, idx2, i);
                 if (cs2 == "?" || cs2 == ".") cs2.Clear();
                 if (cs2 == "") continue;
                 if (cs1 == buffer) {
                      if (oper != "") oper += ", ";
                      oper += cs2;
                      break;
                 }
            }
       }
       delete [] value;
}

static void get_pdb_chain_id(CifString &chn_id, ReVarCifArray<CifString>& pdb_chain_ids,
               ReVarCifArray<_PDB_ASYM_CORR> &chainids)
{
       pdb_chain_ids.Clear();
       int len = strlen(chn_id.Text()) + 1;
       char *value = new char[len];
       memset(value, 0, len);
       strcpy(value, chn_id.Text());
       for (unsigned int i = 0; i < strlen(value); i++) {
            if (value[i] == ',') value[i] = ' ';
       }
       ndb_clean_string(value);
       char buffer[100];
       char *p = value;
       while (p) {
            p = get_word(p, buffer);
            if (!strcmp(buffer, "")) continue; 
            int found = 0;
            for (unsigned int i = 0; i < chainids.Length(); i++) {
                 for (unsigned int j = 0; j < chainids[i].Asym_IDs.Length(); j++) {
                      if (chainids[i].Asym_IDs[j] == buffer) {
                           int exist = 0;
                           for (unsigned int l = 0; l < pdb_chain_ids.Length(); l++) {
                                if (pdb_chain_ids[l] == chainids[i].PDB_chain_ID) {
                                     exist = 1;
                                     break;
                                }
                           }
                           if (!exist) pdb_chain_ids.Add(chainids[i].PDB_chain_ID);
                           found = 1;
                           break;
                      }
                 }
                 if (found) break;
            }
       }
       delete [] value;
}
