/*
FILE:     util.C
*/
/*
VERSION:  8.120
*/
/*
DATE:     2/12/2008
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2008 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "_atom_change.h"
#include "ndb_misclib.h"
#include "partime.h"
#include "erf_connect.h"
#include "erf_iolib.h"
#include "erf_util.h"
#include "graph_container.h"
#include "sg_center.h"
#include "util.h"
#include "util_global.h"

// #define NUM_CENTER       60
#define NUM_OBS_SPR      237
#define NUM_METAL_X_LIST 36

typedef struct {
        char obsolete_residue[5];
        char supercede_residue[5];
} _OBS_SPR;

static _OBS_SPR _obs_spr_list[NUM_OBS_SPR] = {
       { "119", "P4P" },
       { "148", "BTB" },
       // { "1GP", "G3P" },
       { "1NA", "MAG" },
       { "1NI", "LP1" },
       { "1PY", "PPY" },
       { "2AS", "ACB" },
       { "2LU", "DCL" },
       { "2OG", "AKG" },
       { "2PI", "NVA" },
       { "2PL", "PGA" },
       { "2SI", "IDS" },
       { "345", "CBP" },
       { "395", "961" },
       { "450", "DMQ" },
       { "4RR", "ROL" },
       { "55C", "5CM" },
       { "577", "SB2" },
       { "5HP", "PCA" },
       { "5IT", "5IU" },
       { "5SA", "ARE" },
       { "638", "XV6" },
       { "6AB", "BE2" },
       { "7MQ", "MQ7" },
       { "A35", "A40" },
       { "A39", "A2M" },
       { "AA3", "ABA" },
       { "ABK", "FKI" },
       { "AC4", "AMZ" },
       { "ACU", "ACE" },
       { "ADG", "TOA" },
       { "AFL", "FUL" },
       { "AGC", "GLC" },
       { "AHA", "ACA" },
       { "AMA", "PLA" },
       { "ANE", "ADE" },
       { "ASQ", "PHD" },
       { "B1F", "B2F" },
       { "BAR", "TSA" },
       { "BDN", "BEN" },
       { "BFP", "FBP" },
       { "BH4", "H4B" },
       { "BOT", "THZ" },
       { "BOX", "BEZ" },
       { "BP",  "BAP" },
       { "BPC", "BAP" },
       { "BRI", "1GL" },
       { "BRO", "BR"  },
       { "BTA", "NVA" },
       { "BTC", "CYS" },
       { "BUT", "NBU" },
       { "BZO", "PHQ" },
       { "C25", "C5P" },
       { "C32", "CBR" },
       { "CAY", "CCS" },
       { "CBM", "ACY" },
       { "CBX", "FMT" },
       { "CBZ", "PHQ" },
       { "CEA", "CSO" },
       { "CLO", "CL"  },
       { "CM",  "ACY" },
       { "CMN", "ACI" },
       { "CN",  "CYN" },
       { "CNM", "ACM" },
       { "COE", "MOT" },
       { "CPG", "5GP" },
       { "CRC", "DKA" },
       { "CRY", "GOL" },
       { "CYL", "ACI" },
       { "CYM", "SMC" },
       { "CYP", "GPR" },
       { "D1P", "ORP" },
       { "DCG", "DGP" },
       { "DDB", "MDA" },
       { "DDM", "DMJ" },
       { "DGM", "ASO" },
       { "DHN", "AA4" },
       { "DHO", "DXC" },
       { "DHU", "H2U" },
       { "DIS", "HOH" },
       { "DLA", "LAC" },
       { "DNJ", "NOJ" },
       { "DOH", "BHD" },
       { "DOX", "DIO" },
       { "DSP", "DAS" },
       { "DSX", "BAT" },
       { "DUM", "UNX" },
       { "E4N", "NET" },
       { "EGL", "EDO" },
       { "EHP", "MTY" },
       { "EOX", "EOH" },
       { "FA",  "FOL" },
       { "FAT", "PLM" },
       { "FCY", "CYS" },
       { "FLO", "F"   },
       { "FS3", "F3S" },
       { "FS4", "SF4" },
       { "G32", "6OG" },
       { "G42", "8OG" },
       { "GCL", "XAO" },
       { "GCM", "GLM" },
       { "GDB", "GDN" },
       { "GGL", "GLU" },
       { "GLB", "GAL" },
       { "GLI", "PSI" },
       { "GLW", "G6D" },
       { "GS4", "SGC" },
       { "GSA", "G4S" },
       { "GSD", "SGC" },
       { "GTN", "GNP" },
       { "GTO", "GCP" },
       { "GUR", "GLL" },
       { "HAA", "DHY" },
       { "HAC", "ALC" },
       { "HAP", "PLH" },
       { "HBL", "HBI" },
       { "HGC", "MMC" },
       { "HHP", "PH2" },
       { "HMP", "HMI" },
       { "HNP", "H5P" },
       { "HP3", "PGR" },
       { "HPB", "PR0" },
       { "HPC", "PHM" },
       { "HPG", "PDO" },
       { "HQ",  "HQO" },
       { "HV5", "TBG" },
       { "ICI", "ICT" },
       { "IDO", "IOD" },
       { "INY", "CRP" },
       { "IOH", "IPA" },
       { "IPS", "PO4" },
       { "ISB", "ALQ" },
       { "ITS", "4IP" },
       { "IU6", "CHC" },
       { "KGR", "GLR" },
       { "LAU", "DAO" },
       { "LIN", "AAE" },
       { "LLA", "HFA" },
       { "LN4", "BDI" },
       { "LOF", "HFA" },
       { "LP2", "LP1" },
       { "LTR", "TRP" },
       { "MAM", "MMA" },
       { "MBA", "DCI" },
       { "MDG", "M7G" },
       { "MEP", "T23" },
       { "MFA", "MFU" },
       { "MGA", "MBG" },
       { "MGY", "SAR" },
       { "MHM", "HEM" },
       { "MOL", "BIC" },
       { "MP3", "3PG" },
       { "MPS", "VXA" },
       { "MTO", "HOH" },
       { "N2B", "PYL" },
       { "NAC", "A3D" },
       { "NAH", "NAD" },
       { "NAN", "SIA" },
       { "NC2", "NC1" },
       { "NED", "IDG" },
       { "NEM", "HIC" },
       { "NEV", "NVP" },
       { "NEW", "PCQ" },
       { "NFB", "NFO" },
       { "NGL", "ASG" },
       { "NIV", "NVP" },
       { "NMO", "NO"  },
       { "NNS", "TYL" },
       { "NTY", "GHP" },
       { "NWB", "CHH" },
       { "O2",  "OXY" },
       { "OET", "ETH" },
       { "OIP",  "DI" },
       { "OLE", "1LU" },
       { "OLI", "OLA" },
       { "OMB", "MOH" },
       { "OR5", "R5P" },
       { "OTB", "BOC" },
       { "OTE", "C8E" },
       { "OX",  "O"   },
       { "OXO", "O"   },
       { "PA4", "IDG" },
       { "PAG", "2PG" },
       { "PAS", "PHD" },
       { "PC2", "PC1" },
       { "PCC", "PCA" },
       { "PDL", "PP3" },
       { "PEI", "LEA" },
       { "PF",  "PRL" },
       { "PGC", "PGH" },
       { "PGQ", "PGO" },
       { "PGY", "PG9" },
       { "PIG", "PGE" },
       { "PLY", "PLM" },
       { "PNL", "CBG" },
       { "POC", "PC"  },
       { "PPM", "GB"  },
       { "PTP", "THP" },
       { "PVL", "PYR" },
       { "Q72", "672" },
       { "QX",  "QUI" },
       { "RDP", "R1P" },
       { "RIC", "RBZ" },
       { "RON", "NVA" },
       { "RT",  "5MU" },
       { "S4U", "4SU" },
       { "SAA", "APG" },
       { "SBU", "NBU" },
       { "SCP", "DCS" },
       { "SEA", "DHL" },
       { "SEG", "SER" },
       { "SEO", "BME" },
       { "SEU", "ITU" },
       { "SFN", "SO3" },
       { "SI2", "SIA" },
       { "SOM", "MPS" },
       { "SPG", "PGS" },
       { "STO", "STU" },
       { "STY", "TYS" },
       { "SUL", "SO4" },
       { "T37", "NYM" },
       { "TFH", "HDZ" },
       { "THB", "H4B" },
       { "THQ", "TZP" },
       { "TIB", "TB9" },
       { "TMB", "TBM" },
       { "TMN", "TRS" },
       { "TPH", "HPH" },
       { "TRB", "TB9" },
       { "TRG", "MLY" },
       { "TS",  "PST" },
       { "TS3", "GCG" },
       { "U18", "F89" },
       { "U25", "U5P" },
       { "UIC", "GRL" },
       { "XL1", "SCC" },
       { "XPP", "HP2" },
       { "ZN2", "ZN"  }
};

typedef struct {
       char name[3];
       int  flag;
} _METAL_X_LIST;

static _METAL_X_LIST _metal_x_list[NUM_METAL_X_LIST] = {
       { "AL", 2 },
       { "AS", 2 },
       { "BA", 2 },
       { "C",  1 },
       { "CA", 2 },
       { "CD", 2 },
       { "CO", 2 },
       { "CS", 2 },
       { "CU", 2 },
       { "FE", 2 },
       { "GD", 2 },
       { "HG", 2 },
       { "IN", 2 },
       { "IR", 2 },
       { "K",  2 },
       { "LI", 2 },
       { "MG", 2 },
       { "MN", 2 },
       { "MO", 2 },
       { "N",  1 },
       { "NA", 2 },
       { "NI", 2 },
       { "O",  1 },
       { "P",  1 },
       { "PB", 2 },
       { "PT", 2 },
       { "RB", 2 },
       { "S",  1 },
       { "SM", 2 },
       { "SR", 2 },
       { "TE", 2 },
       { "TL", 2 },
       { "V",  2 },
       { "W",  2 },
       { "YB", 2 },
       { "ZN", 2 }
};
/*
static const char *_full_name_and_initials[NUM_CENTER] = {
       "Accelerated Technologies Center for Gene to 3D Structure",
       "Accelerated Technologies Center for Gene to 3D Structure (ATCG3D)",
       "Bacterial targets at IGS-CNRS, France",
       "Bacterial targets at IGS-CNRS, France (BIGS)",
       "Berkeley Structural Genomics Center",
       "Berkeley Structural Genomics Center (BSGC)",
       "Center for Eukaryotic Structural Genomics",
       "Center for Eukaryotic Structural Genomics (CESG)",
       "Center for High-Throughput Structural Biology",
       "Center for High-Throughput Structural Biology (CHTSB)",
       "Center for Structures of Membrane Proteins",
       "Center for Structures of Membrane Proteins (CSMP)",
       "Integrated Center for Structure and Function Innovation",
       "Integrated Center for Structure and Function Innovation (ISFI)",
       "Israel Structural Proteomics Center",
       "Israel Structural Proteomics Center (ISPC)",
       "Joint Center for Structural Genomics",
       "Joint Center for Structural Genomics (JCSG)",
       "Marseilles Structural Genomics Program @ AFMB",
       "Marseilles Structural Genomics Program @ AFMB (MSGP)",
       "Midwest Center for Structural Genomics",
       "Midwest Center for Structural Genomics (MCSG)",
       "Montreal-Kingston Bacterial Structural Genomics Initiative",
       "Montreal-Kingston Bacterial Structural Genomics Initiative (BSGI)",
       "Mycobacterium Tuberculosis Structural Proteomics Project",
       "Mycobacterium Tuberculosis Structural Proteomics Project (XMTB)",
       "New York Consortium on Membrane Protein Structure",
       "New York Consortium on Membrane Protein Structure (NYCOMPS)",
       "New York Structural Genomics Research Consortium",
       "New York Structural Genomics Research Consortium (NYSGRC)",
       "New York Structural GenomiX Research Consortium",
       "New York Structural GenomiX Research Consortium (NYSGXRC)",
       "Northeast Structural Genomics Consortium",
       "Northeast Structural Genomics Consortium (NESG)",
       "Ontario Centre for Structural Proteomics",
       "Ontario Centre for Structural Proteomics (OCSP)",
       "Oxford Protein Production Facility",
       "Oxford Protein Production Facility (OPPF)",
       "Paris-Sud Yeast Structural Genomics",
       "Paris-Sud Yeast Structural Genomics (YSG)",
       "Protein Structure Factory",
       "Protein Structure Factory (PSF)",
       "RIKEN Structural Genomics/Proteomics Initiative",
       "RIKEN Structural Genomics/Proteomics Initiative (RSGI)",
       "Scottish Structural Proteomics Facility",
       "Scottish Structural Proteomics Facility (SSPF)",
       "Southeast Collaboratory for Structural Genomics",
       "Southeast Collaboratory for Structural Genomics (SECSG)",
       "Structural Genomics Consortium",
       "Structural Genomics Consortium (SGC)",  
       "Structural Genomics of Pathogenic Protozoa Consortium",
       "Structural Genomics of Pathogenic Protozoa Consortium (SGPP)",
       "Structural Proteomics in Europe",
       "Structural Proteomics in Europe (SPINE)",
       "Structure 2 Function Project",
       "Structure 2 Function Project (S2F)",
       "TB Structural Genomics Consortium",
       "TB Structural Genomics Consortium (TBSGC)",
       "The Integrated Center for Structure and Function Innovation",
       "The Integrated Center for Structure and Function Innovation (ISFI)"
};
*/
static int  read_reserved_word_file(const char *file);
static void change_reserved_word(char *string);

#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int corres_compare(const void *coores1, const void *coores2);
static int residue_name_compare(const void *group1, const void *group2);
static int atom_name_compare(const void *group1, const void *group2);
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif

int ndb_filterlib_init(const char *path)
{  
       if (!read_ndb_tokens()) return FALSE;
       if (!init_connect_dictionary(path)) return FALSE;
       init_graph_container();
       atom_changes_init();

       Num_Reserved_Word = read_reserved_word_file(Reserved_Word_File);
 
       return TRUE;
}

void ndb_filterlib_close()
{
       close_connect_dictionary();
       close_graph_container();
}

int find_residue_type(const char *residueName, int &exist_unrecognized)
{
       exist_unrecognized = 0;
       ConnectFormat *drug = find_drug(residueName);
       if (drug) return (drug->ndbAtomType());
       else {
            if (strcasecmp(residueName, "UNL")) {
                 ndb_log_message(NDB_MSG_ERR, "Unrecognized residue name %s\n",
                        residueName);
                 exist_unrecognized = 1;
            }
            return (HETAIN_TOKEN);
       }
}

ConnectFormat *find_drug(const char *drugName)
{
       ConnectFormat *drug = ConnectDictionary->find_drug(drugName);
       if (drug && drug->ndbAtomType() < 0) {
            int card_id = find_card_id(drug->AtomType()) + 1;
            drug->set_ndbAtomType(card_id);
       }
       return drug;
}

ConnectFormat *find_drug(const char *drugName, const int no_oxygen)
{
       char name[5];
       strcpy(name, drugName);
       if (no_oxygen) {
            if (drugName[0] == '+')
                 sprintf(name, "%sNOO", &drugName[1]);
            else sprintf(name, "%sNOO", drugName);
       }
       return (find_drug(name));
}

void ndb_name_conversion_first_last(char *string)
{
       int i, len;
       char first[500], last[500], center[200];

       if (string == NULL || string[0] == '\0') return;

       for (i = 0; i < NUM_CENTER; i++) {
            if (!strcasecmp(string, _sg_info[i]._center_name)) break;
            sprintf(center, "%s (%s)", _sg_info[i]._center_name, _sg_info[i]._center_initial);
            if (!strcasecmp(string, center)) break;
       }
       if (i < NUM_CENTER) return;

       len = strlen(string);
       for (i = len - 1; i >= 0; i--)
            if (string[i] == ',') break;

       if (i > 0) {
            strcpy(first, &string[i+1]);
            string[i] = '\0';
            strcpy(last, string);
            ndb_clean_string(first);
            ndb_clean_string(last);
            if (first[strlen(first)-1] == '.')
                 sprintf(string, "%s%s", first, last);
            else sprintf(string, "%s %s", first, last);
       }
}

void ndb_name_conversion_last_first(char *string)
{
       int i, len;
       char first[500], last[500], center[200];

       if (string == NULL || string[0] == '\0') return;

       for (i = 0; i < NUM_CENTER; i++) {
            if (!strcasecmp(string, _sg_info[i]._center_name)) break;
            sprintf(center, "%s (%s)", _sg_info[i]._center_name, _sg_info[i]._center_initial);
            if (!strcasecmp(string, center)) break;
       }
       if (i < NUM_CENTER) return;

       len = strlen(string);
       for (i = len - 1; i >= 0; i--)
            if (string[i] == '.' && i != (len - 1)) break;

       if (i > 0) {
            strcpy(last, &string[i+1]);
            string[i+1] = '\0';
            strcpy(first, string);
            sprintf(string, "%s, %s", last, first);
            ndb_clean_string(string);
            return;
       }

       len = strlen(string);
       for (i = len - 1; i >= 0; i--)
            if (string[i] == ' ') break;

       if (i > 0) {
            strcpy(last, &string[i+1]);
            string[i+1] = '\0';
            strcpy(first, string);
            sprintf(string, "%s, %s", last, first);
            ndb_clean_string(string);
       }
}

int check_extra_space_in_name(const char *string)
{
       int i, len;
       char first[500], center[200];

       if (string == NULL || string[0] == '\0') return 0;

       for (i = 0; i < NUM_CENTER; i++) {
            if (!strcasecmp(string, _sg_info[i]._center_name)) break;
            sprintf(center, "%s (%s)", _sg_info[i]._center_name, _sg_info[i]._center_initial);
            if (!strcasecmp(string, center)) break;
       }
       if (i < NUM_CENTER) return 0;

       len = strlen(string);
       for (i = len - 1; i >= 0; i--)
            if (string[i] == ',') break;

       if (i > 0) {
            strcpy(first, &string[i+1]);
            ndb_clean_string(first);
            for (i = 0; i < (int) strlen(first); i++) {
                 if (first[i] == ' ') return 1;
            }
       }
       return 0;
}

void get_field_format(const int card_id, const int field_no, char *FormatString,
         const int IsJRNL, const int format)
{
       int  tmp_field_no;
     
       if (IsJRNL == FALSE) {
            if (card_id != TER_TOKEN - 1) 
                 tmp_field_no = get_field_no(card_id, field_no, format);
            else tmp_field_no = field_no;
            if (NdbToken[card_id].FieldList[tmp_field_no].FieldType > 2) {
                 if ((NdbToken[card_id].FieldList[tmp_field_no].FieldType % 2) == 1)
                      sprintf(FormatString, "%%-%ds",
                              NdbToken[card_id].FieldList[field_no].FieldWidth);
                 else sprintf(FormatString, "%%%ds",
                              NdbToken[card_id].FieldList[field_no].FieldWidth);
            } else sprintf(FormatString, "%%%d.%dlf",
                     NdbToken[card_id].FieldList[field_no].FieldWidth,
                     NdbToken[card_id].FieldList[field_no].FieldPrec);
       } else {
            if (JrnlToken[card_id].FieldList[field_no].FieldType > 2) {
                 if ((JrnlToken[card_id].FieldList[field_no].FieldType % 2) == 1)
                      sprintf(FormatString, "%%-%ds",
                              JrnlToken[card_id].FieldList[field_no].FieldWidth);
                 else sprintf(FormatString, "%%%ds",
                              JrnlToken[card_id].FieldList[field_no].FieldWidth);
            } else sprintf(FormatString, "%%%dd",
                     JrnlToken[card_id].FieldList[field_no].FieldWidth);
       }
}

int get_field_no(const int card_id, const int field_no, const int format)
{
       int  atomNameField = 2, residueNameField = 4;
       int  chainIdField = 5, residueNoField = 6;
       int  pdbResidueNoField = 15, pdbResidueNameField = 16;
       int  pdbChainIdField = 17, pdbAtomNameField = 18;
       int  altResidueNoField = 19, altResidueNameField = 20;
       int  altChainIdField = 21, altAtomNameField = 22;
       int  tmp_field_no;
     
       tmp_field_no = field_no;
       if (is_atom_card(card_id)) {
            if (format == NDB_FILE_FORMAT_PDB) {
                 if (field_no == atomNameField)
                      tmp_field_no = pdbAtomNameField;
                 else if (field_no == residueNoField)
                      tmp_field_no = pdbResidueNoField;
                 else if (field_no == residueNameField)
                      tmp_field_no = pdbResidueNameField;
                 else if (field_no == chainIdField)
                      tmp_field_no = pdbChainIdField;
            } else if (format == NDB_FILE_FORMAT_ALIAS) {
                 if (field_no == atomNameField)
                      tmp_field_no = altAtomNameField;
                 else if (field_no == residueNoField)
                      tmp_field_no = altResidueNoField;
                 else if (field_no == residueNameField)
                      tmp_field_no = altResidueNameField;
                 else if (field_no == chainIdField)
                      tmp_field_no = altChainIdField;
            }
       }
       return tmp_field_no;
}

void get_print_atom(const char *residueName, const char *atomName, char *printName)
{
       ConnectFormat *drug = find_drug(residueName);
       if (strlen(atomName) == 4 || !drug) {
            sprintf(printName, "%4s", atomName);
            return;
       }

       _SIndex *index = drug->find_atom(atomName);
       if (!index) {
            sprintf(printName, "%4s", atomName);
            return;
       }

       const char *p = strstr(atomName, drug->atoms(index->pos())->atomtype());
       if (p == atomName) {
            if (strlen(drug->atoms(index->pos())->atomtype()) == 2)
                 sprintf(printName, "%-4s", atomName);
            else sprintf(printName, " %-3s", atomName);
       } else sprintf(printName, "%-4s", atomName);
}

void print_atom_atomName_field(const char *atomtype, const int pos, char *temp_line,
            char *atomName, const char *residueName, const int align_i)
{
       int l, len, n, found = 0, align;
       char atom_type[5], *p = NULL;
       _SIndex *index = NULL;
       ConnectFormat *drug = NULL;

       align = align_i;
       if (residueName && !strcmp(residueName, "HOH") && atomName[0] == 'O') {
            sprintf(&temp_line[pos]," %c  ", 'O');
            return;
       }
       l = strlen(atomName);
       if (l == 4 || l == 0) {
            sprintf(&temp_line[pos], "%-4s", atomName);
            return;
       }

       if (!strcmp(atomtype, "") || align < 0) {
            if (residueName && strcmp(residueName, ""))
                 drug = find_drug(residueName);
            if (drug) {
                 index = drug->find_atom(atomName);
                 if (index) {
                      align = drug->atoms(index->pos())->align_pos();
                      found = 1;
                 }
            }
            if (!found) {
                 if (!strcmp(atomtype, "")) {
                      memset(atom_type, 0, 5);
                      ndb_get_atom_symbol(atomName, atom_type);
                 } else strcpy(atom_type, atomtype);
                 p = strstr(atomName, atom_type);
                 n = p - atomName;
                 len = strlen(atom_type);
                 if (n == 1 || len == 2 || len == 1 && atomName[1] == atom_type[0] &&
                     atom_type[0] != 'D' && atom_type[0] != 'H' && atom_type[0] != 'C')
                      align = 0;
                 else align = 1;
            }
       }
       if (align == 0) {
            if (l == 1) sprintf(&temp_line[pos], "%s   ", atomName);
            if (l == 2) sprintf(&temp_line[pos], "%2s  ", atomName);
            else if (l == 3) sprintf(&temp_line[pos], "%3s ", atomName);
       } else {
            if (l == 1) sprintf(&temp_line[pos], " %s  ", atomName);
            else if (l == 2) sprintf(&temp_line[pos], " %2s ", atomName);
            else if (l == 3) sprintf(&temp_line[pos], " %3s", atomName);
       }
}

int is_a_nonstandard_trna(const char *residueName)
{
       int i;
       if (residueName == NULL) return FALSE;

       for (i = 0; i < MxTRnaNo; i++) {
            if (!strcmp(tRnaList[i], residueName)) return TRUE;
       }
       return FALSE;
}

void ndb_convert_time_format(const int format, char *_indatestring,
       char *outdatestring, int fieldLength)
/* ---------------------------------------------------------------
 * converts different format of time output
 * NDB format: fieldLength <= 9 day-mon-year mon is in JAN, FEB,...format
 *                                           and year is two digit
 *             fieldLength <16  year-mon-day:hour:min 
 *                                           mon are in 1, 2, 3 
 *                                           year is four digit
 *             fieldLength 
 * SYBASE format: year-mon-day hour:min mon is in JAN, FEB, ... format
 * PDB format: day-mon-year mon is in JAN, FEB, ... form
 *             is fieldLength <10, year is two digit such as 96
 *                            > 10 year is four digit such as 1996
 * ---------------------------------------------------------------*/
{
  char *buf;
  char mon[MxStrLen], day[MxStrLen], year[MxStrLen], tmp[MxStrLen];
  char hour[MxStrLen], min[MxStrLen], datestring[MxStrLen];
  char indatestring[MxStrLen];
  struct partime p;
  unsigned int i, j;
  char mons[12][MxStrLen] =  
    { "JAN", "FEB", "MAR", "APR", "MAY", "JUN", 
	"JUL", "AUG", "SEP", "OCT", "NOV", "DEC"};
  
  if (!_indatestring) {
    strcpy(outdatestring, "");
    return;
  }

  strcpy(indatestring, _indatestring);
  buf = strchr(indatestring, ';');
  if (buf) *buf = '\0';
  buf = strchr(indatestring, ' ');
  if (buf) *buf = '\0';

  if (format == NDB_FILE_FORMAT_NDB && fieldLength < 11)
       fieldLength = 11;
  /*
   *  Try to catch year ranges like 19xx-19xx as these are not
   *  handled by the partime() function.  
   */
  int exist_day = 1;
  if (strlen(indatestring) == 4 && isdigit(indatestring[0]) &&
      isdigit(indatestring[1]) && isdigit(indatestring[2]) && isdigit(indatestring[3])) {
    strcpy(outdatestring, indatestring);
    return;
  } else if ((strlen(indatestring) == 9) && (indatestring[4] == '-')) {
    strcpy(datestring,"01-JAN-");
    strncat(datestring,indatestring,4); 
  } else if (strlen(indatestring) == 8 && isalpha(indatestring[0]) && isalpha(indatestring[1]) &&
     isalpha(indatestring[2]) && indatestring[3] == '-' && isdigit(indatestring[4]) &&
     isdigit(indatestring[5]) && isdigit(indatestring[6]) && isdigit(indatestring[7])) {
    sprintf(datestring,"01-%s", indatestring);
    exist_day = 0;
  } else if (strlen(indatestring) == 6 && isalpha(indatestring[0]) && isalpha(indatestring[1]) &&
     isalpha(indatestring[2]) && indatestring[3] == '-' && isdigit(indatestring[4]) &&
     isdigit(indatestring[5])) {
    sprintf(datestring,"01-%s", indatestring);
    exist_day = 0; 
  } else if (strlen(indatestring) == 7 && isdigit(indatestring[0]) && isdigit(indatestring[1]) &&
     isdigit(indatestring[2]) && isdigit(indatestring[3]) && indatestring[4] == '-' &&
     isdigit(indatestring[5]) && isdigit(indatestring[6])) {
    sprintf(datestring,"%s-01", indatestring);
    exist_day = 0;
  }  else {
    strcpy(datestring, indatestring);
  }
  /* This part is for sybase standard date string
   *  sybase date string format is like Mon Date Year Hour:MIN:MIN:XXXAM
   *  The place :XXX is not able to be scaned by partime
   */

  for (i=0; i<strlen(datestring); i++)
    if (!strncmp(&datestring[i], ":000",4)) break;
  if (i != strlen(datestring)) {
    j = i+4;
    for (; i<j; i++)
      datestring[i] = ' ';
  }

  if (format == NDB_FILE_FORMAT_SYBASE && isalpha(datestring[0])) {
       sprintf(tmp, "01-%s", datestring);
       strcpy(datestring, tmp);
  }

  buf = partime(datestring, &p);
  
  if (strcmp(buf,"")) {
    // pdb_err_type_bib_section++;
    ndb_log_message(NDB_MSG_ERR,"Date %s format error at  *%s*\n", 
		    indatestring, buf);
    if (format != NDB_FILE_FORMAT_SYBASE) {
      ndb_log_message(NDB_MSG_ERR,"Date format is not converted\n");
      strncpy(outdatestring, indatestring, fieldLength);
      return;
    }
  }
  if (format == NDB_FILE_FORMAT_PDB || format == NDB_FILE_FORMAT_SYBASE ||
	   fieldLength < 10) {
    if (p.tm.tm_mon < 0 || p.tm.tm_mon >= 12)
      strcpy(mon, mons[0]);
    else
      strcpy(mon, mons[p.tm.tm_mon]);
  }
  else {
    if (p.tm.tm_mon <0 || p.tm.tm_mon >=12)
      sprintf(mon, "01");
    else if (p.tm.tm_mon < 9)
      sprintf(mon, "0%d", p.tm.tm_mon+1);
    else
      sprintf(mon, "%d", p.tm.tm_mon+1);
  }
  if (p.tm.tm_mday > 0 && p.tm.tm_mday <= 31) {
    if (p.tm.tm_mday < 10)
      sprintf(day, "0%d", p.tm.tm_mday);
    else
      sprintf(day, "%d", p.tm.tm_mday);
  }
  else
    strcpy(day,  "01");

  if (format == NDB_FILE_FORMAT_PDB) {
    if (fieldLength <= 10)  {
      if (p.tm.tm_year <0) 
	  sprintf(year, "00");
      else if (p.tm.tm_year % 100 < 10) 
	  sprintf(year, "0%d", p.tm.tm_year %100);
      else 
	sprintf(year, "%d", p.tm.tm_year %100);
    }
    else {
      if (p.tm.tm_year > 1000)
	sprintf(year, "%d", p.tm.tm_year);
      else if (p.tm.tm_year > 100)
	sprintf(year, "1%d", p.tm.tm_year);
      else if (p.tm.tm_year > 60)
        sprintf(year, "19%d", p.tm.tm_year);
      else if (p.tm.tm_year > 10)
	sprintf(year, "20%d", p.tm.tm_year);
      else
	sprintf(year, "200%d", p.tm.tm_year);
    }
    if (exist_day)
         sprintf(outdatestring, "%s-%s-%s", day, mon, year);
    else sprintf(outdatestring, "%s-%s", mon, year);
  }
  else {
    if (p.tm.tm_year <0 && fieldLength >= 10)
      sprintf(year, "2000");
    else if (p.tm.tm_year <0)
      sprintf(year, "00");
    else if (fieldLength <= 9 && format != NDB_FILE_FORMAT_SYBASE) 
      sprintf(year, "%d", p.tm.tm_year%100);
    else if (p.tm.tm_year < 10)
      sprintf(year, "200%d", p.tm.tm_year);
    else if (p.tm.tm_year < 60)
      sprintf(year, "20%d", p.tm.tm_year);
    else if (p.tm.tm_year < 100)
      sprintf(year, "19%d", p.tm.tm_year);
    else if (p.tm.tm_year < 1000)
      sprintf(year, "1%d", p.tm.tm_year );
    else
      sprintf(year, "%d", p.tm.tm_year );
    
    if (p.tm.tm_min > 60 || p.tm.tm_min < 0 )
      strcpy(min,"00");
    else if (p.tm.tm_min < 10)
      sprintf(min, "0%d", p.tm.tm_min );
    else
      sprintf(min, "%d", p.tm.tm_min );
    
    if (p.tm.tm_hour > 24 || p.tm.tm_hour < 0 )
      strcpy(hour,"00");
    else if ( p.tm.tm_hour < 10 )
      sprintf(hour, "0%d", p.tm.tm_hour );
    else
      sprintf(hour, "%d", p.tm.tm_hour );

    if (format == NDB_FILE_FORMAT_SYBASE)
      sprintf(outdatestring, "%s-%s-%s %s:%s", year, mon, day, hour, min);
    else if (fieldLength <= 9) {
      if (exist_day)
           sprintf(outdatestring, "%s-%s-%s", day,  mon, year);
      else sprintf(outdatestring, "%s-%s", mon, year);
    } else if (fieldLength < 16) {
      if (exist_day)
           sprintf(outdatestring, "%s-%s-%s", year, mon, day);
      else sprintf(outdatestring, "%s-%s", year, mon);
    } else {
      if (exist_day)
           sprintf(outdatestring, "%s-%s-%s:%s:%s", year, mon, day, hour, min);
      else sprintf(outdatestring, "%s-%s", year, mon);
    }
  }
}

void ndb_set_string_to_mixed(char *string)
{
      unsigned int i, k;

      ndb_clean_string_dash(string);
      
      k = strlen(string);
      for (i = 0; i < k; i++) {
           if (islower(string[i])) return;
      }

      for (i = 0; i < k; i++) {
           if (string[i] == '*' || string[i] == '$') {
                strcpy(&string[i], &string[i+1]);
                i--; k--;
           }
      }      

      for (i = 0; i < strlen(string); i++)
           if (islower(string[i])) break;

      if (i == strlen(string)) {
           ndb_set_string_to_lower(string);
           string[0] = toupper(string[0]);
           for (i = 1; i < strlen(string); i++)
                if ((string[i - 1] < 'a' || string[i - 1] > 'z') &&
                    (string[i - 1] < 'A' || string[i - 1] > 'Z'))
                     string[i] = toupper(string[i]);
      }

      change_reserved_word(string);
      string[0] = toupper(string[0]);
}

float get_molecule_weight(const char *name)
{
       char *p, mol[MxStrLen], buffer[10];
       float weight;

       if (!strcmp(name, "HOH") ||
           !strcmp(name, "DOD")) return (18.0152);
       weight = 0.0;

       ConnectFormat *drug = find_drug(name);
       if (drug == NULL) return (-1.0);

       strcpy(mol, drug->formula());
       ndb_clean_string(mol);
       if (mol[0] == '\0') return (-1.0);
       p = mol;
       while (p) {
            p = get_word(p, buffer);
            if (buffer[0] != '\0') weight += get_weight(buffer);
       }

       return (weight);
}

double mod(const double a, const double low, const double high)
{
       double ranged = high - low;
       int i = (int) floor(a / ranged);
       double v = a - (double) i * ranged;
       if (v > high) v -= ranged;
       return v;
}

int is_a_standard_residue(const char *resname)
{
       static char standard_residue_list[36][5] = {"A", "DA", "C", "DC", "G", "DG", "I",
             "DI",  "T",   "DT",  "U",   "DU",  "N",   "ALA", "ARG", "ASN", "ASP", "ASX",
             "CYS", "GLN", "GLU", "GLX", "GLY", "HIS", "ILE", "LEU", "LYS", "MET", "PHE",
             "PRO", "SER", "THR", "TRP", "TYR", "VAL", "UNK"};
       char Name[10];
       int i;

       if (resname[0] == '+') strcpy(Name, &resname[1]);
       else                   strcpy(Name, resname);

       for (i = 0; i < 36; i++) {
            if (!strcasecmp(Name, standard_residue_list[i])) return 1;
       }
       return 0;
}

int is_a_nonstandard_residue(const char *resname)
{
       static char standard_residue_list[46][5] = {"A", "DA", "C", "DC", "G", "DG", "I",
             "DI",  "T",   "DT",  "U",   "DU",  "N",   "ALA", "ARG", "ASN", "ASP", "ASX",
             "CYS", "GLN", "GLU", "GLX", "GLY", "HIS", "ILE", "LEU", "LYS", "MET", "PHE",
             "PRO", "SER", "THR", "TRP", "TYR", "VAL", "UNK", "H2O", "WAT", "TIP", "WTR",
             "HOH", "URI", "THY", "ADE", "CYT", "GUA" };
       char Name[10];
       int i;

       if (resname[0] == '+') strcpy(Name, &resname[1]);
       else                   strcpy(Name, resname);

       for (i = 0; i < 46; i++) {
            if (!strcasecmp(Name, standard_residue_list[i])) return 0;
       }
       return 1;
}

int is_standard_aa_residue(const char *resname)
{
       static char standard_residue_list[22][5] = {
             "ALA", "ARG", "ASN", "ASP", "ASX", "CYS", "GLN", "GLU", "GLX", "GLY", "HIS",
             "ILE", "LEU", "LYS", "MET", "PHE", "PRO", "SER", "THR", "TRP", "TYR", "VAL"};

       for (int i = 0; i < 22; i++) {
            if (!strcasecmp(resname, standard_residue_list[i])) return 1;
       }
       return 0;
}

int is_aa_residue(const char *resname)
{
       static char standard_aa_residue_list[170] = "ALA ARG ASN ASP CYS GLN GLU GLY HIS ILE LEU LYS MET PHE PRO SER THR TRP TYR VAL DAL DAR DAS DSP DCY DGN DGL DIL DLE DLY DPN DPR DSN DTH DTR DTY DVA";
       char *p = strstr(standard_aa_residue_list, resname);
       if (p) return 1;
       else   return 0;
}

int is_na_residue(const char *resname)
{
       int i;
       static char na_residue_list[12][5] = { "A", "DA", "C", "DC", "G", "DG", "I", "DI", "T", "DT", "U", "DU" };
       char Name[10];
       if (resname[0] == '+') strcpy(Name, &resname[1]);
       else                   strcpy(Name, resname);
       for (i = 0; i < 12; i++)
            if (!strcmp(Name, na_residue_list[i])) break;
       if (i < 12) return 1;
       else return 0;
}

int res_name_cmp(const char *residueName1, const char *residueName2)
{
       if (residueName1 == residueName2) return 0;
       else if (residueName1 == NULL || 
                residueName2 == NULL) return 1;
       if (!strcmp(residueName1, residueName2)) return 0;
       else if (residueName1[0] == '+' && 
             !strcmp(&residueName1[1], residueName2)) return 0;
       else if (residueName2[0] == '+' && 
             !strcmp(&residueName2[1], residueName1)) return 0;
       else if (!strcmp(residueName1, "DA") && !strcmp(residueName2, "A") ||
                !strcmp(residueName1, "A") && !strcmp(residueName2, "DA") ||
                !strcmp(residueName1, "DC") && !strcmp(residueName2, "C") ||
                !strcmp(residueName1, "C") && !strcmp(residueName2, "DC") ||
                !strcmp(residueName1, "DG") && !strcmp(residueName2, "G") ||
                !strcmp(residueName1, "G") && !strcmp(residueName2, "DG") ||
                !strcmp(residueName1, "DI") && !strcmp(residueName2, "I") ||
                !strcmp(residueName1, "I") && !strcmp(residueName2, "DI") ||
                !strcmp(residueName1, "DT") && !strcmp(residueName2, "T") ||
                !strcmp(residueName1, "T") && !strcmp(residueName2, "DT") ||
                !strcmp(residueName1, "DU") && !strcmp(residueName2, "U") ||
                !strcmp(residueName1, "U") && !strcmp(residueName2, "DU"))
            return 0;
       else return 1;
}

int res_name_cmp_na(const char *residueName1, const char *residueName2)
{
       char name1[5], name2[5];
       strcpy(name1, residueName1);
       strcpy(name2, residueName2);
       get_parent_name(name1);
       get_parent_name(name2);
       return (res_name_cmp(name1, name2));
}

void get_parent_name(char *real_name)
{
       NAME_CORR tmp, *exist;
       strcpy(tmp.real_name, real_name);
       exist = (NAME_CORR *) bsearch(&tmp, name_corr, NUM_NA,
                 sizeof(NAME_CORR), corres_compare);
       if (exist) strcpy(real_name, exist->parent);
       else {
            ConnectFormat *drug = find_drug(real_name);
            if (drug && drug->parent_residue() &&
                strcmp(drug->parent_residue(), "")) {
                 strcpy(tmp.real_name, drug->parent_residue());
                 exist = (NAME_CORR *) bsearch(&tmp, name_corr, NUM_NA,
                           sizeof(NAME_CORR), corres_compare);
                 if (exist) strcpy(real_name, exist->parent);
            }
       }
}

char *get_oneletter_code(char *threeletter)
{
       for (int i = 0; i < AANUMBER; i++) {
            if (!strncasecmp(aacodes[i].a3let, threeletter, 3))
                 return aacodes[i].a1let;
       }

       return threeletter;
}

char *get_threeletter_code(const char c)
{
      static char buff[4];
      sprintf(buff, "%c", c);
      for (int i = 0; i < AANUMBER; i++) {
           if (!strncasecmp(aacodes[i].a1let, buff, 1))
                return aacodes[i].a3let;
      }
      return buff;
}

int find_obsolete_supercede_residue(char *resname)
{
       _OBS_SPR obs_spr_tmp, *exist = NULL;
       strcpy(obs_spr_tmp.obsolete_residue, resname);
       exist = (_OBS_SPR *) bsearch(&obs_spr_tmp, _obs_spr_list, NUM_OBS_SPR,
                     sizeof(_OBS_SPR), residue_name_compare);
       if (exist) {
            strcpy(resname, exist->supercede_residue);
            return 1;
       }
       return 0;
}

int is_obsolete_residue(const char *resname)
{
       _OBS_SPR obs_spr_tmp, *exist = NULL;
       strcpy(obs_spr_tmp.obsolete_residue, resname);
       exist = (_OBS_SPR *) bsearch(&obs_spr_tmp, _obs_spr_list, NUM_OBS_SPR,
                     sizeof(_OBS_SPR), residue_name_compare);
       if (exist)
            return 1;
       else return 0;
}

int find_metal_flag(const char *name)
{
       _METAL_X_LIST tmp, *exist = NULL;
       strcpy(tmp.name, name);
       exist = (_METAL_X_LIST *) bsearch(&tmp, _metal_x_list, NUM_METAL_X_LIST,
                     sizeof(_METAL_X_LIST), atom_name_compare);
       if (exist) return exist->flag;
       return 0;
}

static int read_reserved_word_file(const char *file)
{
       int  num;
       char line[1000], buf1[30], buf2[30], buf3[30];
       FILE *fp;

       sprintf(line, "%s/data/ascii/%s", root_dir, file);
       if ((fp = fopen(line, "r")) == NULL) {
            printf("Open input-file: %s error!\n", line);
            return 0;
       }
       num = 0;
       while (!feof(fp)) {
            fgets(line, 99, fp);
            if (feof(fp)) break;
            num++;
       }
       rewind(fp);

       NUMBER_OF_WORD = num;
       if ((Reserved_Word = (RESERVED_WORD *) calloc(num,
           sizeof(RESERVED_WORD))) == NULL) {
           printf("Allocation Reserved_Word failed\n");
           exit (1);
       }

       num = 0;
       while (!feof(fp)) {
            fgets(line, 99, fp);
            if (feof(fp)) break;
            sscanf(line, "%s%s%s", buf1, buf2, buf3);
            strcpy(Reserved_Word[num].old_word, buf1);
            strcpy(Reserved_Word[num].new_word, buf2);
            Reserved_Word[num].len = atoi(buf3);
            num++;
       }
       fclose(fp);
       return (num);
}

static void change_reserved_word(char *string)
{
       int  i, j, found, len;
       char *p;

       for (i = 0; i < Num_Reserved_Word; i++) {
            p = string;
            while ((p = strstr(p, Reserved_Word[i].old_word))) {
                 if (p == string &&
                     (*(p + Reserved_Word[i].len) < 'a' ||
                      *(p + Reserved_Word[i].len) > 'z') &&
                     (*(p + Reserved_Word[i].len) < 'A' ||
                      *(p + Reserved_Word[i].len) > 'Z')) {
                      for (j = 0; j < Reserved_Word[i].len; j++)
                          *(p + j) = Reserved_Word[i].new_word[j];
                 } else if (p != string && (*(p-1) < 'a' || *(p-1) > 'z') &&
                          (*(p-1) < 'A' || *(p-1) > 'Z') &&
                       (*(p + Reserved_Word[i].len) < 'a' ||
                       *(p + Reserved_Word[i].len) > 'z') &&
                       (*(p + Reserved_Word[i].len) < 'A' ||
                        *(p + Reserved_Word[i].len) > 'z')) {
                      if (!strcmp(Reserved_Word[i].old_word, "A") && (*(p-1) != ' '
                         || *(p + Reserved_Word[i].len) != ' ')) {
                           p += Reserved_Word[i].len;
                           continue;
                      }
                      for (j = 0; j < Reserved_Word[i].len; j++)
                           *(p + j) = Reserved_Word[i].new_word[j];
                 }
                 p += Reserved_Word[i].len;
            }
       }

       found = FALSE;
       len = strlen(string);
       for (i = 0; i < len; i++) {
           if (string[i] == ':' || string[i] == '.') found = TRUE;
           if (found && isalpha(string[i])) {
               found = FALSE;
               string[i] = toupper(string[i]);
           }
       }

       p = strstr(string, " a Resolution");
       if (p && isdigit(*(p-1))) {
            p++;
            *p = 'A';
       }

       p = strstr(string, "Ribonuclease a");
       if (p) {
           p += 13;
           *p = 'A';
       }
}

#ifdef SUN_OS
#ifdef __cplusplus
extern "C" {
#endif
#endif
static int corres_compare(const void *coores1, const void *coores2)
{
       return (strcmp(((NAME_CORR *) coores1)->real_name,
                      ((NAME_CORR *) coores2)->real_name));
}

static int residue_name_compare(const void *group1, const void *group2)
{
       return (strcmp(((_OBS_SPR *) group1)->obsolete_residue,
                      ((_OBS_SPR *) group2)->obsolete_residue));
}

static int atom_name_compare(const void *group1, const void *group2)
{
       return (strcasecmp(((_METAL_X_LIST *) group1)->name,
                          ((_METAL_X_LIST *) group2)->name));
}
#ifdef SUN_OS
#ifdef __cplusplus
}
#endif
#endif

