#include <stdio.h>
#include <string.h>
#include <math.h>

#include "_fileobj_ioutil.h"
#include "_maxit.h"
#include "assembly_parser_util.h"
#include "bond_angle_torsion.h"
#include "erf_util.h"
#include "util.h"

#define NUM_MER  20

const static char *_num_mers[NUM_MER+1] = {
       "UNKNOWN",
       "MONOMER",
       "DIMER",
       "TRIMER",
       "TETRAMER",
       "PENTAMER",
       "HEXAMER",
       "HEPTAMER",
       "OCTAMER",
       "NONAMER",
       "DECAMER",
       "UNDECAMER",
       "DODECAMER",
       "TRIDECAMER",
       "TETRADECAMER",
       "PENTADECAMER",
       "HEXADECAMER",
       "HEPTADECAMER",
       "OCTADECAMER",
       "NONADECAMER",
       "EICOSAMER"
};

typedef struct {
       CifString inst_ids;
       ReVarCifArray<CifString> oplist;
} _CHAIN_OPLIST;

typedef struct {
       CifString assembly_id;
       CifString type;
       CifString software;
       CifString ssa;
       CifString absa;
       int total_chains;
       ReVarCifArray<_CHAIN_OPLIST> coplists;
} _ASSEMBLY;

typedef struct {
       CifString asym_id;
       CifString pdb_chnid;
       int       is_polymer;
} _ASYM_CHAINID_CORR;

typedef struct {
       CifString sym;
       CifString oper_id;
       CifString catchainids;
       ReVarCifArray<CifString> chainids;
} _SYM_CHAIN_CORR;
/*
typedef struct {
       CifString PDB_chain_ID;
       ReVarCifArray<CifString> Asym_IDs;
} _PDB_ASYM_CORR;

static void get_pdb_asym_correspondence(CifFileObj *fobj, const char *blockId,
                ReVarCifArray<_PDB_ASYM_CORR>& chainids);
*/
void _MAXIT::ndb_to_pdb_get_remark_475_and_480()
{
       if (_experiment_type != EXPERIMENT_TYPE_XRAY) return;

       ReVarCifArray<ReVarCifArray<CifString> > residue_lists, atom_lists;
       ReVarCifArray<CifString> list, atom_list;
       residue_lists.Clear();
       atom_lists.Clear();
       int Mol_ID = -1;
       for (int i = 0; i < _num_mol; i++) {
            if (_num_mol > 1) Mol_ID = _molecules[i]->Mol_ID();
            _CHAIN *chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() == ATOMN_TOKEN ||
                     chain->chain_type() == ATOMP_TOKEN) {
                      _RESIDUE *residue = chain->GetFirstResidue();
                      while (residue) {
                           int ret = residue->is_zero_occupancy_residue(Mol_ID, atom_list);
                           if (ret) {
                                list.Clear();
                                for (int j = 0; j < 5; j++) list.Add(atom_list[j]);
                                 residue_lists.Add(list);
                           } else if (atom_list.Length() > 0) {
                                atom_lists.Add(atom_list);
                           }
                           residue = chain->GetNextResidue();
                      }
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }

       One_Card_Info_Format *newCard = NULL;
       char mol_id[4], chain_id[2], insert_code[2], buffer[80];
       if (residue_lists.Length() > 0) {
            ndb_to_pdb_get_general_remark(Num_Remark_475, Remark_475, 0, 0, 0, 0);
            for (unsigned int i = 0; i < residue_lists.Length(); i++) {
                 if (residue_lists[i][0] != "-1")
                      strcpy(mol_id, residue_lists[i][0].Text());
                 else strcpy(mol_id, "   ");
                 if (residue_lists[i][1] != "")
                      strcpy(chain_id, residue_lists[i][1].Text());
                 else strcpy(chain_id, " ");
                 if (residue_lists[i][4] != "")
                      strcpy(insert_code, residue_lists[i][4].Text());
                 else strcpy(insert_code, " ");
                 sprintf(buffer, "%3s %3s %c%5s%c", mol_id, residue_lists[i][2].Text(),
                           chain_id[0], residue_lists[i][3].Text(), insert_code[0]);
                 newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(475, 1);
                 newCard->AddElementToField(buffer, 2);
            }
       }

       if (atom_lists.Length() > 0) {
            char head[80];
            ndb_to_pdb_get_general_remark(Num_Remark_480, Remark_480, 0, 0, 0, 0);
            for (unsigned int i = 0; i < atom_lists.Length(); i++) {
                 if (atom_lists[i][0] != "-1")
                      strcpy(mol_id, atom_lists[i][0].Text());
                 else strcpy(mol_id, "   ");
                 if (atom_lists[i][1] != "")
                      strcpy(chain_id, atom_lists[i][1].Text());
                 else strcpy(chain_id, " ");
                 if (atom_lists[i][4] != "")
                      strcpy(insert_code, atom_lists[i][4].Text());
                 else strcpy(insert_code, " ");
                 newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(480, 1);
                 sprintf(head, "%3s %3s %c %4s%c  ", mol_id, atom_lists[i][2].Text(),
                           chain_id[0], atom_lists[i][3].Text(), insert_code[0]);
                 newCard->AddElementToField(head, 2);
                 int num = 0;
                 for (unsigned int j = 5; j < atom_lists[i].Length(); j++) {
                      if (num == 7) {
                           newCard = _CardInfo[REMARK_TOKEN-1].add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                           newCard->AddElementToField(480, 1);
                           sprintf(head, "%3s %3s %c %4s%c  ", mol_id, atom_lists[i][2].Text(),
                                     chain_id[0], atom_lists[i][3].Text(), insert_code[0]);
                           newCard->AddElementToField(head, 2);
                           num = 0;
                      }
                      sprintf(buffer, "%-6s", atom_lists[i][j].Text());
                      newCard->AttachToField(buffer, 2);
                      num++;
                 }
            }
       }
}

void _MAXIT::ndb_to_pdb_get_remark_610()
{
       ReVarCifArray<ReVarCifArray<CifString> > residue_lists;
       ReVarCifArray<CifString> list;
       residue_lists.Clear();
       int Mol_ID = -1;
       char buffer[80];
       CifString cs;

       finding_missing_and_extra_atoms(0);

       for (int i = 0; i < _num_mol; i++) {
            if (_num_mol > 1) Mol_ID = _molecules[i]->Mol_ID();
            _CHAIN *chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMN_TOKEN &&
                     chain->chain_type() != ATOMP_TOKEN) {
                      _RESIDUE *residue = chain->GetFirstResidue();
                      while (residue) {
                           if (strcmp(residue->missing().Text(), "")) {
                                _ATOM *atom = residue->GetFirstAtom();
                                list.Clear();
                                sprintf(buffer, "%d", Mol_ID);
                                cs.Copy(buffer);
                                list.Add(cs);
                                cs.Copy(atom->pdb_chnid());
                                list.Add(cs);
                                cs.Copy(atom->pdb_resnam());
                                list.Add(cs);
                                sprintf(buffer, "%d", atom->pdb_resnum());
                                cs.Copy(buffer);
                                list.Add(cs);
                                cs.Copy(atom->ins_code());
                                list.Add(cs);
                                residue_lists.Add(list);
                           }
                           residue = chain->GetNextResidue();
                      }
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }

       if (residue_lists.Length() == 0) return;

       ndb_to_pdb_get_general_remark(Num_Remark_610, Remark_610, 0, 0, 0, 0);

       char mol_id[4], chain_id[2], insert_code[2];
       for (unsigned int i2 = 0; i2 < residue_lists.Length(); i2++) {
            if (residue_lists[i2][0] != "-1")
                 strcpy(mol_id, residue_lists[i2][0].Text());
            else strcpy(mol_id, "   ");
            if (residue_lists[i2][1] != "")
                 strcpy(chain_id, residue_lists[i2][1].Text());
            else strcpy(chain_id, " ");
            if (residue_lists[i2][4] != "")
                 strcpy(insert_code, residue_lists[i2][4].Text());
            else strcpy(insert_code, " ");
            sprintf(buffer, "%3s %3s %c%5s%c", mol_id, residue_lists[i2][2].Text(),
                      chain_id[0], residue_lists[i2][3].Text(), insert_code[0]);
            One_Card_Info_Format *newCard = _CardInfo[REMARK_TOKEN-1].
                      add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(610, 1);
            newCard->AddElementToField(buffer, 2);
       }
}

void _MAXIT::ndb_to_pdb_get_remark_615()
{
       if (_experiment_type != EXPERIMENT_TYPE_XRAY) return;

       ReVarCifArray<ReVarCifArray<CifString> > residue_lists;
       ReVarCifArray<CifString> list, atom_list;
       residue_lists.Clear();
       int Mol_ID = -1;
       for (int i = 0; i < _num_mol; i++) {
            if (_num_mol > 1) Mol_ID = _molecules[i]->Mol_ID();
            _CHAIN *chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMN_TOKEN &&
                     chain->chain_type() != ATOMP_TOKEN) {
                      _RESIDUE *residue = chain->GetFirstResidue();
                      while (residue) {
                           int ret = residue->is_zero_occupancy_residue(Mol_ID, atom_list);
                           if (ret || atom_list.Length() > 0)  {
                                list.Clear();
                                for (int j = 0; j < 5; j++) list.Add(atom_list[j]);
                                 residue_lists.Add(list);
                           }
                           residue = chain->GetNextResidue();
                      }
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }

       if (residue_lists.Length() == 0) return;

       ndb_to_pdb_get_general_remark(Num_Remark_615, Remark_615, 0, 0, 0, 0);

       char mol_id[4], chain_id[2], insert_code[2], buffer[80];
       for (unsigned int i2 = 0; i2 < residue_lists.Length(); i2++) {
            if (residue_lists[i2][0] != "-1")
                 strcpy(mol_id, residue_lists[i2][0].Text());
            else strcpy(mol_id, "   ");
            if (residue_lists[i2][1] != "")
                 strcpy(chain_id, residue_lists[i2][1].Text());
            else strcpy(chain_id, " ");
            if (residue_lists[i2][4] != "")
                 strcpy(insert_code, residue_lists[i2][4].Text());
            else strcpy(insert_code, " ");
            sprintf(buffer, "%3s %3s %c%5s%c", mol_id, residue_lists[i2][2].Text(),
                      chain_id[0], residue_lists[i2][3].Text(), insert_code[0]);
             One_Card_Info_Format *newCard = _CardInfo[REMARK_TOKEN-1].
                  add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(615, 1);
            newCard->AddElementToField(buffer, 2);
       }
}

int _MAXIT::get_atom_pair(_ATOM &metal_atom, _ATOM &polar_atom, _LINK *link)
{
       if (_num_mol == 0 || _molecules == NULL) return 0;

       _CHAIN *ch1 = _molecules[link->mol_index]->GetIndexChain(link->FstIndex);
       _CHAIN *ch2 = _molecules[link->mol_index]->GetIndexChain(link->SndIndex);
       if (!ch1 || !ch2) return 0;
       _FIELD *Init = ch1->SeqResIndex(link->FstRes);
       _FIELD *End  = ch2->SeqResIndex(link->SndRes);
       if (!Init || !End) return 0;
       _RESIDUE *beg = Init->Residue;
       _RESIDUE *end = End->Residue;
       if (!beg || !end) return 0;;

       int found = 0, num_a, num_b;
       _ATOM **atom_a = beg->find_atom(link->FstAtom, num_a);
       _ATOM **atom_b = end->find_atom(link->SndAtom, num_b);
       if (atom_a && atom_b) {
            _ATOM *atom1 = NULL;
            _ATOM *atom2 = NULL;
            for (int i = 0; i < num_a; i++) {
                 if (!strcmp(atom_a[i]->alt_loc(), link->altLoc1) &&
                     !strcmp(atom_a[i]->restype(), link->FstName)) {
                      atom1 = atom_a[i];
                      break;
                 }
            }
            for (int i2 = 0; i2 < num_b; i2++) {
                 if (!strcmp(atom_b[i2]->alt_loc(), link->altLoc2) &&
                     !strcmp(atom_b[i2]->restype(), link->SndName)) {
                      atom2 = atom_b[i2];
                      break;
                 }
            }
            if (atom1 && atom2) {
                 found = 1;
                 COORD coord = atom2->orig();
                 if (strcmp(link->SymOP_2, "") && strcmp(link->SymOP_2, "1_555")) {
                      int op = atoi(link->SymOP_2) - 1;
                      int t[3], j;
                      int len = strlen(link->SymOP_2);
                      t[0] = t[1] = t[2] = 0;
                      for (j = 0; j < len; j++) {
                           if (link->SymOP_2[j] == '_') break;
                      }
                      if (j + 3 < len) {
                           for (int k = 0; k < 3; k++) {
                                t[k] = link->SymOP_2[j + k + 1] - '5';
                           }
                      }
                      symmetry_operation(coord, _cell, op, t[0], t[1], t[2]);
                 }
                 if (find_metal_flag(link->FstAtom) == 2) {
                      metal_atom = *atom1;
                      polar_atom = *atom2;
                      polar_atom.set_orig(coord);
                 } else {
                      metal_atom = *atom2;
                      metal_atom.set_orig(coord);
                      polar_atom = *atom1;
                 }
            }
       }
       if (atom_a) delete [] atom_a;
       if (atom_b) delete [] atom_b;

       return found;
}

void _MAXIT::ndb_to_pdb_get_remark_620(ReVarCifArray<_METAL_COORDINATION> &MetalList)
{
       if (MetalList.Length() == 0) return;

       ndb_to_pdb_get_general_remark(Num_Remark_620, Remark_620, 0, 0, 0, 0);

       char buffer[100], buffer1[100], chain_id, insert_code, name[5];
       One_Card_Info_Format *pCard = NULL;
       for (unsigned int i = 0; i < MetalList.Length(); i++) {
            pCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            pCard->AddElementToField(620, 1);
            pCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            pCard->AddElementToField(620, 1);
            pCard->AddElementToField("COORDINATION ANGLES FOR:  M RES CSSEQI METAL", 2);

            if (strcmp(MetalList[i].metal_atom.pdb_chnid(), ""))
                 chain_id = MetalList[i].metal_atom.pdb_chnid()[0];
            else chain_id = ' ';
            if (strcmp(MetalList[i].metal_atom.ins_code(), ""))
                 insert_code = MetalList[i].metal_atom.ins_code()[0];
            else insert_code = ' ';
            get_print_atom(MetalList[i].metal_atom.pdb_resnam(),
                       MetalList[i].metal_atom.pdb_atmnam(), name);
            if (MetalList[i].Mol_ID > 0)
                 sprintf(buffer, "                        %3d %3s %c%4d%c %4s",  
                        MetalList[i].Mol_ID, MetalList[i].metal_atom.pdb_resnam(), chain_id,
                        MetalList[i].metal_atom.pdb_resnum(), insert_code, name);
            else sprintf(buffer, "                            %3s %c%4d%c %4s",
                        MetalList[i].metal_atom.pdb_resnam(), chain_id,
                        MetalList[i].metal_atom.pdb_resnum(), insert_code, name);
            pCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            pCard->AddElementToField(620, 1);
            pCard->AddElementToField(buffer, 2);
            pCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            pCard->AddElementToField(620, 1);
            pCard->AddElementToField("N RES CSSEQI ATOM", 2);
            for (unsigned int j = 0; j < MetalList[i].polar_atoms.Length(); j++) {
                 if (strcmp(MetalList[i].polar_atoms[j].pdb_chnid(), ""))
                      chain_id = MetalList[i].polar_atoms[j].pdb_chnid()[0];
                 else chain_id = ' ';
                 if (strcmp(MetalList[i].polar_atoms[j].ins_code(), ""))
                      insert_code = MetalList[i].polar_atoms[j].ins_code()[0];
                 else insert_code = ' ';
                 get_print_atom(MetalList[i].polar_atoms[j].pdb_resnam(),
                         MetalList[i].polar_atoms[j].pdb_atmnam(), name);
                 sprintf(buffer, "%d %3s %c%4d%c %4s", j + 1,
                          MetalList[i].polar_atoms[j].pdb_resnam(), chain_id,
                          MetalList[i].polar_atoms[j].pdb_resnum(), insert_code, name);
                 strcpy(buffer1, "N                 ");
                 for (unsigned int k = 0; k < j; k++) {
                      double angle = cal_angle(MetalList[i].polar_atoms[k],
                                          MetalList[i].metal_atom,
                                          MetalList[i].polar_atoms[j]);
                      int len = strlen(buffer);
                      sprintf(&buffer[len], "%6.1f", angle);
                      len = strlen(buffer1);
                      sprintf(&buffer1[len], "   %d  ", k + 1);
                 }
                 pCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 pCard->AddElementToField(620, 1);
                 pCard->AddElementToField(buffer, 2);
            }
            pCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            pCard->AddElementToField(620, 1);
            pCard->AddElementToField(buffer1, 2);
       }
}

void _MAXIT::find_secondary_citations()
{
       if (_JrnlCardInfo == NULL) return;

       int vField = 5;
       int jrnl_id = find_jrnl_id("REF");
       One_Card_Info_Format *pJrnlCard = _JrnlCardInfo[jrnl_id].Head();
       while (pJrnlCard) {
            if (strcmp(pJrnlCard->Field(vField+1), "") &&
               !strcmp(pJrnlCard->Field(vField), ""))
                 pJrnlCard->AddElementToField("V.", vField);
            else if (!strcmp(pJrnlCard->Field(vField+1), ""))
                 pJrnlCard->AddElementToField("", vField);
            pJrnlCard = pJrnlCard->Next();
       }

       int astmField = 3;
       int issnField = 6;
       jrnl_id = find_jrnl_id("REFN");
       pJrnlCard = _JrnlCardInfo[jrnl_id].Head();
       while (pJrnlCard) {
            if (strcmp(pJrnlCard->Field(astmField+1), "") &&
               !strcmp(pJrnlCard->Field(astmField), ""))
                 pJrnlCard->AddElementToField("ASTM", astmField);
            else if (!strcmp(pJrnlCard->Field(astmField+1), ""))
                 pJrnlCard->AddElementToField("", astmField);

            if (strcmp(pJrnlCard->Field(issnField+1), "") &&
               !strcmp(pJrnlCard->Field(issnField), ""))
                 pJrnlCard->AddElementToField("ISSN", issnField);
            else if (!strcmp(pJrnlCard->Field(issnField+1), ""))
                 pJrnlCard->AddElementToField("", issnField);
            pJrnlCard = pJrnlCard->Next();
       }

       int maxJrnlSerialNo = 0;
       for (jrnl_id = 0; jrnl_id < MxJrnlTokens; jrnl_id++) {
            pJrnlCard = _JrnlCardInfo[jrnl_id].Head();
            while (pJrnlCard) {
                 if (atoi(pJrnlCard->Field(1)) > maxJrnlSerialNo)
                      maxJrnlSerialNo = atoi(pJrnlCard->Field(1));
                 pJrnlCard = pJrnlCard->Next();
            }
       }

       char temp_line[MxLen], FieldFormat[MxStrLen];
       One_Card_Info_Format *newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
       newCard->AddElementToField(1, 1);
       for (int JrnlSerialNo = 1; JrnlSerialNo <= maxJrnlSerialNo; JrnlSerialNo++) {
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(1, 1);
            sprintf(temp_line, "REFERENCE %d", JrnlSerialNo);
            newCard->AddElementToField(temp_line, 2);
            for (int displayOrder = 0; displayOrder < MxJrnlTokens; displayOrder++) {
                 for (jrnl_id = 0; jrnl_id < MxJrnlTokens; jrnl_id++) {
                      if (JrnlToken[jrnl_id].DisplayOrder == displayOrder) break;
                 }
                 if (jrnl_id == MxJrnlTokens) continue;

                 pJrnlCard = _JrnlCardInfo[jrnl_id].Head();
                 while (pJrnlCard) {
                      if (atoi(pJrnlCard->Field(1)) == JrnlSerialNo) break;
                      pJrnlCard = pJrnlCard->Next();
                 }

                 if (pJrnlCard == NULL) continue;

                 int ContNo = 1;
                 int start = 0;
                 while (1) {
                      int pos = 1;
                      strcpy(temp_line, " ");
                      for (int field_no = 2; field_no < JrnlToken[jrnl_id].NumField; field_no++) {
                           if (JrnlToken[jrnl_id].FieldList[field_no].NdbOnly) continue;
                           pJrnlCard->UpperCaseField(field_no);
                           if (field_no > 2) {
                                int width = JrnlToken[jrnl_id].FieldList[field_no].FieldStcol
                                          - JrnlToken[jrnl_id].FieldList[field_no-1].FieldStcol
                                          - JrnlToken[jrnl_id].FieldList[field_no-1].FieldWidth;
                                sprintf(FieldFormat, "%%-%ds", width);
                                sprintf(&temp_line[pos], FieldFormat, "");
                                pos += width;
                           }
                           get_field_format(jrnl_id, field_no, FieldFormat, 1, NDB_FILE_FORMAT_PDB);
                           if (JrnlToken[jrnl_id].ConField == field_no + 1) {
                                if (ContNo != 1)
                                     sprintf(&temp_line[pos], FieldFormat, ContNo);
                                else {
                                     sprintf(FieldFormat, "%%%ds",
                                            JrnlToken[jrnl_id].FieldList[field_no].FieldWidth);
                                     sprintf(&temp_line[pos], FieldFormat, "");
                                }                      
                           } else if (JrnlToken[jrnl_id].ContinuedF == field_no + 1)
                                print_jrnl_continuedf_field(jrnl_id, pJrnlCard, pos, start,
                                                     temp_line, FieldFormat);
                           else if (strcmp(pJrnlCard->Field(field_no), "") == 0 ||
                                   ContNo != 1 && field_no != 2) {
                                sprintf(FieldFormat, "%%%ds",
                                       JrnlToken[jrnl_id].FieldList[field_no].FieldWidth);
                                sprintf(&temp_line[pos], FieldFormat, "");
                           } else if (JrnlToken[jrnl_id].FieldList[field_no].FieldType <= 2) {
                                double temp_double = atof(pJrnlCard->Field(field_no));
                                sprintf(&temp_line[pos], FieldFormat, (int) temp_double);
                           } else {
                                sprintf(&temp_line[pos], FieldFormat, pJrnlCard->Field(field_no));
                           }
                           pos += JrnlToken[jrnl_id].FieldList[field_no].FieldWidth;
                      }

                      newCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(1, 1);
                      newCard->AddElementToField(temp_line, 2);

                      if (JrnlToken[jrnl_id].ContinuedF != 0 && start <
                              (int) strlen(pJrnlCard->Field(JrnlToken[jrnl_id].ContinuedF-1)))
                           ContNo++;
                      else break;
                 }
            }
       }

       for (int i = 0; i < MxJrnlTokens; i++) {
            pJrnlCard = _JrnlCardInfo[i].Head();
            while (pJrnlCard) {
                 if (atoi(pJrnlCard->Field(1)) > 0)
                      pJrnlCard = _JrnlCardInfo[i].delete_a_card(pJrnlCard);
                 else pJrnlCard = pJrnlCard->Next();
            }
       }
}

int _MAXIT::generate_biological_assembly()
{
       if (!_CifObj) return 0;

       if (is_empty_table(_CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly")) ||
           is_empty_table(_CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly_gen")) ||
           is_empty_table(_CifObj->GetTablePtr(_StructureId, "pdbx_struct_oper_list"))) {
            return 0;
       }

       ISTable *t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly");

       CifString cs1, cs2, cs3;
       ReVarCifArray<_ASSEMBLY> assembly_lists;
       _ASSEMBLY assembly;
       assembly_lists.Clear();
       int rowNo = t->GetNumRows();
       int idx1 = t->GetColumnIndex("id");
       int idx2 = t->GetColumnIndex("details");
       int idx3 = t->GetColumnIndex("method_details");
       for (int i2 = 0; i2 < rowNo; i2++) {
            get_value(cs2, t, idx2, i2);
            cs2.ToLower();
            if (cs2 == "representative helical assembly" ||
                cs2 == "complete icosahedral assembly" ||
                cs2 == "complete point assembly" ||
                cs2 == "author_and_software_defined_assembly" ||
                cs2 == "author_defined_assembly" ||
                cs2 == "software_defined_assembly") {
                 assembly.type.Copy(cs2);
                 get_value(cs1, t, idx1, i2);
                 assembly.assembly_id.Copy(cs1);
                 get_value(cs3, t, idx3, i2);
                 assembly.software.Copy(cs3);
                 assembly.ssa.Clear();
                 assembly.absa.Clear();  
                 assembly.coplists.Clear();
                 assembly_lists.Add(assembly);
            }
       }
       if (assembly_lists.Length() == 0) return 0;

       t = _CifObj->GetTablePtr(_StructureId, "struct_asym");

       ReVarCifArray<_ASYM_CHAINID_CORR> asym_chainid_corrs;
       asym_chainid_corrs.Clear();
       if (!is_empty_table(t)) {
            rowNo = t->GetNumRows();
            idx1 = t->GetColumnIndex("id");
            idx2 = t->GetColumnIndex("ndb_pdb_id");
            idx3 = t->GetColumnIndex("ndb_type");
            for (int i = 0; i < rowNo; i++) {
                 get_value(cs1, t, idx1, i);
                 get_value(cs2, t, idx2, i);
                 get_value(cs3, t, idx3, i);
                 if (cs1 != "" && cs2 != "") {
                      _ASYM_CHAINID_CORR coor;
                      coor.asym_id.Copy(cs1);
                      coor.pdb_chnid.Copy(cs2);
                      if (cs3 == "ATOMP" || cs3 == "ATOMN")
                           coor.is_polymer = 1;
                      else coor.is_polymer = 0;
                      asym_chainid_corrs.Add(coor);
                 }
            }
       }

       t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly_gen");
       rowNo = t->GetNumRows();
       idx1 = t->GetColumnIndex("assembly_id");
       idx2 = t->GetColumnIndex("asym_id_list");
       idx3 = t->GetColumnIndex("oper_expression");
       _CHAIN_OPLIST coplist;
       char *buff = NULL, *p = NULL, buffer[80];
       int len;
       TblIndexObj ChainIDIndex, PolymerChainIDIndex;
       for (unsigned int i = 0; i < assembly_lists.Length(); i++) {
            assembly_lists[i].total_chains = 0;
            for (int j = 0; j < rowNo; j++) {
                 get_value(cs1, t, idx1, j);
                 get_value(cs2, t, idx2, j);
                 get_value(cs3, t, idx3, j);
                 if (cs1 == assembly_lists[i].assembly_id) {
                      len = strlen(cs2.Text());
                      buff = new char[len + 1];
                      memset(buff, 0, len + 1);
                      strcpy(buff, cs2.Text());
                      for (int k = 0; k < len; k++) {
                           if (buff[k] == ',' || buff[k] == ';') buff[k] = ' ';
                      }
                      ndb_clean_string(buff);
                      p = buff;
                      coplist.inst_ids.Clear();
                      len = 0;
                      ChainIDIndex.Clear();
                      PolymerChainIDIndex.Clear();
                      while (p) {
                           p = get_word(p, buffer);
                           if (!strcmp(buffer, "")) continue;
                           cs2.Copy(buffer);
                           int is_polymer = 0;
                           for (unsigned int k = 0; k < asym_chainid_corrs.Length(); k++) {
                                if (asym_chainid_corrs[k].asym_id == buffer) {
                                     cs2.Copy(asym_chainid_corrs[k].pdb_chnid);
                                     if (asym_chainid_corrs[k].is_polymer) is_polymer = 1;
                                     break;
                                }
                           }
                           if (is_polymer) {
                                if (PolymerChainIDIndex.Seek(cs2) < 0) {
                                     PolymerChainIDIndex.Add(cs2);
                                     len++;
                                }
                           }
                           if (ChainIDIndex.Seek(cs2) < 0) {
                                ChainIDIndex.Add(cs2);
                                if (coplist.inst_ids != "") coplist.inst_ids += ", ";
                                coplist.inst_ids += cs2;
                           }
                      }
                      delete [] buff;
    
                      coplist.oplist.Clear();
                      parseString(cs3, coplist.oplist);
                      if (coplist.oplist.Length() > 0) {
                           assembly_lists[i].coplists.Add(coplist);
                      }
                      assembly_lists[i].total_chains += len * coplist.oplist.Length();
                 }
            }
       }

       t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly_prop");
       if (!is_empty_table(t)) {
            rowNo = t->GetNumRows();
            idx1 = t->GetColumnIndex("biol_id");
            idx2 = t->GetColumnIndex("type");
            idx3 = t->GetColumnIndex("value");
            for (unsigned int i = 0; i < assembly_lists.Length(); i++) {
                 for (int j = 0; j < rowNo; j++) {
                      get_value(cs1, t, idx1, j);
                      get_value(cs2, t, idx2, j);
                      get_value(cs3, t, idx3, j);
                      if (cs1 == assembly_lists[i].assembly_id) {
                           if (cs2 == "SSA (A^2)")
                                assembly_lists[i].ssa.Copy(cs3);
                           else if (cs2 == "ABSA (A^2)")
                                assembly_lists[i].absa.Copy(cs3);
                      }
                 }
            }
       }

       t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_oper_list");

       ndb_to_pdb_get_general_remark(Num_Remark_350 - 1, Remark_350, 0, 0, 0, 0);

       One_Card_Info_Format *newCard = NULL;
       ReVarCifArray<CifString> id_list;
       double mat[4][4];
       char FieldFormat[82];
       for (unsigned int l = 0; l < assembly_lists.Length(); l++) {
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(350, 1);
            newCard->AddElementToField("", 2);
            newCard = _CardInfo[REMARK_TOKEN-1].
                       add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
            newCard->AddElementToField(350, 1);
            sprintf(buffer, "BIOMOLECULE: %d", l + 1);
            newCard->AddElementToField(buffer, 2);

            if (assembly_lists[l].total_chains <= NUM_MER)
                 strcpy(FieldFormat, _num_mers[assembly_lists[l].total_chains]);
            else sprintf(FieldFormat, "%d-MER", assembly_lists[l].total_chains);
            if (assembly_lists[l].type == "author_and_software_defined_assembly" ||
                assembly_lists[l].type == "author_defined_assembly") {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(350, 1);
                 sprintf(buffer, "AUTHOR DETERMINED BIOLOGICAL UNIT: %s", FieldFormat);
                 newCard->AddElementToField(buffer, 2);
            }
            if (assembly_lists[l].type == "author_and_software_defined_assembly" ||
                assembly_lists[l].type == "software_defined_assembly") {
                 newCard = _CardInfo[REMARK_TOKEN-1].
                            add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                 newCard->AddElementToField(350, 1);
                 sprintf(buffer, "SOFTWARE DETERMINED QUATERNARY STRUCTURE: %s", FieldFormat);
                 newCard->AddElementToField(buffer, 2);
                 if (assembly_lists[l].software == "") assembly_lists[l].software.Copy("PISA");
                 if (assembly_lists[l].software != "") {
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(350, 1);
                      sprintf(buffer, "SOFTWARE USED: %s", assembly_lists[l].software.Text());
                      newCard->AddElementToField(buffer, 2);
                 }
                 if (assembly_lists[l].absa != "") {
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(350, 1);
                      sprintf(buffer, "AVERAGE BURIED SURFACE AREA: %s ANGSTROM**2",
                                         assembly_lists[l].absa.Text());
                      newCard->AddElementToField(buffer, 2);
                 }
            }

            int serialNo = 0;
            for (unsigned int m = 0; m < assembly_lists[l].coplists.Length(); m++) {
                 cs2.Copy("APPLY THE FOLLOWING TO CHAINS: ");
                 cs2 += assembly_lists[l].coplists[m].inst_ids;
                 len = strlen(cs2.Text());
                 buff = new char[len + 1];
                 memset(buff, 0, len + 1);
                 strcpy(buff, cs2.Text());
                 p = buff;
                 while (p) {
                      p = get_max_length_word(p, buffer, 60);
                      newCard = _CardInfo[REMARK_TOKEN-1].
                                 add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                      newCard->AddElementToField(350, 1);
                      newCard->AddElementToField(buffer, 2);
                 }
                 delete [] buff;

                 for (unsigned int i = 0; i < assembly_lists[l].coplists[m].oplist.Length(); i++) {
                      get_multi_matrix_list(assembly_lists[l].coplists[m].oplist[i].Text(), id_list);
                      if (get_final_matrix(mat, id_list, t)) {
                           serialNo++;
                           for (int k = 0; k < 3; k++) {
                                newCard = _CardInfo[REMARK_TOKEN-1].
                                         add_a_new_card(NdbToken[REMARK_TOKEN-1], 0);
                                newCard->AddElementToField(Remark_SYMMA.Remark_No, 1);
                                for (int j = 0; j < Remark_SYMMA.NumField; j++) {
                                     print_space_between_remark_field(newCard, &Remark_SYMMA, j);
                                     get_remark_field_format(FieldFormat, &Remark_SYMMA, j);
                                     int field_no = Remark_SYMMA.FieldList[j].FieldId-1;
                                     if (!strcmp("BIOMT", Remark_SYMMA.FieldList[j].Text))
                                          sprintf(buffer, "%s%d",
                                              Remark_SYMMA.FieldList[j].Text, k + 1);
                                     else if (field_no == 1)
                                          sprintf(buffer, FieldFormat, (double) serialNo);
                                     else if (j != Remark_SYMMA.NumField - 1)
                                          sprintf(buffer, FieldFormat, mat[k][j-2]);
                                     else sprintf(buffer, FieldFormat, mat[k][3]);
                                     newCard->AttachToField(buffer, 2);
                                }
                           }
                      }
                 }
            }
       }

       return 1;
}

void _MAXIT::recover_assembly_info()
{
       if (__assembly_array.Length() == 0) return;

       ReVarCifArray<_PDB_ASYM_CORR> chainids;
       get_pdb_asym_correspondence(_CifObj, _StructureId, chainids);

       ISTable *Table = new ISTable("pdbx_struct_assembly");
       Table->AddColumn("id");
       Table->AddColumn("details");
       Table->AddColumn("method_details");
       ISTable *t1 = new ISTable("pdbx_struct_assembly_gen");
       t1->AddColumn("assembly_id");
       t1->AddColumn("oper_expression");
       t1->AddColumn("asym_id_list");
       ISTable *t2 = new ISTable("pdbx_struct_assembly_prop");
       t2->AddColumn("biol_id");
       t2->AddColumn("type");
       t2->AddColumn("value");
       t2->AddColumn("details");
       ISTable *t3 = new ISTable("pdbx_struct_oper_list");
       t3->AddColumn("id");
       t3->AddColumn("type");
       t3->AddColumn("name");
       t3->AddColumn("matrix[1][1]");
       t3->AddColumn("matrix[1][2]");
       t3->AddColumn("matrix[1][3]");
       t3->AddColumn("vector[1]");
       t3->AddColumn("matrix[2][1]");
       t3->AddColumn("matrix[2][2]");
       t3->AddColumn("matrix[2][3]");
       t3->AddColumn("vector[2]");
       t3->AddColumn("matrix[3][1]");
       t3->AddColumn("matrix[3][2]");
       t3->AddColumn("matrix[3][3]");
       t3->AddColumn("vector[3]");
       CifString cifstring, assembly_id, sym, oper_id;
       int row, len, k;
       char *value = NULL;
       ReVarCifArray<_SYM_CHAIN_CORR> sym_chain_corrs;
       BTreeObj<_IIndex> Sort_Index;
       _IIndex idx;
       Sort_Index.Clear();
       for (unsigned int i2 = 0; i2 < __assembly_array.Length(); i2++) {
            idx.set_value(atoi(__assembly_array[i2].assembly_id.Text()));
            idx.set_pos(i2);
            Sort_Index.Add(idx);
       }

       ReVarCifArray<_ASSEMBLY_ARRAY> assembly_array_sort;
       assembly_array_sort.Clear();
       Sort_Index.GoFirst();
       idx = Sort_Index.GetCurrData();
       assembly_array_sort.Add(__assembly_array[idx.pos()]);
       while (Sort_Index.GoToNext() >= 0) {
            idx = Sort_Index.GetCurrData();
            assembly_array_sort.Add(__assembly_array[idx.pos()]);
       }

       char buffer[100], *p = NULL;
       for (unsigned int i = 0; i < assembly_array_sort.Length(); i++) {
            row = Table->GetNumRows();
            Table->AddRow();
            sprintf(buffer, "%d", i + 1);
            assembly_id.Copy(buffer);
            Table->UpdateCell(assembly_id, 0, row);
            Table->UpdateCell(assembly_array_sort[i].method, 1, row);
            Table->UpdateCell(assembly_array_sort[i].software, 2, row);
            if (assembly_array_sort[i].ssa != "") {
                 row = t2->GetNumRows();
                 t2->AddRow();
                 t2->UpdateCell(assembly_id, 0, row);
                 cifstring.Copy("SSA (A^2)");
                 t2->UpdateCell(cifstring, 1, row);
                 t2->UpdateCell(assembly_array_sort[i].ssa, 2, row);
            }
            if (assembly_array_sort[i].absa != "") {
                 row = t2->GetNumRows();
                 t2->AddRow();
                 t2->UpdateCell(assembly_id, 0, row);
                 cifstring.Copy("ABSA (A^2)");
                 t2->UpdateCell(cifstring, 1, row);
                 t2->UpdateCell(assembly_array_sort[i].absa, 2, row);
            }

            sym_chain_corrs.Clear();
            for (unsigned int j = 0; j < assembly_array_sort[i].assembly_chains.Length(); j++) {
                 len = strlen(assembly_array_sort[i].assembly_chains[j].sym_operators.Text())+1;
                 value = new char[len];
                 memset(value, 0, len);
                 strcpy(value, assembly_array_sort[i].assembly_chains[j].sym_operators.Text()); 
                 for (k = 0; k < len - 1; k++) {
                      if (value[k] == ',') value[k] = ' ';
                 }
                 ndb_clean_string(value);
                 p = value;
                 while (p) {
                      p = get_word(p, buffer);
                      if (!strcmp(buffer, "")) continue;
                      row = t3->GetNumRows();
                      for (k = 0; k < row; k++) {
                           t3->GetCell(sym, 2, k);
                           if (sym == buffer) {
                                t3->GetCell(oper_id, 0, k);
                                break;
                           }
                      }
                      if (k == row) {
                           t3->AddRow();
                           sym.Copy(buffer);
                           t3->UpdateCell(sym, 2, row);
                           sprintf(buffer, "%d", row + 1);
                           oper_id.Copy(buffer);
                           t3->UpdateCell(oper_id, 0, row);
                           if (sym == "1_555")
                                cifstring.Copy("identity operation");
                           else cifstring.Copy("crystal symmetry operation");
                           t3->UpdateCell(cifstring, 1, row);
                      }
                      for (k = 0; k < (int) sym_chain_corrs.Length(); k++) {
                           if (sym_chain_corrs[k].sym == sym) {
                                sym_chain_corrs[k].catchainids += assembly_array_sort[i].assembly_chains[j].PDB_chain_ID; 
                                sym_chain_corrs[k].chainids.Add(assembly_array_sort[i].assembly_chains[j].PDB_chain_ID);
                                break;
                           }
                      }
                      if (k == (int) sym_chain_corrs.Length()) {
                           _SYM_CHAIN_CORR sym_corr;
                           sym_corr.sym.Copy(sym);
                           sym_corr.oper_id.Copy(oper_id);
                           sym_corr.catchainids.Copy(assembly_array_sort[i].assembly_chains[j].PDB_chain_ID);
                           sym_corr.chainids.Clear();
                           sym_corr.chainids.Add(assembly_array_sort[i].assembly_chains[j].PDB_chain_ID);
                           sym_chain_corrs.Add(sym_corr);
                      }
                 }
                 delete [] value;
            }
            if (sym_chain_corrs.Length() > 0) {
                 Sort_Index.Clear();
                 for (unsigned int j = 0; j < sym_chain_corrs.Length(); j++) {
                      idx.set_value(sym_chain_corrs[j].chainids.Length());
                      idx.set_pos(j);
                      Sort_Index.Add(idx);
                 }
                 if (Sort_Index.Root() >= 0) {
                      Sort_Index.GoFirst();
                      int *index = new int[sym_chain_corrs.Length()];
                      idx = Sort_Index.GetCurrData();
                      k = 0;
                      index[k++] = idx.pos();
                      while (Sort_Index.GoToNext() >= 0) {
                           idx = Sort_Index.GetCurrData();
                           index[k++] = idx.pos();
                      }
                      for (int f = 0; f < k; f++) {
                           cifstring.Copy(sym_chain_corrs[index[f]].oper_id);
                           int last = f;
                           for (int s = f + 1; s < k; s++) {
                                if (sym_chain_corrs[index[s]].catchainids ==
                                    sym_chain_corrs[index[f]].catchainids) {
                                     cifstring += ",";
                                     cifstring += sym_chain_corrs[index[s]].oper_id;
                                     last = s;
                                }
                           }
                           row = t1->GetNumRows();
                           t1->AddRow();
                           t1->UpdateCell(assembly_id, 0, row);
                           t1->UpdateCell(cifstring, 1, row);
                           cifstring.Clear();
                           for (unsigned int j = 0; j < sym_chain_corrs[index[f]].chainids.Length(); j++) {
                                for (unsigned int l = 0; l < chainids.Length(); l++) {
                                     if (sym_chain_corrs[index[f]].chainids[j] ==
                                         chainids[l].PDB_chain_ID) {
                                          for (unsigned int m = 0; m < chainids[l].Asym_IDs.Length(); m++) {
                                               if (cifstring != "") cifstring += ",";
                                               cifstring += chainids[l].Asym_IDs[m]; 
                                          }
                                          break;
                                     }
                                }
                           }
                           t1->UpdateCell(cifstring, 2, row);
                           f = last;
                      }
                      delete [] index;
                 }
            }
       }

       int rowNo = t3->GetNumRows(); 
       int j, op, lx, ly, lz;
       double matrix[3][3], vector[3];
       for (int i3 = 0; i3 < rowNo; i3++) {
            t3->GetCell(cifstring, 2, i3);
            strcpy(buffer, cifstring.Text());
            op = atoi(buffer) - 1;
            len = strlen(buffer);
            for (j = 0; j < len; j++) {
                 if (buffer[j] == '_') break;
            }
            lx = buffer[j + 1] - '5';
            ly = buffer[j + 2] - '5';
            lz = buffer[j + 3] - '5';
            ndb_get_symmetry_matrix(_cell, op, lx, ly, lz, matrix, vector);
            sprintf(buffer, "%.10f", matrix[0][0]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  3, i3);
            sprintf(buffer, "%.10f", matrix[0][1]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  4, i3);
            sprintf(buffer, "%.10f", matrix[0][2]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  5, i3);
            sprintf(buffer, "%.10f", vector[0]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  6, i3);
            sprintf(buffer, "%.10f", matrix[1][0]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  7, i3);
            sprintf(buffer, "%.10f", matrix[1][1]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  8, i3);
            sprintf(buffer, "%.10f", matrix[1][2]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring,  9, i3);
            sprintf(buffer, "%.10f", vector[1]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring, 10, i3);
            sprintf(buffer, "%.10f", matrix[2][0]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring, 11, i3);
            sprintf(buffer, "%.10f", matrix[2][1]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring, 12, i3);
            sprintf(buffer, "%.10f", matrix[2][2]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring, 13, i3);
            sprintf(buffer, "%.10f", vector[2]);
            cifstring.Copy(buffer);
            t3->UpdateCell(cifstring, 14, i3);
       }

       _CifObj->WriteTable(Table, _StructureId, "pdbx_struct_assembly");
       _CifObj->WriteTable(t1, _StructureId, "pdbx_struct_assembly_gen");
       _CifObj->WriteTable(t2, _StructureId, "pdbx_struct_assembly_prop");
       _CifObj->WriteTable(t3, _StructureId, "pdbx_struct_oper_list");
}

void _MAXIT::version_update_read()
{
       if (!__need_file_version_update) return;

       if (_num_mol == 0 || _molecules == NULL) return;

       if (!_CifObj) return;

       ISTable *Table = _CifObj->GetTablePtr(_StructureId, "struct_site_gen");
       if (!is_empty_table(Table)) cifndb_process_struct_site_gen(Table);
       Table = _CifObj->GetTablePtr(_StructureId, "rcsb_struct_link");
       if (!is_empty_table(Table)) cifndb_process_rcsb_link(Table);
       Table = _CifObj->GetTablePtr(_StructureId, "struct_conn");
       if (!is_empty_table(Table)) cifndb_process_struct_conn(Table);

       for (int i2 = 0; i2 < _n_site; i2++) {
            if (_site[i2]->numRes == 0) continue;
            _site[i2]->residue = new _RESIDUE*[_site[i2]->numRes];
            for (int j = 0; j < _site[i2]->numRes; j++) {
                 _site[i2]->residue[j] = NULL;
                 _CHAIN *chain = _molecules[0]->GetIndexChain(_site[i2]->initIndex[j]);
                 if (!chain) continue;
                 _FIELD *Init = chain->SeqResIndex(_site[i2]->initRes[j]);
                 if (!Init) continue;
                 _site[i2]->residue[j] = Init->Residue;
            }
       }

       for (int i = 0; i < _n_link; i++) {
            _CHAIN *fst = _molecules[0]->GetIndexChain(_link[i]->FstIndex);
            _CHAIN *snd = _molecules[0]->GetIndexChain(_link[i]->SndIndex);
            if (!fst || !snd) continue;
            _FIELD *Fst = fst->SeqResIndex(_link[i]->FstRes);
            _FIELD *Snd = snd->SeqResIndex(_link[i]->SndRes);
            if (!Fst || !Snd) continue;
            _link[i]->fresidue = Fst->Residue;
            _link[i]->sresidue = Snd->Residue;
            if (!_link[i]->fresidue || !_link[i]->sresidue) continue;

            _ATOM *atom = _link[i]->fresidue->find_atom(_link[i]->FstAtom);
            if (atom) strcpy(_link[i]->FstAtom, atom->pub_atmnam()); 
            atom = _link[i]->sresidue->find_atom(_link[i]->SndAtom);
            if (atom) strcpy(_link[i]->SndAtom, atom->pub_atmnam()); 
       }
}

void _MAXIT::version_update_write()
{
       if (!__need_file_version_update) return;

       if (_num_mol == 0 || _molecules == NULL) return;

       _ATOM *atom1 = NULL, *atom2 = NULL;
       int FstIndex = -1, FstRes = -1, SndIndex = -1, SndRes = -1;

       for (int i2 = 0; i2 < _n_site; i2++) {
            for (int j = 0; j < _site[i2]->numRes; j++) {
                 if (_site[i2]->residue[j] == NULL) {
                      _site[i2]->initIndex[j] = -1;
                      _site[i2]->initRes[j] = -1;
                      continue;
                 }
                 atom1 = _site[i2]->residue[j]->GetFirstAtom();
                 if (_molecules[0]->find_pdb_residue(atom1->pdb_chnid(),
                         atom1->pdb_resnam(), atom1->pdb_resnum(),
                         atom1->ins_code(), FstIndex, FstRes)) {
                      _site[i2]->initIndex[j] = FstIndex;
                      _site[i2]->initRes[j] = FstRes;
                 }
            }
       }

       for (int i = 0; i < _n_link; i++) {
            _link[i]->FstIndex = -1;
            _link[i]->FstRes = -1;
            _link[i]->SndIndex = -1;
            _link[i]->SndRes = -1;
            if (!_link[i]->fresidue || !_link[i]->sresidue) continue;

            atom1 = _link[i]->fresidue->GetFirstAtom();
            while (atom1) {
                 if (!strcmp(_link[i]->FstAtom, atom1->pub_atmnam())) break;
                 atom1 = _link[i]->fresidue->GetNextAtom();
            }
            atom2 = _link[i]->sresidue->GetFirstAtom();
            while (atom2) {
                 if (!strcmp(_link[i]->SndAtom, atom2->pub_atmnam())) break;
                 atom2 = _link[i]->sresidue->GetNextAtom();
            }
            if (!atom1 || !atom2) continue;

            if (_molecules[0]->find_pdb_residue(atom1->pdb_chnid(),
                    atom1->pdb_resnam(), atom1->pdb_resnum(),
                    atom1->ins_code(), FstIndex, FstRes) &&
                _molecules[0]->find_pdb_residue(atom2->pdb_chnid(),
                    atom2->pdb_resnam(), atom2->pdb_resnum(),
                    atom2->ins_code(), SndIndex, SndRes)) {
                 _link[i]->FstIndex = FstIndex;
                 _link[i]->FstRes = FstRes;
                 _link[i]->SndIndex = SndIndex;
                 _link[i]->SndRes = SndRes;
                 strcpy(_link[i]->FstAtom, atom1->pdb_atmnam());
                 strcpy(_link[i]->SndAtom, atom2->pdb_atmnam());
            }
       }
}
/*
static void get_pdb_asym_correspondence(CifFileObj *fobj, const char *blockId,
                ReVarCifArray<_PDB_ASYM_CORR>& chainids)
{
       chainids.Clear();
       ISTable *Table = fobj->GetTablePtr(blockId, "struct_biol_gen");
       CifString cifstring, cifstring1;
       _PDB_ASYM_CORR asym_corr;
       if (Table) {
            int rowNo = Table->GetNumRows();
            int idx1 = Table->GetColumnIndex("symmetry");
            int idx2 = Table->GetColumnIndex("ndb_new_pdb_asym_id");
            int idx3 = Table->GetColumnIndex("asym_id");
            for (int i = 0; i < rowNo; i++) {
                 Table->GetCell(cifstring, idx1, i);
                 if (cifstring != "1_555") continue;

                 Table->GetCell(cifstring, idx2, i);
                 Table->GetCell(cifstring1, idx3, i);
                 if (cifstring != "" && cifstring != "?" && cifstring != ".") {
                      int found = 0;
                      for (unsigned int j = 0; j < chainids.Length(); j++) {
                           if (chainids[j].PDB_chain_ID == cifstring) {
                                found = 1;
                                chainids[j].Asym_IDs.Add(cifstring1);
                                break;
                           }
                      }
                      if (!found) {
                           asym_corr.PDB_chain_ID.Copy(cifstring);
                           asym_corr.Asym_IDs.Clear();
                           asym_corr.Asym_IDs.Add(cifstring1);
                           chainids.Add(asym_corr);
                      }
                 }
            }
       }
}
*/
