/*
FILE:     pdb2cif_form.C
*/
/*
VERSION:  8.120
*/
/*
DATE:     2/12/2008
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2008 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

#include "DICFileObj.h"
#include "erf_iolib.h"
#include "erf_category.h"
#include "erf_util.h"
#include "ndb_misclib.h"

#include "remark.h"
#include "remark_100s.h"
#include "remark_200s.h"
#include "remark_300s.h"
#include "remark_general.h"
#include "remark_3_prolsqs.h"
#include "remark_3_refmac5.h"
#include "remark_3_phenix.h"
#include "remark_3_shelxl.h"
#include "remark_3_tnt.h"
#include "remark_3_xplors.h"

typedef struct {
        int num_word;
        char word[50][100];
} WORD;

static DICFileObj *dict = NULL;
static ISTable *aliases = NULL;
static ISTable *items = NULL;
static ReVarCifArray<CifString> queryCol1;
static ReVarCifArray<CifString> queryCol2;
static int iCol1, colIndex;

static int exchange_flag = 0;

static char defalut_dir[] = "/rcsbapps";

static void write_remark_to_cif_txt(void);
static int  database_stored(const char *token, int field, WORD *cif_category,
               const int is_explicit, const char *program);
// static void get_word_for_special_line(const char *line, char *word, const int added);
// static int  get_word_for_refmac(char *line, WORD& cif_category);
static void write_remark_cif(FILE *fp, FILE *fp1, REMARKS *remark, const int num_remark,
                             const char *program);
static void write_remark_group_cif(FILE *fp, FILE *fp1, GROUP_REMARKS *group_remark,
                             const char *program);
static void read_binary_dictionary();
static void find_exchange_items(char *item);

int main(int argc, char **argv)
{
       char logfile[80];

       root_dir = getenv("RCSBROOT");
       if (root_dir == NULL) root_dir = defalut_dir;

       if (argc == 2 && !strcmp(argv[1], "-exchange")) exchange_flag = 1;

       strcpy(logfile, "pdb2cif_form.err");
       ndb_set_debug_level(NDB_MSG_INFO);
       ndb_open_log("pdb2cif_form", logfile);

       if (!read_ndb_tokens()) {
            printf("Read ndb_tokens failed\n");
            return -1;
       }
       if (!read_cif_db() || !read_cif_rms()) {
            printf("Read ndb_cif or ndb_cif_rms failed\n");
            return -1;
       }

       if (exchange_flag) {
            read_binary_dictionary();
            if (aliases && items) {
                 queryCol1.Clear();
                 colIndex = aliases->GetColumnIndex("name");
                 queryCol1.Add("alias_namelc");
                 queryCol1.Add("dictionary");
                 queryCol1.Add("version");
                 aliases->CreateIndex("ind0", queryCol1);

                 queryCol2.Clear();
                 queryCol2.Add("namelc");
                 items->CreateIndex("ind1", queryCol2);
            }
       }

       write_remark_to_cif_txt();

       ndb_close_log();
       return 0;
}

static void write_remark_to_cif_txt(void)
{
       FILE *fp = fopen("remark_to_mmcif.txt", "w");

       // fprintf(fp, "Xplor:\n");
       fprintf(fp, "#\ndata_XPLOR\n#\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_mapping.pdb_remark\n");
       fprintf(fp, "_mapping.category\n");
       FILE *fp1 = fopen("xplor.html", "w");
       fprintf(fp1, "<html>\n<body>\n<pre>\n");
       write_remark_group_cif(fp, fp1, &Remark_XPlor, "x-plor");
       fprintf(fp1, "</pre>\n</body>\n</html>\n");
       fclose (fp1);
       // fprintf(fp, "\014\n");


       // fprintf(fp, "CNS:\n");
       fprintf(fp, "#\ndata_CNS\n#\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_mapping.pdb_remark\n");
       fprintf(fp, "_mapping.category\n");
       fp1 = fopen("cns", "w");
       write_remark_group_cif(fp, fp1, &Remark_CNS, "cns");
       fclose (fp1);
       // fprintf(fp, "\014\n");

       // fprintf(fp, "CNX:\n");
       fprintf(fp, "#\ndata_CNX\n#\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_mapping.pdb_remark\n");
       fprintf(fp, "_mapping.category\n");
       fp1 = fopen("cnx", "w");
       write_remark_group_cif(fp, fp1, &Remark_CNX, "cnx");
       fclose (fp1);
       // fprintf(fp, "\014\n");

       // fprintf(fp, "Nuclsq:\n");
       fprintf(fp, "#\ndata_NUCLSQ\n#\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_mapping.pdb_remark\n");
       fprintf(fp, "_mapping.category\n");
       fp1 = fopen("nuclsq", "w");
       write_remark_group_cif(fp, fp1, &Remark_Nuclsq, "nuclsq");
       fclose (fp1);
       // fprintf(fp, "\014\n");
       
       // fprintf(fp, "Prolsq:\n");
       fprintf(fp, "#\ndata_PROLSQ\n#\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_mapping.pdb_remark\n");
       fprintf(fp, "_mapping.category\n");
       fp1 = fopen("prolsq", "w");
       write_remark_group_cif(fp, fp1, &Remark_Prolsq, "prolsq");
       fclose (fp1);
       // fprintf(fp, "\014\n");
       
       // fprintf(fp, "Refmac:\n");
       fprintf(fp, "#\ndata_REFMAC\n#\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_mapping.pdb_remark\n");
       fprintf(fp, "_mapping.category\n");
       fp1 = fopen("refmac", "w");
       write_remark_group_cif(fp, fp1, &Remark_Refmac, "refmac");
       fclose (fp1);
       // fprintf(fp, "\014\n");

       // fprintf(fp, "Refmac5.0 & plus:\n");
       fprintf(fp, "#\ndata_REFMAC5\n#\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_mapping.pdb_remark\n");
       fprintf(fp, "_mapping.category\n");
       fp1 = fopen("refmac5", "w");
       write_remark_group_cif(fp, fp1, &Remark_Refmac_5_0, "refmac 5");
       fclose (fp1);
       // fprintf(fp, "\014\n");
       
       // fprintf(fp, "Shelx:\n");
       fprintf(fp, "#\ndata_SHELX\n#\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_mapping.pdb_remark\n");
       fprintf(fp, "_mapping.category\n");
       fp1 = fopen("shelx", "w");
       write_remark_group_cif(fp, fp1, &Remark_Shelxl, "shelx");
       fclose (fp1);
       // fprintf(fp, "\014\n");
       
       // fprintf(fp, "Tnt:\n");
       fprintf(fp, "#\ndata_TNT\n#\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_mapping.pdb_remark\n");
       fprintf(fp, "_mapping.category\n");
       fp1 = fopen("tnt", "w");
       write_remark_group_cif(fp, fp1, &Remark_Tnt, "tnt");
       fclose (fp1);
       // fprintf(fp, "\014\n");

       // fprintf(fp, "Buster_tnt:\n");
       fprintf(fp, "#\ndata_BUSTER_TNT\n#\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_mapping.pdb_remark\n");
       fprintf(fp, "_mapping.category\n");
       fp1 = fopen("buster", "w");
       write_remark_group_cif(fp, fp1, &Remark_Buster_Tnt, "Buster_tnt");
       fclose (fp1);
       // fprintf(fp, "\014\n");

       // fprintf(fp, "Remark_200:\n");
       fprintf(fp, "#\ndata_REMARK_200\n#\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_mapping.pdb_remark\n");
       fprintf(fp, "_mapping.category\n");
       fp1 = fopen("remark200", "w");
       write_remark_cif(fp, fp1, Remark_200, Num_Remark_200, "");
       fclose (fp1);
       // fprintf(fp, "\014\n");

       // fprintf(fp, "Remark_210:\n");
       fprintf(fp, "#\ndata_REMARK_210\n#\n");
       fprintf(fp, "loop_\n");
       fprintf(fp, "_mapping.pdb_remark\n");
       fprintf(fp, "_mapping.category\n");
       fp1 = fopen("remark210", "w");
       write_remark_cif(fp, fp1, Remark_210, Num_Remark_210, "");
       fclose (fp1);

       fclose(fp);
}

static int database_stored(const char *token, int field, WORD *cif_category,
               const int is_explicit, const char *program)
{
       char name[20], abbreviation = ' ';
       int  id, i, j, k, card_id;

       if (!cif_category) return 0;

       if ((!strcmp(token, "ALTSQN") || !strcmp(token, "ALTSQR") ||
            !strcmp(token, "PDBSQN") || !strcmp(token, "PDBSQR") ||
            !strcmp(token, "NDBSQN") || !strcmp(token, "SEQRES")) &&
            field > 5) field = 5;
          
       cif_category->num_word = 0;
       for (id = 0; id < 50; id++) {
            memset(cif_category->word[id], '\0', 100);
       }
       if (!strcmp(token, "ATOMMN") || !strcmp(token, "ATOMMP") ||
           !strcmp(token, "ATOMP") || !strcmp(token, "HETAC") ||
           !strcmp(token, "HETAD") || !strcmp(token, "HETAI") ||
           !strcmp(token, "HETAS") || !strcmp(token, "HETIC") ||
           !strcmp(token, "HETAIN"))
            strcpy(name, "ATOMN");
       else strcpy(name, token);
       for (i = 0; i < MxCategory; i++) {
            for (j = 0; j < CifCategory[i].nkeywords; j++) {
                 for (k = 0; k < CifCategory[i].keywords[j].numNdbInfo; k++) {
                      if (!strcmp(CifCategory[i].keywords[j].NdbInfo[k].NdbTokenName, name) &&
                          strcmp(CifCategory[i].category, "rcsb_nmr_refine") &&
                          field == CifCategory[i].keywords[j].NdbInfo[k].NdbFieldNo) {
                           sprintf(cif_category->word[cif_category->num_word],
                                 "%s.%s", CifCategory[i].category,
                                 CifCategory[i].keywords[j].name);
                           if (exchange_flag) find_exchange_items(cif_category->word[cif_category->num_word]);
                           cif_category->num_word += 1;
// new
                           strcpy(cif_category->word[0], CifCategory[i].category);
                           strcpy(cif_category->word[1], CifCategory[i].keywords[j].name);
                      }
                 }
            }
       }
       if (cif_category->num_word == 0) {
            if (strcmp(program, "")) {
                 for (i = 0; i < mxPrograms; i++) {
                      if (strstr(program, programs[i].program)) {
                           abbreviation = programs[i].abbreviation[0];
                           break;
                      }
                 }
            }

            card_id = find_card_id(token);
            for (i = 0; i < mxRmsRecords; i++) {
                 if (abbreviation == ' ' || abbreviation == rmsRecords[i].type[0] &&
                     rmsRecords[i].type[1] == '_') {
                      for (j = 0; j < NDB_CIF_RMS_ITEMS; j++) {
                           if (card_id == rmsRecords[i].tokens[j] &&
                               (field - 1) == rmsRecords[i].fields[j]) {
                                strcpy(cif_category->word[0], "refine_ls_restr");
                                if (j == 0) 
                                     strcpy(cif_category->word[1], "dev_ideal");
                                if (j == 1)
                                     strcpy(cif_category->word[1], "dev_ideal_target");
                                if (j == 2)
                                     strcpy(cif_category->word[1], "ndb_weight");
                                if (j == 3)
                                     strcpy(cif_category->word[1], "ndb_count");
                                strcpy(cif_category->word[2], "type");
                                strcpy(cif_category->word[3], rmsRecords[i].type);
/*
                                if (is_explicit == 2) {
                                     strcpy(cif_category->word[cif_category->num_word], rmsRecords[i].type);
                                     cif_category->num_word += 1;
                                     return 1;
                                } else if (is_explicit) {
                                     strcpy(cif_category->word[cif_category->num_word],
                                        "refine_ls_restr.type \"");
                                     strcat(cif_category->word[cif_category->num_word], rmsRecords[i].type);
                                     strcat(cif_category->word[cif_category->num_word], "\"");
                                } else if (exchange_flag)
                                     find_exchange_items(cif_category->word[cif_category->num_word]);
                                cif_category->num_word += 1;
*/
                                return 0;
                           }
                      }
                 }
            }
       }
       return 0;
}
/*
static void get_word_for_special_line(const char *line, char *word, const int added)
{
       if (!strcmp(line, " RMS DEVIATIONS FROM IDEAL VALUES.")) {
            if (added) strcpy(word, "refine_ls_restr.dev_ideal");
       } else if (!strcmp(line, " ISOTROPIC THERMAL FACTOR RESTRAINTS.    RMS    SIGMA") ||
                  !strcmp(line, " DISTANCE RESTRAINTS.                    RMS     SIGMA") ||
                  !strcmp(line, " DISTANCE RESTRAINTS.                    RMS    SIGMA") ||
                  !strcmp(line, " NON-BONDED CONTACT RESTRAINTS.") ||
                  !strcmp(line, " CONFORMATIONAL TORSION ANGLE RESTRAINTS."))
            strcpy(word, "refine_ls_restr.dev_ideal and .dev_ideal_target");
       else if (!strcmp(line, " RMS DEVIATIONS FROM IDEAL VALUES.    RMS    WEIGHT  COUNT"))
            strcpy(word, "refine_ls_restr.dev_ideal, .weight and .number");
       else if (!strcmp(line, " ISOTROPIC THERMAL FACTOR RESTRAINTS.     COUNT   RMS    WEIGHT") ||
                !strcmp(line, " ANISOTROPIC THERMAL FACTOR RESTRAINTS.    COUNT   RMS   WEIGHT") ||
                !strcmp(line, " RMS DEVIATIONS FROM IDEAL VALUES        COUNT    RMS    WEIGHT"))
            strcpy(word, "refine_ls_restr.number, dev_ideal and .dev_ideal_target");
       else if (!strcmp(line, " RMS DEVIATIONS FROM RESTRAINT TARGET VALUES."))
            strcpy(word, "refine_ls_restr.dev_ideal");
}

static int  get_word_for_refmac(char *line, WORD& cif_category)
{
       int ret = 0;

       if (!strcmp(line, "NCS GROUP NUMBER               :")) {
            cif_category.num_word = 1;
            strcpy(cif_category.word[0], "struct_ncs_dom.id");
       } else if (!strcmp(line, "CHAIN NAMES                    :")) {
            cif_category.num_word = 1;
            strcpy(cif_category.word[0], "struct_ncs_dom.details");
       // } else if (!strcmp(line, "NUMBER OF COMPONENTS NCS GROUP :")) {
       } else if (!strcmp(line, "COMPONENT C  SSSEQI  TO  C   SSSEQI   CODE")) {
            cif_category.num_word = 6;
            strcpy(cif_category.word[0], "struct_ncs_dom_lim.pdbx_component_id");
            strcpy(cif_category.word[1], "struct_ncs_dom_lim.beg_auth_asym_id");
            strcpy(cif_category.word[2], "struct_ncs_dom_lim.beg_auth_seq_id");
            strcpy(cif_category.word[3], "struct_ncs_dom_lim.end_auth_asym_id");
            strcpy(cif_category.word[4], "struct_ncs_dom_lim.end_auth_seq_id");
            strcpy(cif_category.word[5], "struct_ncs_dom_lim.pdbx_refine_code");
       } else if (!strcmp(line, "GROUP CHAIN        COUNT   RMS     WEIGHT")) {
            cif_category.num_word = 6;
            strcpy(cif_category.word[0], "refine_ls_restr_ncs.pdbx_type");
            strcpy(cif_category.word[1], "refine_ls_restr_ncs.dom_id");
            strcpy(cif_category.word[2], "refine_ls_restr_ncs.pdbx_asym_id");
            strcpy(cif_category.word[3], "refine_ls_restr_ncs.pdbx_number");
            strcpy(cif_category.word[4], "refine_ls_restr_ncs.rms_dev_position");
            strcpy(cif_category.word[5], "refine_ls_restr_ncs.weight_position");
       } else if (!strcmp(line, "TIGHT POSITIONAL")) {
            strcpy(line, "TIGHT POSITIONAL             (A):"); 
            ret = 1;
            cif_category.num_word = 1;
            strcpy(cif_category.word[0], "tight positional");
       } else if (!strcmp(line, "MEDIUM POSITIONAL")) {
            strcpy(line, "MEDIUM POSITIONAL            (A):");
            ret = 1;
            cif_category.num_word = 1;
            strcpy(cif_category.word[0], "medium positional");
       } else if (!strcmp(line, "LOOSE POSITIONAL")) {
            strcpy(line, "LOOSE POSITIONAL             (A):");
            ret = 1;
            cif_category.num_word = 1;
            strcpy(cif_category.word[0], "loose positional");
       } else if (!strcmp(line, "TIGHT THERMAL")) {
            strcpy(line, "TIGHT THERMAL             (A**2):");
            ret = 1;
            cif_category.num_word = 1;
            strcpy(cif_category.word[0], "tight thermal");
       } else if (!strcmp(line, "MEDIUM THERMAL")) {
            strcpy(line, "MEDIUM THERMAL            (A**2):");
            ret = 1;
            cif_category.num_word = 1;
            strcpy(cif_category.word[0], "medium thermal");
       } else if (!strcmp(line, "LOOSE THERMAL")) {
            strcpy(line, "LOOSE THERMAL      1    A (A**2):");
            ret = 1;
            cif_category.num_word = 1;
            strcpy(cif_category.word[0], "loose thermal");
       } else if (!strcmp(line, "TLS GROUP :")) {
            cif_category.num_word = 1;
            if (exchange_flag)
                 strcpy(cif_category.word[0], "pdbx_refine_tls.id");
            else strcpy(cif_category.word[0], "ccp4_refine_tls.id");
       // } else if (!strcmp(line, "NUMBER OF COMPONENTS GROUP :")) {
       } else if (!strcmp(line, "RESIDUE RANGE :")) {
            cif_category.num_word = 4;
            if (exchange_flag) {
                 strcpy(cif_category.word[0], "pdbx_refine_tls_group.beg_label_asym_id");
                 strcpy(cif_category.word[1], "pdbx_refine_tls_group.beg_label_seq_id");
                 strcpy(cif_category.word[2], "pdbx_refine_tls_group.end_label_asym_id");
                 strcpy(cif_category.word[3], "pdbx_refine_tls_group.end_label_seq_id");
            } else {
                 strcpy(cif_category.word[0], "ccp4_refine_tls_group.beg_label_asym_id");
                 strcpy(cif_category.word[1], "ccp4_refine_tls_group.beg_label_seq_id");
                 strcpy(cif_category.word[2], "ccp4_refine_tls_group.end_label_asym_id");
                 strcpy(cif_category.word[3], "ccp4_refine_tls_group.end_label_seq_id");
            }
       } else if (!strcmp(line, "ORIGIN FOR THE GROUP (A):")) {
            cif_category.num_word = 3;
            if (exchange_flag) {
                 strcpy(cif_category.word[0], "pdbx_refine_tls.origin_x");
                 strcpy(cif_category.word[1], "pdbx_refine_tls.origin_y");
                 strcpy(cif_category.word[2], "pdbx_refine_tls.origin_z");
            } else {
                 strcpy(cif_category.word[0], "ccp4_refine_tls.origin_x");
                 strcpy(cif_category.word[1], "ccp4_refine_tls.origin_y");
                 strcpy(cif_category.word[2], "ccp4_refine_tls.origin_z");
            }
       } else if (!strcmp(line, "T11:")) {
            strcpy(line, "T11:          T22:");
            cif_category.num_word = 2;
            if (exchange_flag) {
                 strcpy(cif_category.word[0], "pdbx_refine_tls.T[1][1]");
                 strcpy(cif_category.word[1], "pdbx_refine_tls.T[2][2]"); 
            } else {
                 strcpy(cif_category.word[0], "ccp4_refine_tls.T[1][1]");
                 strcpy(cif_category.word[1], "ccp4_refine_tls.T[2][2]"); 
            }
       } else if (!strcmp(line, "T33:")) {
            strcpy(line, "T33:          T12:");
            cif_category.num_word = 2;
            if (exchange_flag) {
                 strcpy(cif_category.word[0], "pdbx_refine_tls.T[3][3]");
                 strcpy(cif_category.word[1], "pdbx_refine_tls.T[1][2]"); 
            } else {
                 strcpy(cif_category.word[0], "ccp4_refine_tls.T[3][3]");
                 strcpy(cif_category.word[1], "ccp4_refine_tls.T[1][2]"); 
            }
       } else if (!strcmp(line, "T13:")) {
            strcpy(line, "T13:          T23:");
            cif_category.num_word = 2;
            if (exchange_flag) {
                 strcpy(cif_category.word[0], "pdbx_refine_tls.T[1][3]");
                 strcpy(cif_category.word[1], "pdbx_refine_tls.T[2][3]"); 
            } else {
                 strcpy(cif_category.word[0], "ccp4_refine_tls.T[1][3]");
                 strcpy(cif_category.word[1], "ccp4_refine_tls.T[2][3]"); 
            }
       } else if (!strcmp(line, "L11:")) {
            strcpy(line, "L11:          L22:");
            cif_category.num_word = 2;
            if (exchange_flag) {
                 strcpy(cif_category.word[0], "pdbx_refine_tls.L[1][1]");
                 strcpy(cif_category.word[1], "pdbx_refine_tls.L[2][2]"); 
            } else {
                 strcpy(cif_category.word[0], "ccp4_refine_tls.L[1][1]");
                 strcpy(cif_category.word[1], "ccp4_refine_tls.L[2][2]"); 
            }
       } else if (!strcmp(line, "L33:")) {
            strcpy(line, "L33:          L12:");
            cif_category.num_word = 2;
            if (exchange_flag) {
                 strcpy(cif_category.word[0], "pdbx_refine_tls.L[3][3]");
                 strcpy(cif_category.word[1], "pdbx_refine_tls.L[1][2]"); 
            } else {
                 strcpy(cif_category.word[0], "ccp4_refine_tls.L[3][3]");
                 strcpy(cif_category.word[1], "ccp4_refine_tls.L[1][2]"); 
            }
       } else if (!strcmp(line, "L13:")) {
            strcpy(line, "L13:          L23:");
            cif_category.num_word = 2;
            if (exchange_flag) {
                 strcpy(cif_category.word[0], "pdbx_refine_tls.L[1][3]");
                 strcpy(cif_category.word[1], "pdbx_refine_tls.L[2][3]"); 
            } else {
                 strcpy(cif_category.word[0], "ccp4_refine_tls.L[1][3]");
                 strcpy(cif_category.word[1], "ccp4_refine_tls.L[2][3]"); 
            }
       } else if (!strcmp(line, "S11:")) {
            strcpy(line, "S11:          S12:          S13:");
            cif_category.num_word = 3;
            if (exchange_flag) {
                 strcpy(cif_category.word[0], "pdbx_refine_tls.S[1][1]");
                 strcpy(cif_category.word[1], "pdbx_refine_tls.S[1][2]"); 
                 strcpy(cif_category.word[2], "pdbx_refine_tls.S[1][3]"); 
            } else {
                 strcpy(cif_category.word[0], "ccp4_refine_tls.S[1][1]");
                 strcpy(cif_category.word[1], "ccp4_refine_tls.S[1][2]"); 
                 strcpy(cif_category.word[2], "ccp4_refine_tls.S[1][3]"); 
            }
       } else if (!strcmp(line, "S21:")) {
            strcpy(line, "S21:          S22:          S23:");
            cif_category.num_word = 3;
            if (exchange_flag) {
                 strcpy(cif_category.word[0], "pdbx_refine_tls.S[2][1]");
                 strcpy(cif_category.word[1], "pdbx_refine_tls.S[2][2]"); 
                 strcpy(cif_category.word[2], "pdbx_refine_tls.S[2][3]"); 
            } else {
                 strcpy(cif_category.word[0], "ccp4_refine_tls.S[2][1]");
                 strcpy(cif_category.word[1], "ccp4_refine_tls.S[2][2]"); 
                 strcpy(cif_category.word[2], "ccp4_refine_tls.S[2][3]"); 
            } 
       } else if (!strcmp(line, "S31:")) {
            strcpy(line, "S31:          S32:          S33:");
            cif_category.num_word = 3;
            if (exchange_flag) {
                 strcpy(cif_category.word[0], "pdbx_refine_tls.S[3][1]");
                 strcpy(cif_category.word[1], "pdbx_refine_tls.S[3][2]"); 
                 strcpy(cif_category.word[2], "pdbx_refine_tls.S[3][3]"); 
            } else {
                 strcpy(cif_category.word[0], "ccp4_refine_tls.S[3][1]");
                 strcpy(cif_category.word[1], "ccp4_refine_tls.S[3][2]"); 
                 strcpy(cif_category.word[2], "ccp4_refine_tls.S[3][3]"); 
            }
       }

       return ret;
}
*/
static void write_remark_cif(FILE *fp, FILE *fp1, REMARKS *remark, const int num_remark,
                             const char *program)
{
       int i, j, k;
       char line[500], line1[500], word[100], FieldFormat[82];
       WORD cif_category;

       for (i = 0; i < num_remark; i++) {
            memset(line, 0, 500);
            memset(line1, 0, 500);
            memset(word, 0, 100);
            // sprintf(line, "REMARK %3d", remark[i].Remark_No);
            for (j = 0; j < remark[i].NumField; j++) {
                 if (!strcmp(remark[i].FieldList[j].TokenName, "") &&
                      remark[i].FieldList[j].FieldWidth == 0) {
                      remark[i].FieldList[j].FieldWidth = strlen(remark[i].FieldList[j].Text);
                 }
                    
                 if (j == 0)
                      sprintf(FieldFormat, "%%-%ds",remark[i].FieldList[j].FieldStCol-1);
                 else sprintf(FieldFormat, "%%-%ds",remark[i].FieldList[j].FieldStCol-
                          (remark[i].FieldList[j-1].FieldStCol+
                           remark[i].FieldList[j-1].FieldWidth));
                 sprintf(&line1[strlen(line1)], FieldFormat, ""); 
                 // fprintf(fp1, FieldFormat, "");

                 if (remark[i].FieldList[j].FieldJustification==PDB_REMARK_LEFT_JUSTIFIED)
                      sprintf(FieldFormat, "%%-%ds", remark[i].FieldList[j].FieldWidth);
                 else sprintf(FieldFormat, "%%%ds", remark[i].FieldList[j].FieldWidth);

                 if (!strcmp(remark[i].FieldList[j].TokenName, "")) {
                      sprintf(&line1[strlen(line1)], FieldFormat,
                               remark[i].FieldList[j].Text);
                      // fprintf(fp1, FieldFormat, remark[i].FieldList[j].Text);
                      if (!strcmp(remark[i].FieldList[j].Text, ";")) continue;
/*
                      for (k = 0; k < remark[i].FieldList[j].FieldStCol; k++) {
                           strcat(line, " ");
                      }
                      sprintf(&line[strlen(line)], "%s",
                                 remark[i].FieldList[j].Text);
*/
                      if (strcmp(remark[i].FieldList[j].Text, "") &&
                          strcmp(remark[i].FieldList[j].Text, "GROUP") && !strcmp(line, ""))
                           strcpy(line, remark[i].FieldList[j].Text);
                 } else {
                      if (strcmp(line1, "")) {
                           fprintf(fp1, "\"%s\" ? ? ? ? ? ?\n", line1);
                           memset(line1, 0, 500);
                      }
/*
                      if (j == (remark[i].NumField - 1))
                           fprintf(fp1, "<input type=\"text\" name=\"r2\" size=\"%d\">",
                                    72 - remark[i].FieldList[j].FieldStCol);
                      else fprintf(fp1, "<input type=\"text\" name=\"r2\" size=\"%d\">",
                                    remark[i].FieldList[j].FieldWidth); 
*/
                      k = remark[i].FieldList[j].FieldWidth;
                      if (j == (remark[i].NumField - 1))
                           k = 72 - remark[i].FieldList[j].FieldStCol;
                      if (!strcmp(remark[i].FieldList[j].TokenName, "XXX")) continue;
                      database_stored(remark[i].FieldList[j].TokenName,
                          remark[i].FieldList[j].FieldId, &cif_category, 1, program);
                      strcpy(word, cif_category.word[0]);
                      fprintf(fp1, "? R %d ", k);
                      for (k = 0; k < 4; k++) {
                           if (strcmp(cif_category.word[k], ""))
                                fprintf(fp1, "\"%s\" ", cif_category.word[k]);
                           else fprintf(fp1, "? ");
                      }
                      fprintf(fp1, "\n");
                 }
            }
            // ndb_set_string_to_lower(word);
            if (strcmp(line, "") && strcmp(word, "")) fprintf(fp, "'%s' '%s'\n", line, word);
            // fprintf(fp1, "%s\n", word);
            if (strcmp(line1, "")) {
                 fprintf(fp1, "\"%s\" ? ? ? ? ? ?\n", line1);
                 memset(line1, 0, 500);
            }
            fprintf(fp1, "\"%s\" ? ? ? ? ? ?\n", "return");
       }
}

static void write_remark_group_cif(FILE *fp, FILE *fp1, GROUP_REMARKS *group_remark,
                             const char *program)
{
       int n;

       for (n = 0; n < group_remark->Remarks_No; n++) {
            if (group_remark->remarks[n] == Remark_3_AtomCount ||
                group_remark->remarks[n] == Remark_Shelxl_AtomCount ||
                group_remark->remarks[n] == Remark_3_REFMAC_AtomCount) continue;
            write_remark_cif(fp, fp1, group_remark->remarks[n],
                                group_remark->num_remarks[n], program);
            fprintf(fp1, "\n");
            // fprintf(fp, "\n");
       }
}

static void read_binary_dictionary()
{
       unsigned int i;
       int iCol2, iCol3, iCol4, iCol5, iCol6;
       ReVarCifArray<CifString> *col0 = NULL, *col1 = NULL, *col2 = NULL;
       char binaryfile[300], *BlockName = NULL;
       struct stat stat1;

       sprintf(binaryfile, "%s/data/binary/%s", root_dir, "mmcif_pdbx.odb");
       if (!stat(binaryfile, &stat1)) {
            dict = new DICFileObj(binaryfile, READ_MODE);
            BlockName = dict->GetBlockName(0);
            aliases = dict->GetTablePtr(BlockName, "item_aliases");
            items   = dict->GetTablePtr(BlockName, "item");
            if (BlockName) delete [] BlockName;

            iCol1 = items->GetColumnIndex("name");
            col0 = items->GetColumn(iCol1);
            for (i = 0; i < col0->Length(); i++) {
              (*col0)[i].ToLower();
            }
            items->AddColumn("namelc");
            iCol2 = items->GetColumnIndex("namelc");
            items->FillColumn(*col0, iCol2);

            iCol3 = aliases->GetColumnIndex("name");
            col1 = aliases->GetColumn(iCol3);
            for (i = 0; i < col1->Length(); i++) {
              (*col1)[i].ToLower();
            }
            aliases->AddColumn("namelc");
            iCol4 = aliases->GetColumnIndex("namelc");
            aliases->FillColumn(*col1, iCol4);

            iCol5 = aliases->GetColumnIndex("alias_name");
            col2 = aliases->GetColumn(iCol5);
            for (i = 0; i < col2->Length(); i++) {
              (*col2)[i].ToLower();
            }
            aliases->AddColumn("alias_namelc");
            iCol6 = aliases->GetColumnIndex("alias_namelc");
            aliases->FillColumn(*col2, iCol6);
            if (col0) delete col0;
            if (col1) delete col1;
            if (col2) delete col2;
       }
}

static void find_exchange_items(char *item)
{
       int errCode;
       CifString cs, itemName, aliasName;
       ReVarCifArray<CifString> queryTarget1, queryTarget2;
       ReVarPCifArray<int> *queryResult1 = NULL, *queryResult2 = NULL;

       if (!items || !aliases) return;

       cs.Clear(); cs.Copy("_"); cs += item; cs.ToLower();
       queryTarget1.Clear();
       queryTarget1.Add(cs);
       queryTarget1.Add("cif_rcsb.dic");
       queryTarget1.Add("1.1");
       queryResult1 = aliases->Search(queryTarget1, queryCol1, errCode);
       if (errCode < 0) {
            queryTarget2.Clear();
            queryTarget2.Add(cs);
            queryResult2 = items->Search(queryTarget2, queryCol2, errCode);
            if (errCode >= 0) {
                 itemName.Clear();
                 items->GetCell(itemName, iCol1, (*queryResult2)[0]);
                 strcpy(item,  &(itemName.Text()[1]));
            }
            if (queryResult2) delete queryResult2;
       } else {
            aliasName.Clear();
            aliases->GetCell(aliasName, colIndex, (*queryResult1)[0]);
            strcpy(item, &(aliasName.Text()[1]));
       }
       if (queryResult1) delete queryResult1;
}
