/*
FILE:     topology.C
*/
/*
VERSION:  8.120
*/
/*
DATE:     2/12/2008
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2008 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "_validation.h"
#include "bond_angle_torsion.h"
#include "topology_global.h"
#include "util.h"

static void add_single_bond_values(FILE *fp, int &first_write, const int exchange_out,
            const int mol_id, _RESIDUE *residue, const int first);
static void add_mutual_bond_values(FILE *fp, int &first_write, const int exchange_out,
             const int mol_id, const int chain_type, _RESIDUE *prev, _RESIDUE *curr);
static void add_single_angle_values(FILE *fp, int &first_write, const int exchange_out,
                     const int mol_id, _RESIDUE *residue, const int first);
static void add_mutual_angle_values(FILE *fp, int &first_write, const int exchange_out,
            const int mol_id, const int chain_type, _RESIDUE *prev, _RESIDUE *curr);
static int is_residues_linked(const int chain_type, _RESIDUE *prev, _RESIDUE *curr);
static void add_torsions_fp(FILE *fp, int &first_write, const int exchange_out, _ATOM **atom1,
         _ATOM **atom2, _ATOM **atom3, _ATOM **atom4, const int num_a, const int num_b,
         const int num_c, const int num_d, const int chain_type, const int mol_id);

void _VALIDATION::get_molecule_topology()
{
       int i = 0, l = 0, mol_id = 0, is_link = 0, first = 0; 
       int num1 = 0, num2 = 0, num3 = 0, num4 = 0;
       double prev_phi = -360.0, prev_chi = -360.0, tor_value[NUM_TOR_NA];
       double phi = -360.0, psi = -360.0, alpha = -360.0;
       double beta = -360.0, gamma = -360.0, delta = -360.0;
       double epsilon = -360.0, zeta = -360.0, chi = -360.0;
       _CHAIN *chain = NULL;
       _RESIDUE *prev = NULL, *curr = NULL;
       _ATOM **at1 = NULL, **at2 = NULL, **at3 = NULL, **at4 = NULL;

       if (_calculated_bond_angle_torsion) return;

       _dbIndex_init();
       for (i = 0; i < _num_mol; i++) {
            mol_id = _molecules[i]->Mol_ID();
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMP_TOKEN &&
                     chain->chain_type() != ATOMN_TOKEN) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }

                 prev = NULL;
                 prev_phi = -360.0;
                 prev_chi = -360.0;
                 curr = chain->GetFirstResidue();
                 first = 1;
                 while (curr) {
                      if (strcmp(curr->ResName(), "ASI")) {
                           if (chain->chain_type() == ATOMP_TOKEN)
                                curr->set_type(MONOMER_TYPE_A_A);
                           else first = 0;
                           if (chain->IsLastResidue()) first = 2;
                           add_single_values(mol_id, curr, first);
                           is_link = 0;
                           if (prev && curr && strcmp(prev->ResName(), "ASI")) {
                                is_link = add_mutual_values(mol_id,
                                   chain->chain_type(), prev, curr);
                           }
                           if (chain->chain_type() == ATOMP_TOKEN) {
/*
                                if (fp) {
                                     for (l = 0; l < NUM_TOR_AA; l++) {
                                          at1 = curr->find_atom(tor_aa[l].at1, num1);
                                          at2 = curr->find_atom(tor_aa[l].at2, num2);
                                          at3 = curr->find_atom(tor_aa[l].at3, num3);
                                          at4 = curr->find_atom(tor_aa[l].at4, num4);
                                          add_torsions(at1, at2, at3, at4,
                                                   num1, num2, num3, num4,
                                                   tor_aa[l].Rep_Name, tor_aa[l].type,
                                                  chain->chain_type(), mol_id);
                                          if (at1) delete [] at1;
                                          if (at2) delete [] at2;
                                          if (at3) delete [] at3;
                                          if (at4) delete [] at4;
                                     }
                                }
*/

                                if (is_standard_aa_residue(curr->ResName())) {
                                     at1 = curr->find_atom(CALPHA.at1, num1);
                                     at2 = curr->find_atom(CALPHA.at2, num2);
                                     at3 = curr->find_atom(CALPHA.at3, num3);
                                     at4 = curr->find_atom(CALPHA.at4, num4);
                                     add_torsions(at1, at2, at3, at4,
                                                  num1, num2, num3, num4,
                                                  CALPHA.Rep_Name, CALPHA.type,
                                              chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                }
                                phi = -360.0;
                                if (is_link && prev && strcmp(prev->ResName(), "ASI")) {
                                     at1 = prev->find_atom(PHI.at1, num1);
                                     at2 = curr->find_atom(PHI.at2, num2);
                                     at3 = curr->find_atom(PHI.at3, num3);
                                     at4 = curr->find_atom(PHI.at4, num4);
                                     phi = add_torsions(at1, at2, at3, at4,
                                                  num1, num2, num3, num4,
                                                  PHI.Rep_Name, PHI.type,
                                              chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                     at1 = prev->find_atom(PSI.at1, num1);
                                     at2 = prev->find_atom(PSI.at2, num2);
                                     at3 = prev->find_atom(PSI.at3, num3);
                                     at4 = curr->find_atom(PSI.at4, num4);
                                     psi = add_torsions(at1, at2, at3, at4,
                                                  num1, num2, num3, num4,
                                                  PSI.Rep_Name, PSI.type,
                                              chain->chain_type(), mol_id);
                                     if (prev_phi > -360.0 && psi > -360.0) {
                                          insert_torsion_pair(at1[0]->pdb_chnid(),
                                             at1[0]->pdb_resnam(), at1[0]->pdb_resnum(),
                                             at1[0]->ins_code(), prev_phi, psi, mol_id);
                                     }
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                     at1 = prev->find_atom(OMEGA.at1, num1);
                                     at2 = prev->find_atom(OMEGA.at2, num2);
                                     at3 = curr->find_atom(OMEGA.at3, num3);
                                     at4 = curr->find_atom(OMEGA.at4, num4);
                                     add_torsions(at1, at2, at3, at4,
                                                  num1, num2, num3, num4,
                                                  OMEGA.Rep_Name, OMEGA.type,
                                              chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                     at1 = prev->find_atom(PLANAR.at1, num1);
                                     at2 = prev->find_atom(PLANAR.at2, num2);
                                     at3 = curr->find_atom(PLANAR.at3, num3);
                                     at4 = prev->find_atom(PLANAR.at4, num4);
                                     add_torsions(at1, at2, at3, at4,
                                                  num1, num2, num3, num4,
                                                  PLANAR.Rep_Name, PLANAR.type,
                                              chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                }
                                prev_phi = phi;
                           } else {
                                alpha = -360.0;
                                beta = -360.0;
                                gamma = -360.0;
                                delta = -360.0;
                                epsilon = -360.0;
                                zeta = -360.0;
                                chi = -360.0;
                                for (l = 0; l < NUM_TOR_NA; l++) {
                                     at1 = curr->find_atom(tor_na[l].at1, num1);
                                     at2 = curr->find_atom(tor_na[l].at2, num2);
                                     at3 = curr->find_atom(tor_na[l].at3, num3);
                                     at4 = curr->find_atom(tor_na[l].at4, num4);
                                     tor_value[l] = add_torsions(at1, at2, at3, at4,
                                              num1, num2, num3, num4,
                                              tor_na[l].Rep_Name, tor_na[l].type,
                                             chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                }
                                beta = tor_value[0];
                                gamma = tor_value[1];
                                delta = tor_value[2];
                                if (tor_value[3] > -360.0)
                                     chi = tor_value[3];
                                else if (tor_value[4] > -360.0)
                                     chi = tor_value[4];

                                if (is_link && prev) {
                                     at1 = prev->find_atom(ALPHA.at1, num1);
                                     at2 = curr->find_atom(ALPHA.at2, num2);
                                     at3 = curr->find_atom(ALPHA.at3, num3);
                                     at4 = curr->find_atom(ALPHA.at4, num4);
                                     alpha = add_torsions(at1, at2, at3, at4,
                                              num1, num2, num3, num4,
                                              ALPHA.Rep_Name, ALPHA.type,
                                              chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                     at1 = prev->find_atom(EPSILON.at1, num1);
                                     at2 = prev->find_atom(EPSILON.at2, num2);
                                     at3 = prev->find_atom(EPSILON.at3, num3);
                                     at4 = curr->find_atom(EPSILON.at4, num4);
                                     epsilon = add_torsions(at1, at2, at3, at4,
                                              num1, num2, num3, num4,
                                              EPSILON.Rep_Name, EPSILON.type,
                                              chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                     at1 = prev->find_atom(ZETA.at1, num1);
                                     at2 = prev->find_atom(ZETA.at2, num2);
                                     at3 = curr->find_atom(ZETA.at3, num3);
                                     at4 = curr->find_atom(ZETA.at4, num4);
                                     zeta = add_torsions(at1, at2, at3, at4,
                                              num1, num2, num3, num4,
                                              ZETA.Rep_Name, ZETA.type,
                                              chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                }
                                if (chain->na_type() == ATOMN_TYPE_DNA_ONLY) {
                                     if (zeta > -360.0 && alpha > -360.0)
                                          insert_pair_z_a_1(prev->pdb_res_no(), zeta, alpha);
                                     if (zeta > -360.0 && beta > -360.0)
                                          insert_pair_z_b_1(prev->pdb_res_no(), zeta, beta);
                                     if (zeta > -360.0 && epsilon > -360.0)
                                          insert_pair_z_e(prev->pdb_res_no(), zeta, epsilon);
                                     if (alpha > -360.0 && gamma > -360.0)
                                          insert_pair_a_g(curr->pdb_res_no(), alpha, gamma);
                                     if (zeta > -360.0 && prev_chi > -360.0)
                                          insert_pair_z_x(prev->pdb_res_no(), zeta, prev_chi);
                                     if (delta > -360.0 && chi > -360.0)
                                          insert_pair_d_x(curr->pdb_res_no(), delta, chi);
                                }
                                prev_chi = chi;
                           }
                      }
                      prev = curr;
                      curr = chain->GetNextResidue();
                      first = 0;
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }
       _calculated_bond_angle_torsion = 1;
}

void _VALIDATION::get_molecule_topology_fp(FILE *fp, const int exchange_out)
{
       int i = 0, l = 0, mol_id = 0, is_link = 0, first = 0, first_write; 
       int num1 = 0, num2 = 0, num3 = 0, num4 = 0;
       _CHAIN *chain = NULL;
       _RESIDUE *prev = NULL, *curr = NULL;
       _ATOM **at1 = NULL, **at2 = NULL, **at3 = NULL, **at4 = NULL;

       _dbIndex_init();
       first_write = 1;
       for (i = 0; i < _num_mol; i++) {
            mol_id = _molecules[i]->Mol_ID();
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMP_TOKEN &&
                     chain->chain_type() != ATOMN_TOKEN) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }

                 prev = NULL;
                 curr = chain->GetFirstResidue();
                 first = 1;
                 while (curr) {
                      if (strcmp(curr->ResName(), "ASI")) {
                           if (chain->chain_type() == ATOMP_TOKEN)
                                curr->set_type(MONOMER_TYPE_A_A);
                           else first = 0;
                           if (chain->IsLastResidue()) first = 2;
                           add_single_bond_values(fp, first_write, exchange_out,
                                       mol_id, curr, first);
                           if (prev && curr && strcmp(prev->ResName(), "ASI")) {
                                add_mutual_bond_values(fp, first_write, exchange_out,
                                        mol_id, chain->chain_type(), prev, curr);
                           }
                      }
                      prev = curr;
                      curr = chain->GetNextResidue();
                      first = 0;
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }

       first_write = 1;
       for (i = 0; i < _num_mol; i++) {
            mol_id = _molecules[i]->Mol_ID();
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMP_TOKEN &&
                     chain->chain_type() != ATOMN_TOKEN) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }

                 prev = NULL;
                 curr = chain->GetFirstResidue();
                 first = 1;
                 while (curr) {
                      if (strcmp(curr->ResName(), "ASI")) {
                           if (chain->chain_type() == ATOMP_TOKEN)
                                curr->set_type(MONOMER_TYPE_A_A);
                           else first = 0;
                           if (chain->IsLastResidue()) first = 2;
                           add_single_angle_values(fp, first_write, exchange_out,
                                       mol_id, curr, first);
                           if (prev && curr && strcmp(prev->ResName(), "ASI")) {
                                add_mutual_angle_values(fp, first_write, exchange_out,
                                       mol_id, chain->chain_type(), prev, curr);
                           }
                      }
                      prev = curr;
                      curr = chain->GetNextResidue();
                      first = 0;
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }

       first_write = 1;
       for (i = 0; i < _num_mol; i++) {
            mol_id = _molecules[i]->Mol_ID();
            chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMP_TOKEN &&
                     chain->chain_type() != ATOMN_TOKEN) {
                      chain = _molecules[i]->GetNextChain();
                      continue;
                 }

                 prev = NULL;
                 curr = chain->GetFirstResidue();
                 while (curr) {
                      if (strcmp(curr->ResName(), "ASI")) {
                           is_link = 0;
                           if (prev && curr && strcmp(prev->ResName(), "ASI")) {
                                is_link = is_residues_linked(chain->chain_type(), prev, curr);
                           }
                           if (chain->chain_type() == ATOMP_TOKEN) {
                                for (l = 0; l < NUM_TOR_AA; l++) {
                                     at1 = curr->find_atom(tor_aa[l].at1, num1);
                                     at2 = curr->find_atom(tor_aa[l].at2, num2);
                                     at3 = curr->find_atom(tor_aa[l].at3, num3);
                                     at4 = curr->find_atom(tor_aa[l].at4, num4);
                                     add_torsions_fp(fp, first_write, exchange_out,
                                         at1, at2, at3, at4, num1, num2, num3,
                                         num4, chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                }
                                if (is_link && prev && strcmp(prev->ResName(), "ASI")) {
                                     at1 = prev->find_atom(PHI.at1, num1);
                                     at2 = curr->find_atom(PHI.at2, num2);
                                     at3 = curr->find_atom(PHI.at3, num3);
                                     at4 = curr->find_atom(PHI.at4, num4);
                                     add_torsions_fp(fp, first_write, exchange_out,
                                         at1, at2, at3, at4, num1, num2, num3,
                                         num4, chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                     at1 = prev->find_atom(PSI.at1, num1);
                                     at2 = prev->find_atom(PSI.at2, num2);
                                     at3 = prev->find_atom(PSI.at3, num3);
                                     at4 = curr->find_atom(PSI.at4, num4);
                                     add_torsions_fp(fp, first_write, exchange_out,
                                         at1, at2, at3, at4, num1, num2, num3,
                                         num4, chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                     at1 = prev->find_atom(OMEGA.at1, num1);
                                     at2 = prev->find_atom(OMEGA.at2, num2);
                                     at3 = curr->find_atom(OMEGA.at3, num3);
                                     at4 = curr->find_atom(OMEGA.at4, num4);
                                     add_torsions_fp(fp, first_write, exchange_out,
                                         at1, at2, at3, at4, num1, num2, num3,
                                         num4, chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                }
                           } else {
                                for (l = 0; l < NUM_TOR_NA; l++) {
                                     at1 = curr->find_atom(tor_na[l].at1, num1);
                                     at2 = curr->find_atom(tor_na[l].at2, num2);
                                     at3 = curr->find_atom(tor_na[l].at3, num3);
                                     at4 = curr->find_atom(tor_na[l].at4, num4);
                                     add_torsions_fp(fp, first_write, exchange_out,
                                         at1, at2, at3, at4, num1, num2, num3,
                                         num4, chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                }
                                if (is_link && prev) {
                                     at1 = prev->find_atom(ALPHA.at1, num1);
                                     at2 = curr->find_atom(ALPHA.at2, num2);
                                     at3 = curr->find_atom(ALPHA.at3, num3);
                                     at4 = curr->find_atom(ALPHA.at4, num4);
                                     add_torsions_fp(fp, first_write, exchange_out,
                                         at1, at2, at3, at4, num1, num2, num3,
                                         num4, chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                     at1 = prev->find_atom(EPSILON.at1, num1);
                                     at2 = prev->find_atom(EPSILON.at2, num2);
                                     at3 = prev->find_atom(EPSILON.at3, num3);
                                     at4 = curr->find_atom(EPSILON.at4, num4);
                                     add_torsions_fp(fp, first_write, exchange_out,
                                         at1, at2, at3, at4, num1, num2, num3,
                                         num4, chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                     at1 = prev->find_atom(ZETA.at1, num1);
                                     at2 = prev->find_atom(ZETA.at2, num2);
                                     at3 = curr->find_atom(ZETA.at3, num3);
                                     at4 = curr->find_atom(ZETA.at4, num4);
                                     add_torsions_fp(fp, first_write, exchange_out,
                                         at1, at2, at3, at4, num1, num2, num3,
                                         num4, chain->chain_type(), mol_id);
                                     if (at1) delete [] at1;
                                     if (at2) delete [] at2;
                                     if (at3) delete [] at3;
                                     if (at4) delete [] at4;
                                }
                           }
                      }
                      prev = curr;
                      curr = chain->GetNextResidue();
                      first = 0;
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }
}

void _VALIDATION::add_single_values(const int mol_id, _RESIDUE *residue, const int first)
{
       CifString string;
       string.Copy(residue->ResName());
       string += residue->type();
       MONOMER_DB *exist = find_monomer_db(string);
       if (!exist) return;
       int n_bond = exist->mon->nbond;
       MONOMER_BOND *bond = exist->mon->bond;
       int n_angle = exist->mon->nangle;
       MONOMER_ANGLE *angle = exist->mon->angle;

       int i = 0, j = 0, m = 0, check;
       int num1 = 0, num2 = 0, num3 = 0;
       double val = 0, ept = 0, std = 0, diff = 0;
       _ATOM **at1 = NULL, **at2 = NULL, **at3 = NULL;
       _BOND_VALUE *bond_value = NULL;
       _ANGLE_VALUE *angle_value = NULL;

       for (i = 0; i < n_bond; i++) {
            at1 = residue->find_atom(bond[i].at1, num1);
            at2 = residue->find_atom(bond[i].at2, num2);
            check = 1;
            if (residue->type() == MONOMER_TYPE_A_A) {
                if (!strcmp(bond[i].at1, "N") && !strcmp(bond[i].at2, "CA")) {
                     check = 0;
                     if (first == 1) check = 1;
                } else if (!strcmp(bond[i].at1, "C") && !strcmp(bond[i].at2, "O")) {
                     check = 0;
                     if (first == 2) check = 1;
                }
            }
            if (at1 && at2) {
                 bond_value = cal_bonds(at1, num1, at2, num2);
                 if (bond_value) {
                      for (m = 0; m < num1 * num2; m++) {
                           if (!bond_value[m].fst || !bond_value[m].snd) break;
                           val = bond_value[m].val;
                           ept = bond[i].mean_std[0].val;
                           std = bond[i].mean_std[0].std;
                           diff = fabs(val - ept);
/*
                           if (first && residue->type() == MONOMER_TYPE_A_A ||
                              !first && residue->type() != MONOMER_TYPE_A_A) {
*/
                           if (check) {
                                for (j = 1; j < bond[i].num; j++) {
                                     if (fabs(val - bond[i].mean_std[j].val) < diff) {
                                          diff = fabs(val - bond[i].mean_std[j].val);
                                          ept = bond[i].mean_std[j].val;
                                          std = bond[i].mean_std[j].std;
                                     }
                                }
                           }
                           insert_bond(bond_value[m].fst->pdb_chnid(),
                                  bond_value[m].fst->pdb_resnam(),
                                  bond_value[m].fst->pdb_resnum(),
                                  bond_value[m].fst->ins_code(), bond[i].Rep_Name,
                                  bond[i].type, val, ept, std, mol_id,
                                  bond_value[m].fst, bond_value[m].snd);
                      }
                      delete [] bond_value;
                      bond_value = NULL;
                 }
            }
            if (at1) delete [] at1;
            if (at2) delete [] at2;
       }
       for (i = 0; i < n_angle; i++) {
            at1 = residue->find_atom(angle[i].at1, num1);
            at2 = residue->find_atom(angle[i].at2, num2);
            at3 = residue->find_atom(angle[i].at3, num3);
            check = 1;
            if (residue->type() == MONOMER_TYPE_A_A && !strcmp(angle[i].at1, "CA") &&
                !strcmp(angle[i].at2, "C") && !strcmp(angle[i].at3, "O")) {
                 check = 0;
                 if (first == 2) check = 1;
            }

            if (at1 && at2 && at3) {
                 angle_value = cal_angles(at1, num1, at2, num2, at3, num3);
                 if (angle_value) {
                      for (m = 0; m < num1 * num2 * num3; m++) {
                           if (!angle_value[m].fst || !angle_value[m].snd ||
                               !angle_value[m].trd) break;
                           val = angle_value[m].val;
                           ept = angle[i].mean_std[0].val;
                           std = angle[i].mean_std[0].std;
                           diff = fabs(val - ept);
                           if (check) {
                                for (j = 1; j < angle[i].num; j++) {
                                     if (fabs(val - angle[i].mean_std[j].val) < diff) {
                                          diff = fabs(val - angle[i].mean_std[j].val);
                                          ept = angle[i].mean_std[j].val;
                                          std = angle[i].mean_std[j].std;
                                     }
                                }
                           }
                           insert_angle(angle_value[m].fst->pdb_chnid(),
                               angle_value[m].fst->pdb_resnam(),
                               angle_value[m].fst->pdb_resnum(),
                               angle_value[m].fst->ins_code(), angle[i].Rep_Name,
                               angle[i].type, val, ept, std, mol_id,
                               angle_value[m].fst, angle_value[m].snd, angle_value[m].trd);
                      }
                      delete [] angle_value;
                      angle_value = NULL;
                 }
            }
            if (at1) delete [] at1;
            if (at2) delete [] at2;
            if (at3) delete [] at3;
       }
}

int _VALIDATION::add_mutual_values(const int mol_id, const int chain_type,
        _RESIDUE *prev, _RESIDUE *curr)
{
       MONOMER_BOND *bond = NULL;
       MONOMER_ANGLE *angle1 = NULL;
       MONOMER_ANGLE *angle2 = NULL;
       int n_angle1 = 0;
       int n_angle2 = 0;
       int is_link = 0;
       int i = 0, j = 0, m = 0;
       int num1 = 0, num2 = 0, num3 = 0;
       double val = 0, ept = 0, std = 0, diff = 0, cutoff_value = 1.75;
       _ATOM **at1 = NULL, **at2 = NULL, **at3 = NULL;
       _BOND_VALUE *bond_value = NULL;
       _ANGLE_VALUE *angle_value = NULL;

       if (chain_type == ATOMP_TOKEN) {
            if (!strcmp(curr->ResName(), "GLY")) {
                 bond = &A_LINK;
                 angle1 = GLY_ANG1;
                 angle2 = &GLY_ANG2;
                 n_angle1 = 2;
                 n_angle2 = 1;
            } else if (!strcmp(curr->ResName(), "PRO")) {
                 bond = &PRO_LINK;
                 angle1 = PRO_ANG1;
                 angle2 = PRO_ANG2;
                 n_angle1 = 2;
                 n_angle2 = 2;
            } else {
                 bond = &A_LINK;
                 angle1 = AA_ANG1;
                 angle2 = &AA_ANG2;
                 n_angle1 = 2;
                 n_angle2 = 1;
            }
       } else if (chain_type == ATOMN_TOKEN) {
            cutoff_value = 2.10;
            bond = &N_LINK;
            angle1 = &NN_ANG1;
            angle2 = NN_ANG2;
            n_angle1 = 1;
            n_angle2 = 3;
       }

       if (bond) {
            at1 = prev->find_atom(bond->at1, num1);
            at2 = curr->find_atom(bond->at2, num2);
            if (at1 && at2) {
                 bond_value = cal_bonds(at1, num1, at2, num2);
                 if (bond_value) {
                      for (m = 0; m < num1 * num2; m++) {
                           if (!bond_value[m].fst || !bond_value[m].snd) break;
                           if (bond_value[m].val < cutoff_value) {
                                is_link = 1;
                                insert_bond(bond_value[m].fst->pdb_chnid(),
                                  bond_value[m].snd->pdb_resnam(),
                                  bond_value[m].snd->pdb_resnum(),
                                  bond_value[m].snd->ins_code(), bond->Rep_Name,
                                  bond->type, bond_value[m].val, bond->mean_std[0].val,
                                  bond->mean_std[0].std,
                                  mol_id, bond_value[m].fst, bond_value[m].snd);
                           }
                      }
                      delete [] bond_value;
                      bond_value = NULL;
                 }
            }
            if (at1) delete [] at1;
            if (at2) delete [] at2;
       }
       if (!is_link) return is_link;
       for (i = 0; i < n_angle1; i++) {
            at1 = prev->find_atom(angle1[i].at1, num1);
            at2 = prev->find_atom(angle1[i].at2, num2);
            at3 = curr->find_atom(angle1[i].at3, num3);
            if (at1 && at2 && at3) {
                 angle_value = cal_angles(at1, num1, at2, num2, at3, num3);
                 if (angle_value) {
                      for (m = 0; m < num1 * num2 * num3; m++) {
                           if (!angle_value[m].fst || !angle_value[m].snd ||
                               !angle_value[m].trd) break;
                           val = angle_value[m].val;
                           ept = angle1[i].mean_std[0].val;
                           std = angle1[i].mean_std[0].std;
                           diff = fabs(val - ept);
                           for (j = 1; j < angle1[i].num; j++) {
                                if (fabs(val - angle1[i].mean_std[j].val) < diff) {
                                     diff = fabs(val - angle1[i].mean_std[j].val);
                                     ept = angle1[i].mean_std[j].val;
                                     std = angle1[i].mean_std[j].std;
                                }
                           }
                           insert_angle(angle_value[m].fst->pdb_chnid(),
                               angle_value[m].fst->pdb_resnam(),
                               angle_value[m].fst->pdb_resnum(),
                               angle_value[m].fst->ins_code(), angle1[i].Rep_Name,
                               angle1[i].type, val, ept, std, mol_id,
                               angle_value[m].fst, angle_value[m].snd, angle_value[m].trd);
                      }
                      delete [] angle_value;
                      angle_value = NULL;
                 }
            }
            if (at1) delete [] at1;
            if (at2) delete [] at2;
            if (at3) delete [] at3;
       }
       for (i = 0; i < n_angle2; i++) {
            at1 = prev->find_atom(angle2[i].at1, num1);
            at2 = curr->find_atom(angle2[i].at2, num2);
            at3 = curr->find_atom(angle2[i].at3, num3);
            if (at1 && at2 && at3) {
                 angle_value = cal_angles(at1, num1, at2, num2, at3, num3);
                 if (angle_value) {
                      for (m = 0; m < num1 * num2 * num3; m++) {
                           if (!angle_value[m].fst || !angle_value[m].snd ||
                               !angle_value[m].trd) break;
                           val = angle_value[m].val;
                           ept = angle2[i].mean_std[0].val;
                           std = angle2[i].mean_std[0].std;
                           diff = fabs(val - ept);
                           for (j = 1; j < angle2[i].num; j++) {
                                if (fabs(val - angle2[i].mean_std[j].val) < diff) {
                                     diff = fabs(val - angle2[i].mean_std[j].val);
                                     ept = angle2[i].mean_std[j].val;
                                     std = angle2[i].mean_std[j].std;
                                }
                           }
                           insert_angle(angle_value[m].fst->pdb_chnid(),
                               angle_value[m].snd->pdb_resnam(),
                               angle_value[m].snd->pdb_resnum(),
                               angle_value[m].snd->ins_code(), angle2[i].Rep_Name,
                               angle2[i].type, val, ept, std, mol_id,
                               angle_value[m].fst, angle_value[m].snd, angle_value[m].trd);
                      }
                      delete [] angle_value;
                      angle_value = NULL;
                 }
            }
            if (at1) delete [] at1;
            if (at2) delete [] at2;
            if (at3) delete [] at3;
       }
       return is_link;
}

double _VALIDATION::add_torsions(_ATOM **atom1, _ATOM **atom2, _ATOM **atom3,
         _ATOM **atom4, const int num_a, const int num_b, const int num_c,
         const int num_d, const char *Rep_Name, const char *type,
         const int chain_type, const int mol_id)
{
       int i, num;
       double torsion = -360.0, val;
       _TORSION_VALUE *tor_value = NULL;

       if (atom1 == NULL || atom2 == NULL || atom3 == NULL || atom4 == NULL)
            return torsion;

       tor_value = cal_torsions(num, atom1, num_a, atom2, num_b, atom3, num_c, atom4, num_d);
       if (tor_value) {
            for (i = 0; i < num; i++) {
                 val = tor_value[i].val;
                 if (chain_type == ATOMN_TOKEN)
                      val = mod(val, 0, 360);
                 else val = mod(val, -180, 180);
                 if (torsion < -359.0) torsion = val;
                 insert_torsion(tor_value[i].trd->pdb_chnid(), tor_value[i].trd->pdb_resnam(),
                     tor_value[i].trd->pdb_resnum(), tor_value[i].trd->ins_code(), Rep_Name,
                     type, val, mol_id, tor_value[i].fst, tor_value[i].snd, tor_value[i].trd,
                     tor_value[i].fth);
            }
            delete [] tor_value;
            tor_value = NULL;
       }
       return torsion;
}

static void add_single_bond_values(FILE *fp, int &first_write, const int exchange_out,
            const int mol_id, _RESIDUE *residue, const int first)
{
       CifString cifstring;
       cifstring.Copy(residue->ResName());
       cifstring += residue->type();
       MONOMER_DB *exist = find_monomer_db(cifstring);
       if (!exist) return;
       int n_bond = exist->mon->nbond;
       MONOMER_BOND *bond = exist->mon->bond;

       int i = 0, j = 0, m = 0, check;
       int num1 = 0, num2 = 0;
       double val = 0, ept = 0, diff = 0;
       _ATOM **at1 = NULL, **at2 = NULL;
       _BOND_VALUE *bond_value = NULL;

       for (i = 0; i < n_bond; i++) {
            at1 = residue->find_atom(bond[i].at1, num1);
            at2 = residue->find_atom(bond[i].at2, num2);
            check = 1;
            if (residue->type() == MONOMER_TYPE_A_A) {
                if (!strcmp(bond[i].at1, "N") && !strcmp(bond[i].at2, "CA")) {
                     check = 0;
                     if (first == 1) check = 1;
                } else if (!strcmp(bond[i].at1, "C") && !strcmp(bond[i].at2, "O")) {
                     check = 0;
                     if (first == 2) check = 1;
                }
            }
            if (at1 && at2) {
                 bond_value = cal_bonds(at1, num1, at2, num2);
                 if (bond_value) {
                      for (m = 0; m < num1 * num2; m++) {
                           if (!bond_value[m].fst || !bond_value[m].snd) continue;
                           if (bond_value[m].val > 1.75) continue; 

                           if (first_write) {
                                first_write = 0;
                                fprintf(fp, "#\nloop_\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_bond.pdbx_PDB_model_num\n");
                                else fprintf(fp, "_geom_bond.ndb_model_num\n");
                                fprintf(fp, "_geom_bond.atom_site_id_1\n");
                                fprintf(fp, "_geom_bond.atom_site_label_atom_id_1\n");
                                fprintf(fp, "_geom_bond.atom_site_label_asym_id_1\n");
                                fprintf(fp, "_geom_bond.atom_site_label_comp_id_1\n");
                                fprintf(fp, "_geom_bond.atom_site_label_seq_id_1\n");
                                fprintf(fp, "_geom_bond.atom_site_label_alt_id_1\n");
                                fprintf(fp, "_geom_bond.site_symmetry_1\n");
                                fprintf(fp, "_geom_bond.atom_site_auth_atom_id_1\n");
                                fprintf(fp, "_geom_bond.atom_site_auth_asym_id_1\n");
                                fprintf(fp, "_geom_bond.atom_site_auth_comp_id_1\n");
                                fprintf(fp, "_geom_bond.atom_site_auth_seq_id_1\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_bond.pdbx_atom_site_PDB_ins_code_1\n");
                                else fprintf(fp, "_geom_bond.ndb_atom_site_PDB_ins_code_1\n");
                                fprintf(fp, "_geom_bond.atom_site_id_2\n");
                                fprintf(fp, "_geom_bond.atom_site_label_atom_id_2\n");
                                fprintf(fp, "_geom_bond.atom_site_label_asym_id_2\n");
                                fprintf(fp, "_geom_bond.atom_site_label_comp_id_2\n");
                                fprintf(fp, "_geom_bond.atom_site_label_seq_id_2\n");
                                fprintf(fp, "_geom_bond.atom_site_label_alt_id_2\n");
                                fprintf(fp, "_geom_bond.site_symmetry_2\n");
                                fprintf(fp, "_geom_bond.atom_site_auth_atom_id_2\n");
                                fprintf(fp, "_geom_bond.atom_site_auth_asym_id_2\n");
                                fprintf(fp, "_geom_bond.atom_site_auth_comp_id_2\n");
                                fprintf(fp, "_geom_bond.atom_site_auth_seq_id_2\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_bond.pdbx_atom_site_PDB_ins_code_2\n");
                                else fprintf(fp, "_geom_bond.ndb_atom_site_PDB_ins_code_2\n");
                                fprintf(fp, "_geom_bond.dist\n");
                                fprintf(fp, "_geom_bond.dist_esd\n");
                           }
                           val = bond_value[m].val;
                           ept = bond[i].mean_std[0].val;
                           diff = fabs(val - ept);
                           if (check) {
                                for (j = 1; j < bond[i].num; j++) {
                                     if (fabs(val - bond[i].mean_std[j].val) < diff) {
                                          diff = fabs(val - bond[i].mean_std[j].val);
                                          ept = bond[i].mean_std[j].val;
                                     }
                                }
                           }
                           fprintf(fp, "%d ", mol_id);
                           write_atom_to_file(fp, bond_value[m].fst, "1_555");
                           fprintf(fp, " ");
                           write_atom_to_file(fp, bond_value[m].snd, "1_555");
                           fprintf(fp, " %.3f %.3f\n", val, (val - ept));
                      }
                      delete [] bond_value;
                      bond_value = NULL;
                 }
            }
            if (at1) delete [] at1;
            if (at2) delete [] at2;
       }
}

static void add_mutual_bond_values(FILE *fp, int &first_write, const int exchange_out,
             const int mol_id, const int chain_type, _RESIDUE *prev, _RESIDUE *curr)
{
       MONOMER_BOND *bond = NULL;
       int m = 0, num1 = 0, num2 = 0;
       _ATOM **at1 = NULL, **at2 = NULL;
       _BOND_VALUE *bond_value = NULL;

       if (chain_type == ATOMP_TOKEN) {
            if (!strcmp(curr->ResName(), "GLY")) {
                 bond = &A_LINK;
            } else if (!strcmp(curr->ResName(), "PRO")) {
                 bond = &PRO_LINK;
            } else {
                 bond = &A_LINK;
            }
       } else if (chain_type == ATOMN_TOKEN) {
            bond = &N_LINK;
       }
       if (!bond) return;

       at1 = prev->find_atom(bond->at1, num1);
       at2 = curr->find_atom(bond->at2, num2);
       if (at1 && at2) {
            bond_value = cal_bonds(at1, num1, at2, num2);
            if (bond_value) {
                 for (m = 0; m < num1 * num2; m++) {
                      if (!bond_value[m].fst || !bond_value[m].snd) continue;
                      if (bond_value[m].val > 1.75) continue;

                      if (first_write) {
                           first_write = 0;
                           fprintf(fp, "#\nloop_\n");
                           if (exchange_out)
                                fprintf(fp, "_geom_bond.pdbx_PDB_model_num\n");
                           else fprintf(fp, "_geom_bond.ndb_model_num\n");
                           fprintf(fp, "_geom_bond.atom_site_id_1\n");
                           fprintf(fp, "_geom_bond.atom_site_label_atom_id_1\n");
                           fprintf(fp, "_geom_bond.atom_site_label_asym_id_1\n");
                           fprintf(fp, "_geom_bond.atom_site_label_comp_id_1\n");
                           fprintf(fp, "_geom_bond.atom_site_label_seq_id_1\n");
                           fprintf(fp, "_geom_bond.atom_site_label_alt_id_1\n");
                           fprintf(fp, "_geom_bond.site_symmetry_1\n");
                           fprintf(fp, "_geom_bond.atom_site_auth_atom_id_1\n");
                           fprintf(fp, "_geom_bond.atom_site_auth_asym_id_1\n");
                           fprintf(fp, "_geom_bond.atom_site_auth_comp_id_1\n");
                           fprintf(fp, "_geom_bond.atom_site_auth_seq_id_1\n");
                           if (exchange_out)
                                fprintf(fp, "_geom_bond.pdbx_atom_site_PDB_ins_code_1\n");
                           else fprintf(fp, "_geom_bond.ndb_atom_site_PDB_ins_code_1\n");
                           fprintf(fp, "_geom_bond.atom_site_id_2\n");
                           fprintf(fp, "_geom_bond.atom_site_label_atom_id_2\n");
                           fprintf(fp, "_geom_bond.atom_site_label_asym_id_2\n");
                           fprintf(fp, "_geom_bond.atom_site_label_comp_id_2\n");
                           fprintf(fp, "_geom_bond.atom_site_label_seq_id_2\n");
                           fprintf(fp, "_geom_bond.atom_site_label_alt_id_2\n");
                           fprintf(fp, "_geom_bond.site_symmetry_2\n");
                           fprintf(fp, "_geom_bond.atom_site_auth_atom_id_2\n");
                           fprintf(fp, "_geom_bond.atom_site_auth_asym_id_2\n");
                           fprintf(fp, "_geom_bond.atom_site_auth_comp_id_2\n");
                           fprintf(fp, "_geom_bond.atom_site_auth_seq_id_2\n");
                           if (exchange_out)
                                fprintf(fp, "_geom_bond.pdbx_atom_site_PDB_ins_code_2\n");
                           else fprintf(fp, "_geom_bond.ndb_atom_site_PDB_ins_code_2\n");
                           fprintf(fp, "_geom_bond.dist\n");
                           fprintf(fp, "_geom_bond.dist_esd\n");
                      }
                      fprintf(fp, "%d ", mol_id);
                      write_atom_to_file(fp, bond_value[m].fst, "1_555");
                      fprintf(fp, " ");
                      write_atom_to_file(fp, bond_value[m].snd, "1_555");
                      fprintf(fp, " %.3f %.3f\n", bond_value[m].val,
                                       (bond_value[m].val - bond->mean_std[0].val));
                 }
                 delete [] bond_value;
                 bond_value = NULL;
            }
       }
       if (at1) delete [] at1;
       if (at2) delete [] at2;
}

static void add_single_angle_values(FILE *fp, int &first_write, const int exchange_out,
                     const int mol_id, _RESIDUE *residue, const int first)
{
       CifString cifstring;
       cifstring.Copy(residue->ResName());
       cifstring += residue->type();
       MONOMER_DB *exist = find_monomer_db(cifstring);
       if (!exist) return;
       int n_angle = exist->mon->nangle;
       MONOMER_ANGLE *angle = exist->mon->angle;

       int i = 0, j = 0, m = 0, check;
       int num1 = 0, num2 = 0, num3 = 0;
       double val = 0, ept = 0, diff = 0;
       _ATOM **at1 = NULL, **at2 = NULL, **at3 = NULL;
       _ANGLE_VALUE *angle_value = NULL;

       for (i = 0; i < n_angle; i++) {
            at1 = residue->find_atom(angle[i].at1, num1);
            at2 = residue->find_atom(angle[i].at2, num2);
            at3 = residue->find_atom(angle[i].at3, num3);
            check = 1;
            if (residue->type() == MONOMER_TYPE_A_A && !strcmp(angle[i].at1, "CA") &&
                !strcmp(angle[i].at2, "C") && !strcmp(angle[i].at3, "O")) {
                 check = 0;
                 if (first == 2) check = 1;
            }

            if (at1 && at2 && at3) {
                 angle_value = cal_angles(at1, num1, at2, num2, at3, num3);
                 if (angle_value) {
                      for (m = 0; m < num1 * num2 * num3; m++) {
                           if (!angle_value[m].fst || !angle_value[m].snd ||
                               !angle_value[m].trd) continue;

                           if (first_write) {
                                first_write = 0;
                                fprintf(fp, "#\nloop_\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_angle.pdbx_PDB_model_num\n");
                                else fprintf(fp, "_geom_angle.ndb_model_num\n");
                                fprintf(fp, "_geom_angle.atom_site_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_atom_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_asym_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_comp_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_seq_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_alt_id_1\n");
                                fprintf(fp, "_geom_angle.site_symmetry_1\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_atom_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_asym_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_comp_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_seq_id_1\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_angle.pdbx_atom_site_PDB_ins_code_1\n");
                                else fprintf(fp, "_geom_angle.ndb_atom_site_PDB_ins_code_1\n");
                                fprintf(fp, "_geom_angle.atom_site_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_atom_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_asym_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_comp_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_seq_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_alt_id_2\n");
                                fprintf(fp, "_geom_angle.site_symmetry_2\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_atom_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_asym_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_comp_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_seq_id_2\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_angle.pdbx_atom_site_PDB_ins_code_2\n");
                                else fprintf(fp, "_geom_angle.ndb_atom_site_PDB_ins_code_2\n");
                                fprintf(fp, "_geom_angle.atom_site_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_atom_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_asym_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_comp_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_seq_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_alt_id_3\n");
                                fprintf(fp, "_geom_angle.site_symmetry_3\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_atom_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_asym_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_comp_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_seq_id_3\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_angle.pdbx_atom_site_PDB_ins_code_3\n");
                                else fprintf(fp, "_geom_angle.ndb_atom_site_PDB_ins_code_3\n");
                                fprintf(fp, "_geom_angle.value\n");
                                fprintf(fp, "_geom_angle.value_esd\n");
                           }
                           val = angle_value[m].val;
                           ept = angle[i].mean_std[0].val;
                           diff = fabs(val - ept);
                           if (check) {
                                for (j = 1; j < angle[i].num; j++) {
                                     if (fabs(val - angle[i].mean_std[j].val) < diff) {
                                          diff = fabs(val - angle[i].mean_std[j].val);
                                          ept = angle[i].mean_std[j].val;
                                     }
                                }
                           }
                           fprintf(fp, "%d ", mol_id);
                           write_atom_to_file(fp, angle_value[m].fst, "1_555");
                           fprintf(fp, " ");
                           write_atom_to_file(fp, angle_value[m].snd, "1_555");
                           fprintf(fp, " ");
                           write_atom_to_file(fp, angle_value[m].trd, "1_555");
                           fprintf(fp, " %.3f %.3f\n", val, (val - ept));
                      }
                      delete [] angle_value;
                      angle_value = NULL;
                 }
            }
            if (at1) delete [] at1;
            if (at2) delete [] at2;
            if (at3) delete [] at3;
       }
}

static void add_mutual_angle_values(FILE *fp, int &first_write, const int exchange_out,
            const int mol_id, const int chain_type, _RESIDUE *prev, _RESIDUE *curr)
{
       MONOMER_ANGLE *angle1 = NULL;
       MONOMER_ANGLE *angle2 = NULL;
       int n_angle1 = 0;
       int n_angle2 = 0;
       int i = 0, j = 0, m = 0;
       int num1 = 0, num2 = 0, num3 = 0;
       double val = 0, ept = 0, diff = 0;
       _ATOM **at1 = NULL, **at2 = NULL, **at3 = NULL;
       _ANGLE_VALUE *angle_value = NULL;

       if (!is_residues_linked(chain_type, prev, curr)) return;

       if (chain_type == ATOMP_TOKEN) {
            if (!strcmp(curr->ResName(), "GLY")) {
                 angle1 = GLY_ANG1;
                 angle2 = &GLY_ANG2;
                 n_angle1 = 2;
                 n_angle2 = 1;
            } else if (!strcmp(curr->ResName(), "PRO")) {
                 angle1 = PRO_ANG1;
                 angle2 = PRO_ANG2;
                 n_angle1 = 2;
                 n_angle2 = 2;
            } else {
                 angle1 = AA_ANG1;
                 angle2 = &AA_ANG2;
                 n_angle1 = 2;
                 n_angle2 = 1;
            }
       } else if (chain_type == ATOMN_TOKEN) {
            angle1 = &NN_ANG1;
            angle2 = NN_ANG2;
            n_angle1 = 1;
            n_angle2 = 3;
       }

       for (i = 0; i < n_angle1; i++) {
            at1 = prev->find_atom(angle1[i].at1, num1);
            at2 = prev->find_atom(angle1[i].at2, num2);
            at3 = curr->find_atom(angle1[i].at3, num3);
            if (at1 && at2 && at3) {
                 angle_value = cal_angles(at1, num1, at2, num2, at3, num3);
                 if (angle_value) {
                      for (m = 0; m < num1 * num2 * num3; m++) {
                           if (!angle_value[m].fst || !angle_value[m].snd ||
                               !angle_value[m].trd) continue;

                           if (first_write) {
                                first_write = 0;
                                fprintf(fp, "#\nloop_\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_angle.pdbx_PDB_model_num\n");
                                else fprintf(fp, "_geom_angle.ndb_model_num\n");
                                fprintf(fp, "_geom_angle.atom_site_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_atom_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_asym_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_comp_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_seq_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_alt_id_1\n");
                                fprintf(fp, "_geom_angle.site_symmetry_1\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_atom_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_asym_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_comp_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_seq_id_1\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_angle.pdbx_atom_site_PDB_ins_code_1\n");
                                else fprintf(fp, "_geom_angle.ndb_atom_site_PDB_ins_code_1\n");
                                fprintf(fp, "_geom_angle.atom_site_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_atom_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_asym_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_comp_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_seq_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_alt_id_2\n");
                                fprintf(fp, "_geom_angle.site_symmetry_2\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_atom_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_asym_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_comp_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_seq_id_2\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_angle.pdbx_atom_site_PDB_ins_code_2\n");
                                else fprintf(fp, "_geom_angle.ndb_atom_site_PDB_ins_code_2\n");
                                fprintf(fp, "_geom_angle.atom_site_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_atom_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_asym_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_comp_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_seq_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_alt_id_3\n");
                                fprintf(fp, "_geom_angle.site_symmetry_3\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_atom_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_asym_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_comp_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_seq_id_3\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_angle.pdbx_atom_site_PDB_ins_code_3\n");
                                else fprintf(fp, "_geom_angle.ndb_atom_site_PDB_ins_code_3\n");
                                fprintf(fp, "_geom_angle.value\n");
                                fprintf(fp, "_geom_angle.value_esd\n");
                           }
                           val = angle_value[m].val;
                           ept = angle1[i].mean_std[0].val;
                           diff = fabs(val - ept);
                           for (j = 1; j < angle1[i].num; j++) {
                                if (fabs(val - angle1[i].mean_std[j].val) < diff) {
                                     diff = fabs(val - angle1[i].mean_std[j].val);
                                     ept = angle1[i].mean_std[j].val;
                                }
                           }
                           fprintf(fp, "%d ", mol_id);
                           write_atom_to_file(fp, angle_value[m].fst, "1_555");
                           fprintf(fp, " ");
                           write_atom_to_file(fp, angle_value[m].snd, "1_555");
                           fprintf(fp, " ");
                           write_atom_to_file(fp, angle_value[m].trd, "1_555");
                           fprintf(fp, " %.3f %.3f\n", val, (val - ept));
                      }
                      delete [] angle_value;
                      angle_value = NULL;
                 }
            }
            if (at1) delete [] at1;
            if (at2) delete [] at2;
            if (at3) delete [] at3;
       }
       for (i = 0; i < n_angle2; i++) {
            at1 = prev->find_atom(angle2[i].at1, num1);
            at2 = curr->find_atom(angle2[i].at2, num2);
            at3 = curr->find_atom(angle2[i].at3, num3);
            if (at1 && at2 && at3) {
                 angle_value = cal_angles(at1, num1, at2, num2, at3, num3);
                 if (angle_value) {
                      for (m = 0; m < num1 * num2 * num3; m++) {
                           if (!angle_value[m].fst || !angle_value[m].snd ||
                               !angle_value[m].trd) continue;

                           if (first_write) {
                                first_write = 0;
                                fprintf(fp, "#\nloop_\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_angle.pdbx_PDB_model_num\n");
                                else fprintf(fp, "_geom_angle.ndb_model_num\n");
                                fprintf(fp, "_geom_angle.atom_site_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_atom_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_asym_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_comp_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_seq_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_label_alt_id_1\n");
                                fprintf(fp, "_geom_angle.site_symmetry_1\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_atom_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_asym_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_comp_id_1\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_seq_id_1\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_angle.pdbx_atom_site_PDB_ins_code_1\n");
                                else fprintf(fp, "_geom_angle.ndb_atom_site_PDB_ins_code_1\n");
                                fprintf(fp, "_geom_angle.atom_site_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_atom_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_asym_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_comp_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_seq_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_label_alt_id_2\n");
                                fprintf(fp, "_geom_angle.site_symmetry_2\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_atom_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_asym_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_comp_id_2\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_seq_id_2\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_angle.pdbx_atom_site_PDB_ins_code_2\n");
                                else fprintf(fp, "_geom_angle.ndb_atom_site_PDB_ins_code_2\n");
                                fprintf(fp, "_geom_angle.atom_site_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_atom_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_asym_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_comp_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_seq_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_label_alt_id_3\n");
                                fprintf(fp, "_geom_angle.site_symmetry_3\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_atom_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_asym_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_comp_id_3\n");
                                fprintf(fp, "_geom_angle.atom_site_auth_seq_id_3\n");
                                if (exchange_out)
                                     fprintf(fp, "_geom_angle.pdbx_atom_site_PDB_ins_code_3\n");
                                else fprintf(fp, "_geom_angle.ndb_atom_site_PDB_ins_code_3\n");
                                fprintf(fp, "_geom_angle.value\n");
                                fprintf(fp, "_geom_angle.value_esd\n");
                           }
                           val = angle_value[m].val;
                           ept = angle2[i].mean_std[0].val;
                           diff = fabs(val - ept);
                           for (j = 1; j < angle2[i].num; j++) {
                                if (fabs(val - angle2[i].mean_std[j].val) < diff) {
                                     diff = fabs(val - angle2[i].mean_std[j].val);
                                     ept = angle2[i].mean_std[j].val;
                                }
                           }
                           fprintf(fp, "%d ", mol_id);
                           write_atom_to_file(fp, angle_value[m].fst, "1_555");
                           fprintf(fp, " ");
                           write_atom_to_file(fp, angle_value[m].snd, "1_555");
                           fprintf(fp, " ");
                           write_atom_to_file(fp, angle_value[m].trd, "1_555");
                           fprintf(fp, " %.3f %.3f\n", val, (val - ept));
                      }
                      delete [] angle_value;
                      angle_value = NULL;
                 }
            }
            if (at1) delete [] at1;
            if (at2) delete [] at2;
            if (at3) delete [] at3;
       }
}

static int is_residues_linked(const int chain_type, _RESIDUE *prev, _RESIDUE *curr)
{
       MONOMER_BOND *bond = NULL;
       int is_link = 0;
       int m = 0, num1 = 0, num2 = 0;
       _ATOM **at1 = NULL, **at2 = NULL;
       _BOND_VALUE *bond_value = NULL;

       if (chain_type == ATOMP_TOKEN) {
            if (!strcmp(curr->ResName(), "GLY")) {
                 bond = &A_LINK;
            } else if (!strcmp(curr->ResName(), "PRO")) {
                 bond = &PRO_LINK;
            } else {
                 bond = &A_LINK;
            }
       } else if (chain_type == ATOMN_TOKEN) {
            bond = &N_LINK;
       }
       if (!bond) return is_link;

       at1 = prev->find_atom(bond->at1, num1);
       at2 = curr->find_atom(bond->at2, num2);
       if (at1 && at2) {
            bond_value = cal_bonds(at1, num1, at2, num2);
            if (bond_value) {
                 for (m = 0; m < num1 * num2; m++) {
                      if (!bond_value[m].fst || !bond_value[m].snd) continue;
                      if (bond_value[m].val < 1.75) {
                           is_link = 1;
                           break; 
                      }
                 }
                 delete [] bond_value;
                 bond_value = NULL;
            }
       }
       if (at1) delete [] at1;
       if (at2) delete [] at2;

       return is_link;
}

static void add_torsions_fp(FILE *fp, int &first_write, const int exchange_out, _ATOM **atom1,
         _ATOM **atom2, _ATOM **atom3, _ATOM **atom4, const int num_a, const int num_b,
         const int num_c, const int num_d, const int chain_type, const int mol_id)
{
       int i, num;
       double val;
       _TORSION_VALUE *tor_value = NULL;

       if (atom1 == NULL || atom2 == NULL || atom3 == NULL || atom4 == NULL) return;

       tor_value = cal_torsions(num, atom1, num_a, atom2, num_b, atom3, num_c, atom4, num_d);
       if (!tor_value) return;

       for (i = 0; i < num; i++) {
            if (first_write) {
                 first_write = 0;
                 fprintf(fp, "#\nloop_\n");
                 if (exchange_out)
                      fprintf(fp, "_geom_torsion.pdbx_PDB_model_num\n");
                 else fprintf(fp, "_geom_torsion.ndb_model_num\n");
                 fprintf(fp, "_geom_torsion.atom_site_id_1\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_atom_id_1\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_asym_id_1\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_comp_id_1\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_seq_id_1\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_alt_id_1\n");
                 fprintf(fp, "_geom_torsion.site_symmetry_1\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_atom_id_1\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_asym_id_1\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_comp_id_1\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_seq_id_1\n");
                 if (exchange_out)
                      fprintf(fp, "_geom_torsion.pdbx_atom_site_PDB_ins_code_1\n");
                 else fprintf(fp, "_geom_torsion.ndb_atom_site_PDB_ins_code_1\n");
                 fprintf(fp, "_geom_torsion.atom_site_id_2\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_atom_id_2\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_asym_id_2\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_comp_id_2\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_seq_id_2\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_alt_id_2\n");
                 fprintf(fp, "_geom_torsion.site_symmetry_2\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_atom_id_2\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_asym_id_2\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_comp_id_2\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_seq_id_2\n");
                 if (exchange_out)
                      fprintf(fp, "_geom_torsion.pdbx_atom_site_PDB_ins_code_2\n");
                 else fprintf(fp, "_geom_torsion.ndb_atom_site_PDB_ins_code_2\n");
                 fprintf(fp, "_geom_torsion.atom_site_id_3\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_atom_id_3\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_asym_id_3\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_comp_id_3\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_seq_id_3\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_alt_id_3\n");
                 fprintf(fp, "_geom_torsion.site_symmetry_3\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_atom_id_3\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_asym_id_3\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_comp_id_3\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_seq_id_3\n");
                 if (exchange_out)
                      fprintf(fp, "_geom_torsion.pdbx_atom_site_PDB_ins_code_3\n");
                 else fprintf(fp, "_geom_torsion.ndb_atom_site_PDB_ins_code_3\n");
                 fprintf(fp, "_geom_torsion.atom_site_id_4\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_atom_id_4\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_asym_id_4\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_comp_id_4\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_seq_id_4\n");
                 fprintf(fp, "_geom_torsion.atom_site_label_alt_id_4\n");
                 fprintf(fp, "_geom_torsion.site_symmetry_4\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_atom_id_4\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_asym_id_4\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_comp_id_4\n");
                 fprintf(fp, "_geom_torsion.atom_site_auth_seq_id_4\n");
                 if (exchange_out)
                      fprintf(fp, "_geom_torsion.pdbx_atom_site_PDB_ins_code_4\n");
                 else fprintf(fp, "_geom_torsion.ndb_atom_site_PDB_ins_code_4\n");
                 fprintf(fp, "_geom_torsion.value\n");
            }
            val = tor_value[i].val;
            if (chain_type == ATOMN_TOKEN)
                 val = mod(val, 0, 360);
            else val = mod(val, -180, 180);

            fprintf(fp, "%d ", mol_id);
            write_atom_to_file(fp, tor_value[i].fst, "1_555");
            fprintf(fp, " ");
            write_atom_to_file(fp, tor_value[i].snd, "1_555");
            fprintf(fp, "\n");
            write_atom_to_file(fp, tor_value[i].trd, "1_555");
            fprintf(fp, " ");
            write_atom_to_file(fp, tor_value[i].fth, "1_555");
            fprintf(fp, " %.3f\n", val);
       }
       delete [] tor_value;
}

